/**
 * Copyright © 2009-2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * StCore library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StCore library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StApplicationImpl.h"

#include <StCore/StCore.h>
#include <StCore/StWindowInterface.h>
#include <StSettings/StSettings.h>
#include <StStrings/stConsole.h>
#include <StGL/StGL.h>
#include <StGL/StGLEnums.h>
#include <StGLStereo/StFormatEnum.h>
#include <StFile/StFileNode.h>

// Exported class-methods wrappers in StCore library (here!)
ST_EXPORT stBool_t StWindow_getValue(StWindowInterface* , const size_t& , size_t* );

StApplicationImpl::StApplicationImpl()
: stRenderer(),
  sViewPath(),
  stParamRenderer(),
  stParamDeviceString(),
  stParamDeviceInt(StRendererInfo::DEVICE_AUTO),
  stDrawerInfo(),
  stOpenFileInfo(),
  stNativeWinParent(),
  showHelpString(),
  bOpened(false),
  bQuit(false) {
    //
    stMemSet(stMessages, 0, sizeof(stMessages));
    stMemSet(&stNativeWinParent, 0, sizeof(stNativeWinParent));

    // add additional paths
    // TODO (Kirill Gavrilov#1) linux library path
#if(defined(_WIN32) || defined(__WIN32__))
    // requires Windows XP with SP1 or higher!
    StStringUtfWide aStCoreFolder = StProcess::getStCoreFolder().toUtfWide();
    SetDllDirectoryW(aStCoreFolder.toCString());
#endif
    parseProcessArguments();
}

StApplicationImpl::~StApplicationImpl() {
    StSettings::FREE();
}

bool StApplicationImpl::chooseRendererPlugin() {
    StArrayList<StRendererInfo> renderersList = StCore::getRenderersList();
    const StString ST_SETTING_RENDERER_AUTO = "rendererPluginAuto";
    const StString ST_SETTING_RENDERER   = "rendererPlugin";
    const StString ST_SETTING_AUTO_VALUE = "Auto";

    if(renderersList.size() < 1 && (stParamRenderer.isEmpty() || stParamRenderer.isEqualsIgnoreCase(ST_SETTING_AUTO_VALUE))) {
        return false;
    }

    StSettings stSettings("sview");
    bool isOutModeAuto = true; // AUTO by default

    if(stParamRenderer.isEqualsIgnoreCase(ST_SETTING_AUTO_VALUE)) {
        // force autodetection
        stSettings.saveString(ST_SETTING_RENDERER, ST_SETTING_AUTO_VALUE);
        stSettings.saveBool(ST_SETTING_RENDERER_AUTO, isOutModeAuto);
    } else if(!stParamRenderer.isEmpty()) {
        // renderer module selected by program attribute
        if(!stParamRenderer.isContains(SYS_FS_SPLITTER)) {
            // got the title
            stParamRenderer = sViewPath + StCore::getRenderersDir() + SYS_FS_SPLITTER + stParamRenderer + ST_DLIB_SUFFIX;
        }
        StRendererInfo stSettedRendererInfo(stParamRenderer);
        if(stSettedRendererInfo.isValid() && stRenderer.InitLibrary(stSettedRendererInfo.getPath())) {
            // save plugin info
            if(renderersList.contains(stSettedRendererInfo)) {
                stSettings.saveString(ST_SETTING_RENDERER, stSettedRendererInfo.getTitle());
            } else {
                stSettings.saveString(ST_SETTING_RENDERER, stSettedRendererInfo.getPath());
            }
            isOutModeAuto = false;
            stSettings.saveBool(ST_SETTING_RENDERER_AUTO, isOutModeAuto);
            return true;
        } else {
            stError(StString("Failed to load StRenderer Plugin:\n'") + stSettedRendererInfo.getPath() + '\'');
        }
    }

    // load saved parameter
    stSettings.loadBool(ST_SETTING_RENDERER_AUTO, isOutModeAuto);
    if(!isOutModeAuto) {
        stSettings.loadString(ST_SETTING_RENDERER, stParamRenderer);
        if(stParamRenderer.isEqualsIgnoreCase(ST_SETTING_AUTO_VALUE)) {
            // still auto in data!
            isOutModeAuto = true;
            stSettings.saveBool(ST_SETTING_RENDERER_AUTO, isOutModeAuto);
        } else {
            if(!stParamRenderer.isContains(SYS_FS_SPLITTER)) {
                // got the title
                stParamRenderer = sViewPath + StCore::getRenderersDir() + SYS_FS_SPLITTER + stParamRenderer + ST_DLIB_SUFFIX;
            }
            StRendererInfo stSettedRendererInfo(stParamRenderer);
            if(stSettedRendererInfo.isValid() && stRenderer.InitLibrary(stSettedRendererInfo.getPath())) {
                // save plugin info
                if(renderersList.contains(stSettedRendererInfo)) {
                    stSettings.saveString(ST_SETTING_RENDERER, stSettedRendererInfo.getTitle());
                } else {
                    stSettings.saveString(ST_SETTING_RENDERER, stSettedRendererInfo.getPath());
                }
                stSettings.saveBool(ST_SETTING_RENDERER_AUTO, isOutModeAuto);
                return true;
            } else {
                stError(StString("Failed to load StRenderer Plugin:\n'") + stSettedRendererInfo.getPath() + '\'');
            }
        }
    }
    stSettings.saveString(ST_SETTING_RENDERER, ST_SETTING_AUTO_VALUE);
    stSettings.saveInt32(ST_SETTING_RENDERER_AUTO, isOutModeAuto);

    renderersList.sort();
    for(size_t rId = renderersList.size() - 1; rId < size_t(-1); --rId) {
        if(stRenderer.InitLibrary(renderersList[rId].getPath())) {
            return true;
        }
    }
    return false;
}

bool StApplicationImpl::create(const StNativeWin_t* stNativeWinParent) {

    // load settings library
    if(StSettings::INIT() != STERROR_LIBNOERROR) {
        stError("StSettings plugin not available!");
        return false;
    }

    if(!chooseRendererPlugin()) {
        stError("StRenderer plugin (for stereo-device support) not available!");
        return false;
    }

    if(stNativeWinParent != NULL) {
        stMemCpy(&this->stNativeWinParent, stNativeWinParent, sizeof(this->stNativeWinParent));
    } else {
        stMemSet(&this->stNativeWinParent, 0, sizeof(this->stNativeWinParent));
    }

    // TODO (Kirill Gavrilov#9) try open another renderer if initialization failes
    stRenderer.Instantiate();
    bool result = stRenderer.init(stParamDeviceInt, stNativeWinParent);
    ST_DEBUG_LOG(stglInfo());
    return result;
}

bool StApplicationImpl::open(const StOpenInfo& stOpenInfo) {
    if(stOpenInfo.getMIME() == StDrawerInfo::CLOSE_MIME()) {
        if(isOpened()) {
            stRenderer.open(stOpenInfo);
            bQuit = true;
        }
        return true;
    } else if(stOpenInfo.getMIME() == StDrawerInfo::DRAWER_MIME()) {
        stDrawerInfo = stOpenInfo;
        bOpened = stRenderer.open(stOpenInfo);
    } else if(!stOpenInfo.isEmpty()) {
        stOpenFileInfo = stOpenInfo;
        bOpened = stRenderer.open(stOpenInfo);
    } else {

        // use stored info (previous opened file or command line arguments)
        ST_DEBUG_LOG_AT("use command line arguments");
        if(!stDrawerInfo.isEmpty()) {
            if(!stRenderer.open(stDrawerInfo)) {
                bOpened = false;
                return bOpened;
            }
        } else if(!showHelpString.isEmpty()) {
            // just show help
            st::cout << showHelpString;
            stInfo(showHelpString);
            bOpened = false;
            return bOpened;
        }

        bOpened = stOpenFileInfo.isEmpty() || stRenderer.open(stOpenFileInfo);
    }
    return bOpened;
}

void StApplicationImpl::callback(StMessage_t* stMessages) {
    if(!isOpened()) {
        return; // nothing to do
    }
    StMessage_t* usedMessages = stMessages;
    // if application need not parse callback events - use internal buffer
    if(stMessages == NULL) {
        usedMessages = this->stMessages;
    }

    if(bQuit) {
        // force Render to quit
        usedMessages[0].uin = StMessageList::MSG_EXIT;
        usedMessages[1].uin = StMessageList::MSG_NULL;
        // be sure Render plugin process quit correctly
        stRenderer.callback(usedMessages);
        bOpened = false;
        bQuit = false;
        return;
    }

    // common callback call
    stRenderer.callback(usedMessages);
    for(size_t i = 0; usedMessages[i].uin != StMessageList::MSG_NULL; ++i) {
        switch(usedMessages[i].uin) {
            case StMessageList::MSG_EXIT: {
                // TODO (Kirill Gavrilov#1)
                bOpened = false;
                return;
            }
            case StMessageList::MSG_DEVICE_INFO: {
                int newDeviceId(StRendererInfo::DEVICE_AUTO);

                size_t ptrToStruct = 0;
                if(!StWindow_getValue(stRenderer.getStWindow()->getLibImpl(), ST_WIN_DATAKEYS_RENDERER, &ptrToStruct) || (ptrToStruct == 0)) {
                    break;
                }
                StSDOptionsList_t* optionsStruct = (StSDOptionsList_t* )ptrToStruct;
                StString newPluginPath(optionsStruct->curRendererPath);
                newDeviceId = optionsStruct->curDeviceId;

                // force Renderer to quit, plugin should disable this message if parsed it himself
                StOpenInfo stCloseInfo;
                stCloseInfo.setMIME(StDrawerInfo::CLOSE_MIME());
                if(isOpened()) {
                    stRenderer.open(stCloseInfo);
                }
                usedMessages[0].uin = StMessageList::MSG_EXIT;
                usedMessages[1].uin = StMessageList::MSG_NULL;

                // be sure Render plugin process quit correctly
                stRenderer.callback(usedMessages);
                bOpened = false;
                bQuit = false;

                // open new renderer
                stParamRenderer = newPluginPath;
                stParamDeviceInt = newDeviceId;
                stRenderer.Destruct();

                if(!chooseRendererPlugin()) {
                    stError("StRenderer plugin (for stereo-device support) not available!");
                    return;
                }

                stRenderer.Instantiate();
                if(stRenderer.init(stParamDeviceInt, &this->stNativeWinParent)) {
                    /// TODO (Kirill Gavrilov#5) get up-to-date opened file info
                    open(StOpenInfo());
                }
                return;
            }
        }
    }

    // draw iteration
    stRenderer.stglDraw(ST_DRAW_BOTH);
}

static StString getAboutString(const StString& theProgramPath) {
    /// TODO (Kirill Gavrilov#9) call plugins for full help list
    StString anAboutString =
        StString("sView ") + StVersionInfo::getSDKVersionString() + '\n'
        + "Copyright (C) 2007-2011 Kirill Gavrilov (kirill@sview.ru).\n"
        + "Usage: " + theProgramPath + " [options] - file\n"
        + "Available options:\n"
          "  --fullscreen         Open fullscreen\n"
          "  --slideshow          Start slideshow\n"
          "  --in=DRAWER_PATH     Open StDrawer plugin (predefined values: image, video)\n"
          "  --out=RENDERER_PATH  Open StRenderer plugin (auto, StOutAnaglyph, StOutDual,...)\n"
          "  --imageLib=IMGLIB    Setup 3rd-party library for image processing (FFmpeg, FreeImage, DevIL)\n"
          "  --viewMode=MODE      View mode (flat, sphere)\n"
          "  --srcFormat=FORMAT   Setup source format:\n"
          "                       "
            + st::formatToString(ST_V_SRC_AUTODETECT)    + ", "
            + st::formatToString(ST_V_SRC_MONO)          + ", "
            + st::formatToString(ST_V_SRC_SIDE_BY_SIDE)  + ", "
            + st::formatToString(ST_V_SRC_PARALLEL_PAIR) + ",\n"
        + "                       "
            + st::formatToString(ST_V_SRC_OVER_UNDER_LR) + ", "
            + st::formatToString(ST_V_SRC_OVER_UNDER_RL) + ", "
            + st::formatToString(ST_V_SRC_ROW_INTERLACE) + ", "
            + st::formatToString(ST_V_SRC_PAGE_FLIP)     + '\n';
    return anAboutString;
}

void StApplicationImpl::parseProcessArguments() {
    StArrayList<StString> stArguments = StProcess::getArguments();
    StArgumentsMap stDrawerArgs;
    StArgumentsMap stOpenFileArgs;
    size_t filesCount = 0;
    sViewPath = StProcess::getStCoreFolder();
    bool isFilesSection = false;
    const StString ARGUMENT_FILES_SECTION     = '-';
    const StString ARGUMENT_ANY               = "--";
    const StString ARGUMENT_HELP              = "help";
    const StString ARGUMENT_PLUGIN_IN         = "in";
    const StString ARGUMENT_PLUGIN_OUT        = "out";
    const StString ARGUMENT_PLUGIN_OUT_DEVICE = "outDevice";
    const StString ARGUMENT_FILE              = "file";
    const StString ARGUMENT_LEFT_VIEW         = "left";
    const StString ARGUMENT_RIGHT_VIEW        = "right";
    // parse extra parameters
    for(size_t p = 1; p < stArguments.size(); ++p) {
        StString param = stArguments[p];
        ///ST_DEBUG_LOG("param= '" + param + "'");
        if(isFilesSection) {
            // file name
            StString filePath = StProcess::getAbsolutePath(param);
            stOpenFileArgs.add(StArgument(ARGUMENT_FILE + filesCount++, filePath));
            if(!stOpenFileInfo.hasPath()) {
                // first file determines MIME type (needed to autoselect Drawer plugin)
                stOpenFileInfo.setPath(filePath);
            }
        } else if(param == ARGUMENT_FILES_SECTION) {
            isFilesSection = true;
        } else if(param.isStartsWith(ARGUMENT_ANY)) {
            // argument
            StArgument stArg; stArg.parseString(param.subString(2, param.getLength())); // cut suffix --

            if(stArg.getKey().isEqualsIgnoreCase(ARGUMENT_HELP)) {
                showHelpString = getAboutString(stArguments[0]);
            } else if(stArg.getKey().isEqualsIgnoreCase(ARGUMENT_PLUGIN_OUT)) {
                stParamRenderer = stArg.getValue();
                ST_DEBUG_LOG("sView argument, Renderer plugin= '" + stParamRenderer + '\'');
            } else if(stArg.getKey().isEqualsIgnoreCase(ARGUMENT_PLUGIN_OUT_DEVICE)) {
                ST_DEBUG_LOG("sView argument, StereoDevice= '" + stArg.getValue() + '\'');
                if(stUtfTools::isInteger(stArg.getValue())) {
                    stParamDeviceInt = std::atoi(stArg.getValue().toCString());
                } else {
                    stParamDeviceString = stArg.getValue();
                }
            } else if(stArg.getKey().isEqualsIgnoreCase(ARGUMENT_PLUGIN_IN)) {
                // predefined plugins links
                const StString IMAGE_VIEWER("image");
                const StString VIDEO_PLAYER("video");
                if(stArg.getValue().isEqualsIgnoreCase(IMAGE_VIEWER)) {
                    stArg.changeValue() = sViewPath + StCore::getDrawersDir() + SYS_FS_SPLITTER + "StImageViewer" + ST_DLIB_SUFFIX;
                } else if(stArg.getValue().isEqualsIgnoreCase(VIDEO_PLAYER)) {
                    stArg.changeValue() = sViewPath + StCore::getDrawersDir() + SYS_FS_SPLITTER + "StMoviePlayer" + ST_DLIB_SUFFIX;
                } else {
                    StString ext = StFileNode::getExtension(stArg.getValue());
                    if(ext != StString(ST_DLIB_EXTENSION)) {
                        stArg.changeValue() += StString(ST_DLIB_SUFFIX);
                    }
                    if(StFileNode::isRelativePath(stArg.getValue())) {
                        stArg.changeValue() = sViewPath + StCore::getDrawersDir() + SYS_FS_SPLITTER + stArg.getValue();
                    }
                }
                ST_DEBUG_LOG("sView argument, Drawer plugin= '" + stArg.getValue() + '\'');
                stDrawerInfo.setMIME(StDrawerInfo::DRAWER_MIME());
                stDrawerInfo.setPath(stArg.getValue());
            } else if(stArg.getKey().isEqualsIgnoreCase(ARGUMENT_LEFT_VIEW)) {
                // left view
                stArg.setValue(StProcess::getAbsolutePath(stArg.getValue()));
                stOpenFileArgs.add(stArg);
                stOpenFileInfo.setPath(stArg.getValue()); // left file always determines MIME type
            } else if(stArg.getKey().isEqualsIgnoreCase(ARGUMENT_RIGHT_VIEW)) {
                // right view
                stArg.setValue(StProcess::getAbsolutePath(stArg.getValue()));
                stOpenFileArgs.add(stArg);
                if(!stOpenFileInfo.hasPath()) {
                    stOpenFileInfo.setPath(stArg.getValue());
                }
            } else {
                // pass argument unchanged to the Drawer
                stDrawerArgs.add(stArg);
            }
        } else {
            // file name
            StString filePath = StProcess::getAbsolutePath(param);
            stOpenFileArgs.add(StArgument(ARGUMENT_FILE + filesCount++, filePath));
            if(!stOpenFileInfo.hasPath()) {
                // first file determines MIME type (needed to autoselect Drawer plugin)
                stOpenFileInfo.setPath(filePath);
            }
        }
    }

    if(!stDrawerInfo.isEmpty()) {
        stDrawerInfo.setArgumentsMap(stDrawerArgs);
    } else {
        // if no Drawer selected - pass all arguments to open file structure
        for(size_t anArgId = 0; anArgId < stDrawerArgs.size(); ++anArgId) {
            stOpenFileArgs.add(stDrawerArgs.getFromIndex(anArgId));
        }
    }
    stOpenFileInfo.setArgumentsMap(stOpenFileArgs);

    // there are NO default sView plugin to load now!
    if(stDrawerInfo.isEmpty() && stOpenFileInfo.isEmpty()) {
        showHelpString = getAboutString(stArguments.size() > 0 ? stArguments[0] : StString("sView"));
    }
}

// Exported class-methods wpappers.
ST_EXPORT StApplicationInterface* StApplication_new() {
    return new StApplicationImpl();
}

ST_EXPORT void StApplication_del(StApplicationInterface* inst) {
    delete (StApplicationImpl* )inst;
}

ST_EXPORT stBool_t StApplication_isOpened(StApplicationInterface* inst) {
    return ((StApplicationImpl* )inst)->isOpened();
}

ST_EXPORT stBool_t StApplication_create(StApplicationInterface* inst, const StNativeWin_t* stNativeWinParent) {
    return ((StApplicationImpl* )inst)->create(stNativeWinParent);
}

ST_EXPORT stBool_t StApplication_open(StApplicationInterface* inst, const StOpenInfo_t* stOpenInfo) {
    return ((StApplicationImpl* )inst)->open(StOpenInfo(stOpenInfo));
}

ST_EXPORT void StApplication_callback(StApplicationInterface* inst, StMessage_t* stMessages) {
    ((StApplicationImpl* )inst)->callback(stMessages);
}
