/**
 * Copyright © 2007-2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * StCore library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StCore library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StCoreImpl.h"

#include <StFile/StFileNode.h>

namespace {
    static bool HAS_LOGGER_ID = StLogger::IdentifyModule("StCore");
};

StCoreImpl::StCoreImpl()
: stWindow(NULL),
  stDrawer(NULL),
  sViewPath(StProcess::getStCoreFolder()),
  stDrawerPath() {
    //
    stWindow = StWindow_new();

    // Adjust system timer
    // By default Windows2K+ timer has ugly precision
    // Thus - Sleep(1) may be long 14ms!
    // We force best available precision to make Sleep() more adequate
    // This affect whole system while running application!
#if(defined(_WIN32) || defined(__WIN32__))
    TIMECAPS ptc; ptc.wPeriodMax = 0; ptc.wPeriodMin = 0;
    if(timeGetDevCaps(&ptc, sizeof(ptc)) == TIMERR_NOERROR) {
        timeBeginPeriod(ptc.wPeriodMin);
    } else {
        timeBeginPeriod(1);
    }
#endif

    // add additional paths
    // TODO (Kirill Gavrilov#1) linux library path
#if(defined(_WIN32) || defined(__WIN32__))
    // requires Windows XP with SP1 or higher!
    StStringUtfWide aStCoreFolder = StProcess::getStCoreFolder().toUtfWide();
    SetDllDirectoryW(aStCoreFolder.toCString());
#endif
}

StCoreImpl::~StCoreImpl() {
    closeDrawer();
    StWindow_del(stWindow);
#if(defined(_WIN32) || defined(__WIN32__))
    TIMECAPS ptc; ptc.wPeriodMax = 0; ptc.wPeriodMin = 0;
    if(timeGetDevCaps(&ptc, sizeof(ptc)) == TIMERR_NOERROR) {
        timeEndPeriod(ptc.wPeriodMin);
    } else {
        timeEndPeriod(1);
    }
#endif
}

bool StCoreImpl::init(const StString& , const int& , const StNativeWin_t* ) {
    if(!StVersionInfo::checkTimeBomb("sView - Core library")) {
        return false;
    }
    // TODO (Kirill Gavrilov#1) function doesn't called from StRenderer plugins!!!
    return true;
}

void StCoreImpl::closeDrawer() {
    delete stDrawer; stDrawer = NULL;
    stDrawerPath = StString();
}

bool StCoreImpl::open(const StOpenInfo& stOpenInfo) {
    StString pluginFullPath;
    StMIME stOpenMIME = stOpenInfo.getMIME();
    if(stOpenMIME == StDrawerInfo::CLOSE_MIME()) {
        // we got close drawer command
        closeDrawer();
        return true;
    } else if(stOpenMIME == StDrawerInfo::DRAWER_MIME()) {
        // we should open specified drawer
        pluginFullPath = stOpenInfo.getPath();
    } else if(stDrawer != NULL) {
        // use already opened drawer
        return stDrawer->open(stOpenInfo);
    } else {
        // we should autodetect best drawer plugin
        ST_DEBUG_LOG_AT("StCoreImpl::open, filePath= '" + stOpenInfo.getPath() + "'; mime= '" + stOpenInfo.getMIME().toString() + "'");
        StArrayList<StDrawerInfo> stDrawerInfoList = StCore::getDrawersList();
        StString fileExtension = StFileNode::getExtension(stOpenInfo.getPath());

        if(stOpenMIME == StMIME() && !stOpenInfo.hasPath()) {
            stError("Nothing to open with sView!\nCheck program arguments.");
            return false;
        }
        bool isDrawerFound = false;
        // firstly - search by MIME
        if(stOpenMIME != StMIME()) {
            for(size_t drId = 0; !isDrawerFound && (drId < stDrawerInfoList.size()); ++drId) {
                for(size_t drMimeId = 0; drMimeId < stDrawerInfoList[drId].getMIMEList().size(); ++drMimeId) {
                    if(stOpenMIME == stDrawerInfoList[drId].getMIMEList()[drMimeId]) {
                        pluginFullPath = stDrawerInfoList[drId].getPath();
                        isDrawerFound = true;
                        break;
                    }
                }
            }
        }
        // if MIME not available or failed - search by extension
        for(size_t drId = 0; !isDrawerFound && (drId < stDrawerInfoList.size()); ++drId) {
            for(size_t drMimeId = 0; drMimeId < stDrawerInfoList[drId].getMIMEList().size(); ++drMimeId) {
                if(fileExtension.isEqualsIgnoreCase(stDrawerInfoList[drId].getMIMEList()[drMimeId].getExtension())) {
                    pluginFullPath = stDrawerInfoList[drId].getPath();
                    isDrawerFound = true;
                    break;
                }
            }
        }
        if(!isDrawerFound) {
            stError(StString("sView, Extension '") + fileExtension + "' not supported");
            return false;
        }
    }

    // check Drawer already runned
    if(stDrawerPath == pluginFullPath) {
        return stDrawer->open(stOpenInfo);
    }

    if(stDrawer != NULL) {
        // have active Drawer instance
        closeDrawer();
    }

    stDrawer = new StDrawer();
    if(!stDrawer->InitLibrary(pluginFullPath)) {
        StString errString = StString("StDrawer Plugin \"") + pluginFullPath + "\" loading Failed!";
        stError(errString);
        closeDrawer();
        return false;
    }
    stDrawer->Instantiate();

    stWindow->stglMakeCurrent(ST_WIN_MASTER);
    if(!stDrawer->init(stWindow)) {
        stDrawerPath = StString();
        return false;
    }
    if(!stDrawer->open(stOpenInfo)) {
        stDrawerPath = pluginFullPath;
        return false;
    }
    // save current plugin path
    stDrawerPath = pluginFullPath;
    return true;
}

// SDK version was used
ST_EXPORT void getSDKVersion(StVersion* ver) {
    *ver = StVersionInfo::getSDKVersion();
}

ST_EXPORT StRendererInterface* StCore_new() {
    return new StCoreImpl(); }
ST_EXPORT void StCore_del(StRendererInterface* inst) {
    delete (StCoreImpl* )inst; }
ST_EXPORT StWindowInterface* StCore_getStWindow(StRendererInterface* inst) {
    return ((StCoreImpl* )inst)->getStWindow(); }
ST_EXPORT stBool_t StCore_init(StRendererInterface* inst, const stUtf8_t* rendererPath, const int& deviceId, const StNativeWin_t* nativeParent) {
    return ((StCoreImpl* )inst)->init(StString(rendererPath), deviceId, nativeParent); }
ST_EXPORT stBool_t StCore_open(StRendererInterface* inst, const StOpenInfo_t* stOpenInfo) {
    return ((StCoreImpl* )inst)->open(StOpenInfo(stOpenInfo)); }
ST_EXPORT void StCore_callback(StRendererInterface* inst, StMessage_t* stMessages) {
    ((StCoreImpl* )inst)->callback(stMessages); }
ST_EXPORT void StCore_stglDraw(StRendererInterface* inst, unsigned int views) {
    ((StCoreImpl* )inst)->stglDraw(views); }
