/**
 * Copyright © 2010-2012 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLWidgets/StGLImageProgram.h>
#include <StGL/StGLResources.h>

StGLImageProgram::StGLImageProgram(const StString& theTitle)
: StGLMeshProgram(StString("StGLImageProgram, ") + theTitle),
  uniTexMainDataLoc(),
  uniTexUVDataLoc(),
  uniTexSizePxLoc(),
  uniTexelSizePxLoc(),
  uniColorProcessingLoc(),
  uniGammaLoc(),
  fGetColor     (StString("StGLImageProgram::fGetColor, ")      + theTitle),
  fGetColorBlend(StString("StGLImageProgram::fGetColorBlend, ") + theTitle),
  // color conversion
  f2RGBPtr(NULL),
  fRGB2RGB    (StString("StGLImageProgram::fRGB2RGB, ")  + theTitle),
  fRGBA2RGB   (StString("StGLImageProgram::fRGBA2RGB, ") + theTitle),
  fGray2RGB   (StString("StGLImageProgram::fGray2RGB, ") + theTitle),
  fYUV2RGB    (StString("StGLImageProgram::fYUV2RGB, ")  + theTitle),
  fYUV2RGBjpeg(StString("StGLImageProgram::fYUV2RGBjpeg, ") + theTitle),
  // color correction
  fCorrectPtr(NULL),
  fCorrectNO (StString("StGLImageProgram::fCorrectNO, ")  + theTitle),
  fCorrectON (StString("StGLImageProgram::fCorrectON, ")  + theTitle),
  myBrightness(),
  mySaturation(),
  myColorScale(1.0f, 1.0f, 1.0f),
  // gamma correction
  fGammaPtr(NULL),
  fGammaNO (StString("StGLImageProgram::fGammaNO, ")  + theTitle),
  fGammaON (StString("StGLImageProgram::fGammaON, ")  + theTitle) {
    //
    f2RGBPtr    = &fRGB2RGB;
    fCorrectPtr = &fCorrectNO;
    fGammaPtr   = &fGammaNO;

    params.gamma = new StFloat32Param(   1.0f,         // initial value
                                        0.05f, 100.0f, // min, max values
                                         1.0f,         // default value
                                        0.05f,         // incremental step
                                      0.0001f);        // equality tolerance
    params.gamma->signals.onChanged.connect(this, &StGLImageProgram::doGammaChanged);
    params.brightness = new StFloat32Param(   1.0f,         // initial value
                                              0.0f, 100.0f, // min, max values
                                              1.0f,         // default value
                                             0.05f,         // incremental step
                                           0.0001f);        // equality tolerance
    params.brightness->signals.onChanged.connect(this, &StGLImageProgram::setupCorrectionShader);
    params.saturation = new StFloat32Param(   1.0f,         // initial value
                                            -10.0f, 100.0f, // min, max values
                                              1.0f,         // default value
                                             0.05f,         // incremental step
                                           0.0001f);        // equality tolerance
    params.saturation->signals.onChanged.connect(this, &StGLImageProgram::setupCorrectionShader);
}

StGLImageProgram::~StGLImageProgram() {
    //
}

bool StGLImageProgram::isNoColorScale() const {
    return myColorScale.r() > 0.9f
        && myColorScale.g() > 0.9f
        && myColorScale.b() > 0.9f;
}

void StGLImageProgram::setColorScale(const StGLVec3& theScale) {
    myColorScale = theScale;
    setupCorrectionShader(0.0f);
}

void StGLImageProgram::resetColorScale() {
    if(isNoColorScale()) {
        return;
    }
    myColorScale = StGLVec3(1.0f, 1.0f, 1.0f);
    setupCorrectionShader(0.0f);
}

void StGLImageProgram::setCorrectionUniform() {
    if(fCorrectPtr == &fCorrectNO) {
        return;
    }
    StGLMatrix stColorMat = StGLMatrix::multiply(mySaturation, myBrightness);
    stColorMat.scale(myColorScale.r(), myColorScale.g(), myColorScale.b());
    glUniformMatrix4fv(uniColorProcessingLoc, 1, GL_FALSE, stColorMat);
}

void StGLImageProgram::setGammaUniform() {
    if(fGammaPtr == &fGammaNO) {
        return;
    }
    GLfloat aReversed = 1.0f / params.gamma->getValue();
    StGLVec4 aVec(aReversed, aReversed, aReversed, 1.0f);
    glUniform4fv(uniGammaLoc, 1, aVec);
}

void StGLImageProgram::setupCorrectionShader(const float ) {
    myBrightness.setBrightness(params.brightness->getValue());
    mySaturation.setSaturation(params.saturation->getValue());

    if(params.brightness->isDefaultValue()
    && params.saturation->isDefaultValue()
    && isNoColorScale()) {
        StGLProgram::swapShader(*fCorrectPtr, fCorrectNO);
        fCorrectPtr = &fCorrectNO;
        return;
    } else if(fCorrectPtr != &fCorrectON) {
        StGLProgram::swapShader(*fCorrectPtr, fCorrectON);
        fCorrectPtr = &fCorrectON;
    }
    if(isValid()) {
        StGLProgram::use();
            setCorrectionUniform();
        StGLProgram::unuse();
    }
}

void StGLImageProgram::doGammaChanged(const float theGamma) {
    if(stAreEqual(theGamma, 1.0f, 0.0001f)) {
        StGLProgram::swapShader(*fGammaPtr, fGammaNO);
        fGammaPtr = &fGammaNO;
        return;
    } else if(fGammaPtr != &fGammaON) {
        StGLProgram::swapShader(*fGammaPtr, fGammaON);
        fGammaPtr = &fGammaON;
    }
    if(isValid()) {
        StGLProgram::use();
            setGammaUniform();
        StGLProgram::unuse();
    }
}

void StGLImageProgram::setupSrcColorShader(StImage::ImgColorModel theColorModel) {
    if(!isValid()) {
        return;
    }
    switch(theColorModel) {
        case StImage::ImgColor_RGB: {
            StGLProgram::swapShader(*f2RGBPtr, fRGB2RGB);
            f2RGBPtr = &fRGB2RGB;
            break;
        }
        case StImage::ImgColor_RGBA: {
            StGLProgram::swapShader(*f2RGBPtr, fRGBA2RGB);
            f2RGBPtr = &fRGBA2RGB;
            break;
        }
        case StImage::ImgColor_GRAY: {
            StGLProgram::swapShader(*f2RGBPtr, fGray2RGB);
            f2RGBPtr = &fGray2RGB;
            break;
        }
        case StImage::ImgColor_YUV: {
            StGLProgram::swapShader(*f2RGBPtr, fYUV2RGB);
            f2RGBPtr = &fYUV2RGB;
            break;
        }
        case StImage::ImgColor_YUVjpeg: {
            StGLProgram::swapShader(*f2RGBPtr, fYUV2RGBjpeg);
            f2RGBPtr = &fYUV2RGBjpeg;
            break;
        }
        default:
            ST_DEBUG_LOG("No GLSL shader for this color model = " + theColorModel);
            ST_DEBUG_ASSERT(false);
    }
}

bool StGLImageProgram::init() {
    const StGLResources aShaders("StGLWidgets");

    const char F_SHADER_GET_COLOR[] =
       "uniform sampler2D uTexture;"
       "vec4 getColor(in vec2 texCoord) {"
       "    return texture2D(uTexture, texCoord);"
       "}";

    // color correction shaders
    const char F_SHADER_CORRECT_NO[] =
       "void applyCorrection(inout vec4 color) {}";

    const char F_SHADER_CORRECT_ON[] =
       "uniform mat4 uColorProcessing;"
       "void applyCorrection(inout vec4 color) {"
       "    color = uColorProcessing * color;"
       "}";

    // gamma correction shaders
    const char F_SHADER_GAMMA_NO[] =
       "void applyGamma(inout vec4 color) {}";

    const char F_SHADER_GAMMA_ON[] =
       "uniform vec4 uGamma;"
       "void applyGamma(inout vec4 color) {"
       "    color = pow(color, uGamma);"
       "}";

    // color conversion shaders
    const char F_SHADER_RGB2RGB[] =
       "void convertToRGB(inout vec4 color, in vec2 texCoord) {}";

    const char F_SHADER_RGBA2RGB[] =
       "void convertToRGB(inout vec4 color, in vec2 texCoord) {"
       "    vec4 backColor;"
       "    bool evenX = int(mod(floor(gl_FragCoord.x + 1.5), 16.0)) >= 8;" // just simple 8 pixels check-board
       "    bool evenY = int(mod(floor(gl_FragCoord.y + 1.5), 16.0)) >= 8;"
       "    if((evenX && evenY) || (!evenX && !evenY)) {"
       "        backColor = vec4(0.2, 0.2, 0.2, 1.0);"
       "    } else {"
       "        backColor = vec4(1.0, 1.0, 1.0, 1.0);"
       "    }"
       "    color = mix(backColor, color, color.a);"
       "}";

    const char F_SHADER_GRAY2RGB[] =
       "void convertToRGB(inout vec4 color, in vec2 texCoord) {"
       "color.r = color.a;" // gray scale stored in alpha
       "color.g = color.a;"
       "color.b = color.a;"
       //"color.a = 1.0;"
       "}";

    return fGetColor.init (F_SHADER_GET_COLOR)
        && fCorrectNO.init(F_SHADER_CORRECT_NO)
        && fCorrectON.init(F_SHADER_CORRECT_ON)
        && fGammaNO.init  (F_SHADER_GAMMA_NO)
        && fGammaON.init  (F_SHADER_GAMMA_ON)
        && fRGB2RGB.init  (F_SHADER_RGB2RGB)
        && fRGBA2RGB.init (F_SHADER_RGBA2RGB)
        && fGray2RGB.init (F_SHADER_GRAY2RGB)
        && fGetColorBlend.initFile(aShaders.getShaderFile("flatGetColorBlend.shf"))
        && fYUV2RGB.initFile      (aShaders.getShaderFile("convertYUV2RGB.shf"))
        && fYUV2RGBjpeg.initFile  (aShaders.getShaderFile("convertYUV2RGBjpeg.shf"));
}

bool StGLImageProgram::link() {
    if(!StGLProgram::link()) {
        return false;
    }

    uniProjMatLoc      = StGLProgram::getUniformLocation("uProjMat");
    uniModelMatLoc     = StGLProgram::getUniformLocation("uModelMat");
    uniTexMainDataLoc  = StGLProgram::getUniformLocation("uTexData");
    uniTexUVDataLoc    = StGLProgram::getUniformLocation("uTexUVData");
    uniTexSizePxLoc    = StGLProgram::getUniformLocation("uTexSizePx");
    uniTexelSizePxLoc  = StGLProgram::getUniformLocation("uTexelSize");
    uniColorProcessingLoc = StGLProgram::getUniformLocation("uColorProcessing");
    uniGammaLoc        = StGLProgram::getUniformLocation("uGamma");
    atrVVertexLoc      = StGLProgram::getAttribLocation("vVertex");
    atrVTCoordLoc    = StGLProgram::getAttribLocation("vTexCoord");

    StGLVarLocation uniTextureLoc  = StGLProgram::getUniformLocation("uTexture");
    StGLVarLocation uniTextureULoc = StGLProgram::getUniformLocation("uTextureU");
    StGLVarLocation uniTextureVLoc = StGLProgram::getUniformLocation("uTextureV");
    StGLProgram::use();
        glUniform1i(uniTextureLoc,  StGLProgram::TEXTURE_SAMPLE_0);
        glUniform1i(uniTextureULoc, StGLProgram::TEXTURE_SAMPLE_1);
        glUniform1i(uniTextureVLoc, StGLProgram::TEXTURE_SAMPLE_2);
        setCorrectionUniform();
        setGammaUniform();
    StGLProgram::unuse();

    return uniModelMatLoc.isValid()
        && uniTexMainDataLoc.isValid()
        //&& uniTexSizePxLoc.isValid()
        //&& uniTexelSizePxLoc.isValid()
        //&& uniSmoothFilterLoc.isValid()
        //&& uniColorProcessingLoc
        && atrVVertexLoc.isValid()
        && atrVTCoordLoc.isValid()
        && uniTextureLoc.isValid();
}
