/**
 * Copyright © 2009-2012 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLWidgets/StGLTextArea.h>
#include <StGLWidgets/StGLRootWidget.h>

#include <StGL/StGLProgram.h>
#include <StFile/StFileNode.h>
#include <StThreads/StProcess.h>

class StGLTextArea::StTextProgram : public StGLProgram {

        private:

    StGLVarLocation uniProjMatLoc;
    StGLVarLocation uniModelMatLoc;
    StGLVarLocation uniTextColorLoc;

    StGLVarLocation atrVVertexLoc;
    StGLVarLocation atrVTexCoordLoc;

        public:

    StTextProgram()
    : StGLProgram("StGLTextArea, Text Program"),
      uniProjMatLoc(),
      uniModelMatLoc(),
      uniTextColorLoc(),
      atrVVertexLoc(),
      atrVTexCoordLoc() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    StGLVarLocation getVTexCoordLoc() const {
        return atrVTexCoordLoc;
    }

    void setProjMat(const StGLMatrix& theProjMat) {
        glUniformMatrix4fv(uniProjMatLoc, 1, GL_FALSE, theProjMat);
    }

    void setModelMat(const StGLMatrix& theModelMat) {
        glUniformMatrix4fv(uniModelMatLoc, 1, GL_FALSE, theModelMat);
    }

    void setTextColor(const StGLVec4& theTextColorVec) {
        glUniform4fv(uniTextColorLoc, 1, theTextColorVec);
    }

    virtual bool init() {
        const char VERTEX_SHADER[] =
           "uniform mat4 uProjMat; \
            uniform mat4 uModelMat; \
            attribute vec4 vVertex; \
            attribute vec2 vTexCoord; \
            varying vec2 fTexCoord; \
            void main(void) { \
                fTexCoord = vTexCoord; \
                gl_Position = uProjMat * uModelMat * vVertex; \
            }";

        const char FRAGMENT_SHADER[] =
           "uniform sampler2D uTexture;"
           "uniform vec4 uTextColor;"
           "varying vec2 fTexCoord;"
           "void main(void) {"
           "     vec4 color = uTextColor;"
           "     color.a *= texture2D(uTexture, fTexCoord).a;"
           "     gl_FragColor = color;"
           "}";

        StGLVertexShader stVertexShader(StGLProgram::getTitle());
        stVertexShader.init(VERTEX_SHADER);

        StGLFragmentShader stFragmentShader(StGLProgram::getTitle());
        stFragmentShader.init(FRAGMENT_SHADER);
        if(!StGLProgram::create().attachShader(stVertexShader).attachShader(stFragmentShader).link()) {
            return false;
        }

        uniProjMatLoc   = StGLProgram::getUniformLocation("uProjMat");
        uniModelMatLoc  = StGLProgram::getUniformLocation("uModelMat");
        uniTextColorLoc = StGLProgram::getUniformLocation("uTextColor");
        atrVVertexLoc   = StGLProgram::getAttribLocation("vVertex");
        atrVTexCoordLoc = StGLProgram::getAttribLocation("vTexCoord");

        StGLVarLocation uniTextureLoc = StGLProgram::getUniformLocation("uTexture");
        if(uniTextureLoc.isValid()) {
            StGLProgram::use();
            glUniform1i(uniTextureLoc, StGLProgram::TEXTURE_SAMPLE_0);
            StGLProgram::unuse();
        }

        return uniProjMatLoc.isValid()
            && uniModelMatLoc.isValid()
            && uniTextColorLoc.isValid()
            && atrVVertexLoc.isValid()
            && atrVTexCoordLoc.isValid()
            && uniTextureLoc.isValid();
    }

};

class StGLTextArea::StBorderProgram : public StGLProgram {

        private:

    StGLVarLocation uniProjMatLoc;
    StGLVarLocation uniModelMatLoc;
    StGLVarLocation uniColorLoc;

    StGLVarLocation atrVVertexLoc;

        public:

    StBorderProgram()
    : StGLProgram("StGLTextArea, Border Program"),
      uniProjMatLoc(),
      uniModelMatLoc(),
      uniColorLoc(),
      atrVVertexLoc() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    void setProjMat(const StGLMatrix& theProjMat) {
        glUniformMatrix4fv(uniProjMatLoc, 1, GL_FALSE, theProjMat);
    }

    void setModelMat(const StGLMatrix& theModelMat) {
        glUniformMatrix4fv(uniModelMatLoc, 1, GL_FALSE, theModelMat);
    }

    void setColor(const StGLVec4& theColorVec) {
        glUniform4fv(uniColorLoc, 1, theColorVec);
    }

    virtual bool init() {
        const char VERTEX_SHADER[] =
           "uniform mat4 uProjMat; \
            uniform mat4 uModelMat; \
            attribute vec4 vVertex; \
            void main(void) { \
                gl_Position = uProjMat * uModelMat * vVertex; \
            }";

        const char FRAGMENT_SHADER[] =
           "uniform vec4 uColor; \
            void main(void) { \
                gl_FragColor = uColor; \
            }";

        StGLVertexShader stVertexShader(StGLProgram::getTitle());
        stVertexShader.init(VERTEX_SHADER);

        StGLFragmentShader stFragmentShader(StGLProgram::getTitle());
        stFragmentShader.init(FRAGMENT_SHADER);
        if(!StGLProgram::create().attachShader(stVertexShader).attachShader(stFragmentShader).link()) {
            return false;
        }

        uniProjMatLoc  = StGLProgram::getUniformLocation("uProjMat");
        uniModelMatLoc = StGLProgram::getUniformLocation("uModelMat");
        uniColorLoc    = StGLProgram::getUniformLocation("uColor");
        atrVVertexLoc  = StGLProgram::getAttribLocation("vVertex");
        return uniProjMatLoc.isValid() && uniModelMatLoc.isValid() && uniColorLoc.isValid() && atrVVertexLoc.isValid();
    }

};

namespace {
    static const StString CLASS_NAME("StGLTextArea");
    static const size_t SHARE_TEXT_PROGRAM_ID   = StGLRootWidget::generateShareId();
    static const size_t SHARE_BORDER_PROGRAM_ID = StGLRootWidget::generateShareId();
    static const size_t SHARE_FONT_NORMAL_ID    = StGLRootWidget::generateShareId();
    static const size_t SHARE_FONT_SMALL_ID     = StGLRootWidget::generateShareId();
    static const size_t SHARE_FONT_BIG_ID       = StGLRootWidget::generateShareId();

    enum {
        ST_FONT_SERIF,
        ST_FONT_SANS,
        ST_FONT_MONO,
    };

    static StString getFont(const int theFontType) {
    #if (defined(_WIN32) || defined(__WIN32__))
        switch(theFontType) {
            case ST_FONT_SERIF:
                return "times.ttf";
            case ST_FONT_SANS:
                return "micross.ttf";
            case ST_FONT_MONO:
            default:
                return "tahoma.ttf";
        }
    #elif (defined(__APPLE__))
        switch(theFontType) {
            case ST_FONT_SERIF:
                return "Times.dfont";
            case ST_FONT_SANS:
                return "Trebuchet MS.ttf";
                //return "LucidaGrande.ttc";
                //return "Geneva.dfont";
                //return "HelveticaNeue.dfont";
                //return "Helvetica.dfont";
            case ST_FONT_MONO:
            default:
                return "Monaco.dfont";
        }
    #elif (defined(__linux__) || defined(__linux))
        switch(theFontType) {
            case ST_FONT_SERIF:
                return "DejaVuSerif.ttf";
            case ST_FONT_SANS:
                return "DejaVuSans.ttf";
            case ST_FONT_MONO:
            default:
                return "DejaVuSansMono.ttf";
        }
    #else
        #error undefined fonts
    #endif
    }
};

StGLTextArea::StGLTextArea(StGLWidget* theParent,
                           const int theLeft, const int theTop,
                           const StGLCorner theCorner,
                           const int theWidth, const int theHeight, bool theToCutView,
                           const FontSize theSize)
: StGLWidget(theParent, theLeft, theTop, theCorner, theWidth, theHeight),
  myTextProgram(getRoot()->getShare(SHARE_TEXT_PROGRAM_ID)),
  myBorderProgram(getRoot()->getShare(SHARE_BORDER_PROGRAM_ID)),
  myFont(getRoot()->getShare((theSize == SIZE_NORMAL) ? SHARE_FONT_NORMAL_ID
                          : ((theSize == SIZE_SMALL)  ? SHARE_FONT_SMALL_ID : SHARE_FONT_BIG_ID))),
  mySize(theSize),
  myTextColor(0.0f, 0.0f, 0.0f, 1.0f),
  myShadowColor(0.0f, 0.0f, 0.0f, 1.0f),
  myBackColor(0.365f, 0.722f, 1.0f, 1.0f),
  myBorderColor(0.0f, 0.0f, 0.0f, 1.0f),
  myMarginLeft(0),
  myMarginRight(0),
  myMarginTop(0),
  myMarginBottom(0),
  myTextWidth(-1.0f),
  myToRecompute(true),
  myToCutView(theToCutView),
  myToShowBorder(false),
  myToDrawShadow(false),
  myIsInitialized(false) {
    // initialize FreeType font without GL resources
    if(myFont.isNull()) {
        const StString ST_FONT_SANS_PATH = StProcess::getFontsRoot() + getFont(ST_FONT_SANS);
        StStringUtf8 aFontPath = StFileNode::getCompatibleName(ST_FONT_SANS_PATH);

        StHandle<StFTFont> aFont = new StFTFont();
        if(!aFont->init(aFontPath, getFontSize())) {
            ST_ERROR_LOG("Could not load font '" + ST_FONT_SANS_PATH + '\'');
        }
        myFont.create(new StGLFont(aFont));
    }
}

StGLTextArea::~StGLTextArea() {
    //
}

const StString& StGLTextArea::getClassName() {
    return CLASS_NAME;
}

const StString& StGLTextArea::getText() const {
    return myText;
}

void StGLTextArea::setText(const StString& theText) {
    if(myText != theText) {
        myText = theText;
        myToRecompute = true;
    }
}

void StGLTextArea::setTextWidth(const int theWidth) {
    myTextWidth = (GLfloat )theWidth;
    myToRecompute = true;
}

bool StGLTextArea::stglInit() {
    if(myIsInitialized) {
        return true;
    }

    // initialize GL resources for the font
    if(!myFont->wasInitialized()) {
        if(!myFont->getFont()->isValid()) {
            return false; // critical error
        } else if(!myFont->stglInit()) {
            ST_ERROR_LOG("Could not initialize OpenGL resources for font");
            return false;
        }
    } else if(!myFont->isValid()) {
        return false;
    }

    myTextWidth = (GLfloat )getRectPx().width();

    // initialize text program
    if(myTextProgram.isNull()) {
        myTextProgram.create(new StTextProgram());
        if(!myTextProgram->init()) {
            return false;
        }
    } else if(!myTextProgram->isValid()) {
        return false;
    }

    // initialize border program
    if(myBorderProgram.isNull()) {
        myBorderProgram.create(new StBorderProgram());
        if(!myBorderProgram->init()) {
            return false;
        }
    } else if(!myBorderProgram->isValid()) {
        return false;
    }

    myIsInitialized = true;

    myBorderIVertBuf.init();
    myBorderOVertBuf.init();

    return StGLWidget::stglInit();
}

void StGLTextArea::recomputeBorder() {
    GLfloat aTextAreaW = GLfloat(getRectPx().width());
    const GLfloat quadVerticesInner[4 * 4] = {
        3.0f + aTextAreaW, myTextBndBox.top()    + 3.0f, 0.0f, 1.0f, // top-right
        3.0f + aTextAreaW, myTextBndBox.bottom() - 3.0f, 0.0f, 1.0f, // bottom-right
                    -3.0f, myTextBndBox.top()    + 3.0f, 0.0f, 1.0f, // top-left
                    -3.0f, myTextBndBox.bottom() - 3.0f, 0.0f, 1.0f  // bottom-left
    };
    const GLfloat quadVerticesOuter[4 * 4] = {
        4.0f + aTextAreaW, myTextBndBox.top()    + 4.0f, 0.0f, 1.0f, // top-right
        4.0f + aTextAreaW, myTextBndBox.bottom() - 4.0f, 0.0f, 1.0f, // bottom-right
                    -4.0f, myTextBndBox.top()    + 4.0f, 0.0f, 1.0f, // top-left
                    -4.0f, myTextBndBox.bottom() - 4.0f, 0.0f, 1.0f  // bottom-left
    };
    myBorderIVertBuf.init(4, 4, quadVerticesInner);
    myBorderOVertBuf.init(4, 4, quadVerticesOuter);
}

void StGLTextArea::drawText() {
    glActiveTexture(GL_TEXTURE0);
    for(size_t aTextureIter = 0; aTextureIter < myTexturesList.size(); ++aTextureIter) {
        if(!myTextVertBuf[aTextureIter]->isValid() || myTextVertBuf[aTextureIter]->getElemsCount() < 1) {
            continue;
        }

        glBindTexture(GL_TEXTURE_2D, myTexturesList[aTextureIter]);

        myTextVertBuf[aTextureIter]->bindVertexAttrib(myTextProgram->getVVertexLoc());
        myTextTCrdBuf[aTextureIter]->bindVertexAttrib(myTextProgram->getVTexCoordLoc());

            glDrawArrays(GL_TRIANGLES, 0, GLsizei(myTextVertBuf[aTextureIter]->getElemsCount()));

        myTextTCrdBuf[aTextureIter]->unBindVertexAttrib(myTextProgram->getVTexCoordLoc());
        myTextVertBuf[aTextureIter]->unBindVertexAttrib(myTextProgram->getVVertexLoc());
    }
    glBindTexture(GL_TEXTURE_2D, 0);
}

void StGLTextArea::stglDraw(unsigned int theView) {
    if(!myIsInitialized || !isVisible()) {
        return;
    }

    myTextColor.a() = myShadowColor.a() = myBackColor.a() = myBorderColor.a() = (GLfloat )opacityValue;

    if(myToRecompute) {
        myFormatter.reset();
        myFormatter.append(myText, *myFont);
        myFormatter.format(myTextWidth, GLfloat(getRectPx().height()));
        myFormatter.getResult(myTexturesList, myTextVertBuf, myTextTCrdBuf);
        myFormatter.getBndBox(myTextBndBox);
        if(myToShowBorder) {
            recomputeBorder();
        }
        myToRecompute = false;
    }

    StRectI_t aTextRectPx = getRectPx();
    aTextRectPx.left()   += myMarginLeft;
    aTextRectPx.right()  -= myMarginRight;
    aTextRectPx.top()    += myMarginTop;
    aTextRectPx.bottom() -= myMarginBottom;
    StRectD_t aTextRectGl = getRoot()->getRectGl(getAbsolute(aTextRectPx));

    // size corrector for FTGL
    StRectD_t zparams; getCamera()->getZParams(zparams);
    GLfloat aSizeOut = 2.0f * GLfloat(zparams.top()) / GLfloat(getRoot()->getRectPx().height());

    StGLMatrix aModelMat;
    aModelMat.translate(StGLVec3(0.0f, 0.0f, -getCamera()->getZScreen()));
    aModelMat.translate(StGLVec3(GLfloat(aTextRectGl.left()),
                                 GLfloat(aTextRectGl.top()),
                                 0.0f));
    aModelMat.scale(aSizeOut, aSizeOut, 0.0f);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    // draw borders
    if(myToShowBorder) {
        if(myBorderOVertBuf.getElemsCount() == 0) {
            recomputeBorder();
        }

        myBorderProgram->use();
            myBorderProgram->setProjMat(getCamera()->getProjMatrix());
            myBorderProgram->setModelMat(aModelMat);

            myBorderProgram->setColor(myBorderColor);
            myBorderOVertBuf.bindVertexAttrib(myBorderProgram->getVVertexLoc());
                glDrawArrays(GL_TRIANGLE_STRIP, 0, GLsizei(myBorderOVertBuf.getElemsCount()));
            myBorderOVertBuf.unBindVertexAttrib(myBorderProgram->getVVertexLoc());

            myBorderProgram->setColor(myBackColor);
            myBorderIVertBuf.bindVertexAttrib(myBorderProgram->getVVertexLoc());
                glDrawArrays(GL_TRIANGLE_STRIP, 0, GLsizei(myBorderIVertBuf.getElemsCount()));
            myBorderIVertBuf.unBindVertexAttrib(myBorderProgram->getVVertexLoc());
        myBorderProgram->unuse();
    }

    // draw text
    glActiveTexture(GL_TEXTURE0); // our shader is bound to first texture unit
    myTextProgram->use();
        myTextProgram->setProjMat(getCamera()->getProjMatrix());
        myTextProgram->setModelMat(aModelMat);
        myTextProgram->setTextColor(myToDrawShadow ? myShadowColor : myTextColor);

        drawText();

        if(myToDrawShadow) {
            aModelMat.initIdentity();
            aTextRectPx.left() -= 1;
            aTextRectPx.top()  -= 1;
            aTextRectGl = getRoot()->getRectGl(getAbsolute(aTextRectPx));
            aModelMat.translate(StGLVec3(0.0f, 0.0f, -getCamera()->getZScreen()));
            aModelMat.translate(StGLVec3(GLfloat(aTextRectGl.left()),
                                         GLfloat(aTextRectGl.top()),
                                         0.0f));
            aModelMat.scale(aSizeOut, aSizeOut, 0.0f);

            myTextProgram->setModelMat(aModelMat);
            myTextProgram->setTextColor(myTextColor);

            drawText();
        }

    myTextProgram->unuse();

    glDisable(GL_BLEND);

    StGLWidget::stglDraw(theView);
}
