/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLWidgets/StGLTextureButton.h>
#include <StGLWidgets/StGLRootWidget.h>

#include <StGL/StGLTexture.h>
#include <StGL/StGLProgram.h>
#include <StGL/StGLResources.h>
#include <StImage/StImageFile.h>
#include <StThreads/StProcess.h>

class StGLTextureButton::StButtonProgram : public StGLProgram {

        private:

    StGLVarLocation uniProjMatLoc;

    StGLVarLocation atrVVertexLoc;
    StGLVarLocation atrVTexCoordLoc;

    StGLVarLocation uniTimeLoc;
    StGLVarLocation uniClickedLoc;
    StGLVarLocation uniParamsLoc;

        public:

    StButtonProgram()
    : StGLProgram("StGLTextureButton"),
      uniProjMatLoc(),
      atrVVertexLoc(),
      atrVTexCoordLoc(),
      uniTimeLoc(),
      uniClickedLoc(),
      uniParamsLoc() {
        //
    }

    virtual ~StButtonProgram() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    StGLVarLocation getVTexCoordLoc() const {
        return atrVTexCoordLoc;
    }

    void setProjMat(const StGLMatrix& stProjMat) {
        glUniformMatrix4fv(uniProjMatLoc, 1, GL_FALSE, stProjMat);
    }

    virtual bool init() {
        const StGLResources aShaders("StGLWidgets");
        StGLVertexShader stVertexShader(StGLProgram::getTitle());
        stVertexShader.initFile(aShaders.getShaderFile("StGLTextureButton.shv"));

        StGLFragmentShader stFragmentShader(StGLProgram::getTitle());
        stFragmentShader.initFile(aShaders.getShaderFile("StGLTextureButton.shf"));
        if(!StGLProgram::create().attachShader(stVertexShader).attachShader(stFragmentShader).link()) {
            return false;
        }

        uniProjMatLoc   = StGLProgram::getUniformLocation("uProjMat");
        atrVVertexLoc   = StGLProgram::getAttribLocation ("vVertex");
        atrVTexCoordLoc = StGLProgram::getAttribLocation ("vTexCoord");
        uniTimeLoc      = StGLProgram::getUniformLocation("uTime");
        uniClickedLoc   = StGLProgram::getUniformLocation("uClicked");
        uniParamsLoc    = StGLProgram::getUniformLocation("uParams");

        StGLVarLocation uniTextureLoc = StGLProgram::getUniformLocation("uTexture");
        if(uniTextureLoc.isValid()) {
            StGLProgram::use();
            glUniform1i(uniTextureLoc, StGLProgram::TEXTURE_SAMPLE_0);
            StGLProgram::unuse();
        }

        return uniProjMatLoc.isValid()
            && atrVVertexLoc.isValid()
            && atrVTexCoordLoc.isValid()
            && uniTimeLoc.isValid()
            && uniClickedLoc.isValid()
            && uniParamsLoc.isValid()
            && uniTextureLoc.isValid();
    }

    void useTemp() {
        StGLProgram::use();
    }

    void use(double time, double lightX, double lightY, double opacity, bool isClicked) {
        StGLProgram::use();
        glUniform1f(uniTimeLoc, GLfloat(time));
        glUniform1i(uniClickedLoc, (isClicked ? 20 : 2));
        glUniform3f(uniParamsLoc, GLfloat(lightX), GLfloat(lightY), GLfloat(opacity));
    }

};

namespace {
    static const StString CLASS_NAME("StGLTextureButton");
    static const size_t SHARE_PROGRAM_ID = StGLRootWidget::generateShareId();
};

StGLTextureButton::StGLTextureButton(StGLWidget* parent, const int& left, const int& top, const StGLCorner corner, const size_t& facesCount)
: StGLWidget(parent, left, top, corner),
  stVerticesBuf(),
  stTexCoordBuf(),
  faceId(0),
  facesCount(facesCount),
  stTextures(new StGLTexture[facesCount]),
  texturesPaths(new StString[facesCount]),
  stProgram(getRoot()->getShare(SHARE_PROGRAM_ID)),
  stWaveTimer(false) {
    //
    StGLWidget::signals.onMouseUnclick.connect(this, &StGLTextureButton::doMouseUnclick);
}

StGLTextureButton::~StGLTextureButton() {
    //
}

const StString& StGLTextureButton::getClassName() {
    return CLASS_NAME;
}

void StGLTextureButton::glWaveTimerControl() {
    if(isPointIn(getRoot()->getCursorZo())) {
        if(!stWaveTimer.isOn()) {
            stWaveTimer.restart();
        }
    } else {
        stWaveTimer.stop();
    }
}

void StGLTextureButton::setTexturePath(const StString* texturesPaths, const size_t& count) {
    size_t minCount = (count > facesCount) ? facesCount : count;
#ifdef __ST_DEBUG__
    if(count != facesCount) {
        ST_DEBUG_LOG_AT("WARNING, Not inaf textures paths for StGLTextureButton!");
    }
#endif
    for(size_t t = 0; t < minCount; ++t) {
        this->texturesPaths[t] = texturesPaths[t];
    }
}

void StGLTextureButton::stglResize(const StRectI_t& winRectPx) {
    StGLWidget::stglResize(winRectPx);

    // update vertices
    StArray<StGLVec2> aVertices(4);
    getRectGl(aVertices);
    stVerticesBuf.init(aVertices);

    // update projection matrix
    stProgram->useTemp();
        stProgram->setProjMat(getRoot()->getScreenProjection());
    stProgram->unuse();
}

bool StGLTextureButton::stglInit() {
    bool result = true;
    StHandle<StImageFile> stImage = StImageFile::create();
    if(!stImage.isNull()) {
        for(size_t t = 0; t < facesCount; ++t) {
            if(texturesPaths[t].isEmpty()) {
                ST_DEBUG_LOG("StGLTextureButton, texture for face " + t + " not set");
                continue;
            }
            if(!stImage->load(texturesPaths[t], StImageFile::ST_TYPE_PNG)) {
                ST_DEBUG_LOG(stImage->getState());
                continue;
            }
            changeRectPx().right()  = getRectPx().left() + (int )stImage->getSizeX();
            changeRectPx().bottom() = getRectPx().top()  + (int )stImage->getSizeY();
            stTextures[t].init(stImage->getPlane());
        }
        stImage.nullify();
    }

    if(stProgram.isNull()) {
        stProgram.create(new StButtonProgram());
        stProgram->init();
    }

    if(!stProgram->isValid()) {
        return false;
    }

    StArray<StGLVec2> aDummyVert(4);
    StArray<StGLVec2> aTexCoords(4);
    aTexCoords[0] = StGLVec2(1.0f, 0.0f);
    aTexCoords[1] = StGLVec2(1.0f, 1.0f);
    aTexCoords[2] = StGLVec2(0.0f, 0.0f);
    aTexCoords[3] = StGLVec2(0.0f, 1.0f);

    stVerticesBuf.init(aDummyVert);
    stTexCoordBuf.init(aTexCoords);
    return result;
}

void StGLTextureButton::stglDraw(unsigned int ) {
    if(!isVisible()) {
        return;
    }
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    stTextures[faceId].bind();

    StRectD_t butRectGl = getRectGl();
    GLdouble butWGl = butRectGl.right() - butRectGl.left();
    GLdouble butHGl = butRectGl.top() - butRectGl.bottom();

    StPointD_t mouseGl = getPointGl(getRoot()->getCursorZo());
    glWaveTimerControl();

    stProgram->use(stWaveTimer.getElapsedTimeInSec(),
                  (mouseGl.x() - butRectGl.left()) / butWGl,
                  (butRectGl.top()  - mouseGl.y()) / butHGl,
                   opacityValue,
                   isClicked(ST_MOUSE_LEFT));

    stVerticesBuf.bindVertexAttrib(stProgram->getVVertexLoc());
    stTexCoordBuf.bindVertexAttrib(stProgram->getVTexCoordLoc());

    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    stTexCoordBuf.unBindVertexAttrib(stProgram->getVTexCoordLoc());
    stVerticesBuf.unBindVertexAttrib(stProgram->getVVertexLoc());

    stProgram->unuse();
    stTextures[faceId].unbind();
    glDisable(GL_BLEND);
}

bool StGLTextureButton::tryClick(const StPointD_t& cursorZo, const int& mouseBtn, bool& isItemClicked) {
    if(StGLWidget::tryClick(cursorZo, mouseBtn, isItemClicked)) {
        isItemClicked = true;
        return true;
    }
    return false;
}

bool StGLTextureButton::tryUnClick(const StPointD_t& cursorZo, const int& mouseBtn, bool& isItemUnclicked) {
    if(StGLWidget::tryUnClick(cursorZo, mouseBtn, isItemUnclicked)) {
        isItemUnclicked = true;
        return true;
    }
    return false;
}

void StGLTextureButton::doMouseUnclick(const int btnId) {
    if(btnId == ST_MOUSE_LEFT) {
        signals.onBtnClick(getUserData());
    }
}
