/**
 * Copyright © 2009-2012 Kirill Gavrilov <kirill@sview.ru>
 *
 * StMoviePlayer program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StMoviePlayer program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StSeekBar.h"

#include <StGLWidgets/StGLRootWidget.h>
#include <StGL/StGLProgram.h>

namespace {
    static const StString className("StSeekBar");
};

class StSeekBar::StProgramSB : public StGLProgram {

        private:

    StGLVarLocation uniProjMatLoc;
    StGLVarLocation uniOpacityLoc;

    StGLVarLocation atrVVertexLoc;
    StGLVarLocation atrVColorLoc;

        public:

    StProgramSB()
    : StGLProgram("StSeekBar"),
      uniProjMatLoc(),
      uniOpacityLoc(),
      atrVVertexLoc(),
      atrVColorLoc() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    StGLVarLocation getVColorLoc() const {
        return atrVColorLoc;
    }

    void setProjMat(const StGLMatrix& theProjMat) {
        glUniformMatrix4fv(uniProjMatLoc, 1, GL_FALSE, theProjMat);
    }

    void setOpacity(const GLfloat theOpacityValue) {
        glUniform1f(uniOpacityLoc, theOpacityValue);
    }

    bool init() {
        const char VERTEX_SHADER[] =
           "uniform mat4 uProjMatrix; \
            uniform float uOpacity; \
            attribute vec4 vVertex; \
            attribute vec4 vColor; \
            varying vec4 fColor; \
            void main(void) { \
                fColor = vec4(vColor.rgb, vColor.a * uOpacity); \
                gl_Position = uProjMatrix * vVertex; \
            }";

        const char FRAGMENT_SHADER[] =
           "varying vec4 fColor; \
            void main(void) { \
                gl_FragColor = fColor; \
            }";

        StGLVertexShader stVertexShader(StGLProgram::getTitle());
        stVertexShader.init(VERTEX_SHADER);

        StGLFragmentShader aFragmentShader(StGLProgram::getTitle());
        aFragmentShader.init(FRAGMENT_SHADER);
        if(!StGLProgram::create().attachShader(stVertexShader).attachShader(aFragmentShader).link()) {
            return false;
        }

        uniProjMatLoc = StGLProgram::getUniformLocation("uProjMatrix");
        uniOpacityLoc = StGLProgram::getUniformLocation("uOpacity");
        atrVVertexLoc = StGLProgram::getAttribLocation("vVertex");
        atrVColorLoc  = StGLProgram::getAttribLocation("vColor");
        return uniProjMatLoc.isValid() && uniOpacityLoc.isValid() && atrVVertexLoc.isValid() && atrVColorLoc.isValid();
    }

};

StSeekBar::StSeekBar(StGLWidget* theParent,
                     const int   theTop)
: StGLWidget(theParent, 0, theTop, StGLCorner(ST_VCORNER_BOTTOM, ST_HCORNER_CENTER), 512, 12),
  myProgram(new StProgramSB()),
  myProgress(0.0f),
  myProgressPx(0) {
    //
    StGLWidget::signals.onMouseUnclick.connect(this, &StSeekBar::doMouseUnclick);
}

StSeekBar::~StSeekBar() {
    myProgram.nullify();
}

const StString& StSeekBar::getClassName() {
    return className;
}

void StSeekBar::stglResize(const StRectI_t& winRectPx) {
    StGLWidget::stglResize(winRectPx);

    // update pixel coordinates
    changeRectPx().right() = getParent()->getRectPx().width() - 128;

    stglUpdateVertices();

    // update projection matrix
    myProgram->use();
        myProgram->setProjMat(getRoot()->getScreenProjection());
    myProgram->unuse();
}

void StSeekBar::stglUpdateVertices() {
    StArray<StGLVec2> aVertices(12);

    // black border quad
    StRectI_t aRectPx(getRectPxAbsolute());
    getRoot()->getRectGl(aRectPx, aVertices, 0);

    // inner empty quad
    aRectPx.top()    += 1;
    aRectPx.bottom() -= 1;
    aRectPx.left()   += 1;
    aRectPx.right()  -= 1;
    getRoot()->getRectGl(aRectPx, aVertices, 4);

    // inner filled quad
    myProgressPx = int(myProgress * GLfloat(aRectPx.width()));
    aRectPx.right() = aRectPx.left() + myProgressPx;
    getRoot()->getRectGl(aRectPx, aVertices, 8);

    myVertices.init(aVertices);
}

bool StSeekBar::stglInit() {
    const GLfloat COLORS[4 * 12] = {
        // black border colors
        0.0f, 0.0f, 0.0f, 1.0f, // quad's top-right
        0.0f, 0.0f, 0.0f, 1.0f, // quad's bottom-right
        0.0f, 0.0f, 0.0f, 1.0f, // quad's top-left
        0.0f, 0.0f, 0.0f, 1.0f, // quad's bottom-left
        // empty color
        0.3f, 0.3f, 0.3f, 1.0f, // quad's top-right
        0.3f, 0.3f, 0.3f, 1.0f, // quad's bottom-right
        0.3f, 0.3f, 0.3f, 1.0f, // quad's top-left
        0.3f, 0.3f, 0.3f, 1.0f, // quad's bottom-left
        // fill color
        0.13f, 0.35f, 0.49f, 1.0f, // quad's top-right
        0.13f, 0.35f, 0.49f, 1.0f, // quad's bottom-right
        0.13f, 0.35f, 0.49f, 1.0f, // quad's top-left
        0.13f, 0.35f, 0.49f, 1.0f  // quad's bottom-left
    };

    myVertices.init(); // just generate buffer
    myColors.init(4, 12, COLORS);

    stglUpdateVertices();

    return myProgram->init();
}

void StSeekBar::stglDraw(unsigned int ) {

    // need to update vertices buffer?
    if(myProgressPx != int(myProgress * GLfloat((getRectPx().width() - 2)))) {
        stglUpdateVertices();
    }

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    myProgram->use();

    myProgram->setOpacity(GLfloat(opacityValue));

    myVertices.bindVertexAttrib(myProgram->getVVertexLoc());
    myColors.bindVertexAttrib(myProgram->getVColorLoc());

    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    glDrawArrays(GL_TRIANGLE_STRIP, 4, 4);

    if(myProgressPx >= 1) {
        glDrawArrays(GL_TRIANGLE_STRIP, 8, 4);
    }

    myColors.unBindVertexAttrib(myProgram->getVColorLoc());
    myVertices.unBindVertexAttrib(myProgram->getVVertexLoc());

    myProgram->unuse();
    glDisable(GL_BLEND);
}

void StSeekBar::doMouseUnclick(const int mouseBtn) {
    signals.onSeekClick(mouseBtn, getPointIn(getRoot()->getCursorZo()).x());
}
