/**
 * Copyright © 2009-2012 Kirill Gavrilov <kirill@sview.ru>
 *
 * StOutInterlace library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StOutInterlace library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StOutInterlace.h"

#include <StGLEW.h>
#include <StGL/StGL.h>
#include <StGL/StGLEnums.h>
#include <StGL/StGLVec.h>
#include <StCore/StWindow.h>
#include <StSettings/StSettings.h>
#include <StSettings/StTranslations.h>

namespace {
    static bool HAS_LOGGER_ID = StLogger::IdentifyModule("StOutInterlace");

    static const char ST_OUT_PLUGIN_NAME[] = "StOutInterlace";

    // shaders data
    static const char VSHADER_ED[]             = "vED.shv";
    static const char FSHADER_EDINTERLACE_ON[] = "fEDinterlace.shf";
    static const char FSHADER_ED_OFF[]         = "fEDoff.shf";

    static const char ST_SETTING_DEVICE_ID[]      = "deviceId";
    static const char ST_SETTING_WINDOWPOS[]      = "windowPos";
    static const char ST_SETTING_BIND_MONITOR[]   = "bindMonitor";
    static const char ST_SETTING_VSYNC[]          = "vsync";
    static const char ST_SETTING_REVERSE[]        = "reverse";
    static const char ST_SETTING_FRBUFF_X[]       = "frbufferX";
    static const char ST_SETTING_FRBUFF_Y[]       = "frbufferY";
    static const char ST_SETTING_USER_FRMB_SIZE[] = "useUserFrmbSize";

    // Zalman Trimon (horizontal interlace)
    static const StString PNP_ID_ZALMAN_TRIMON_M190S = "ZMT1900";
    static const StString PNP_ID_ZALMAN_TRIMON_M220W = "ZMT2200";

    // Envision (row interlace)
    static const StString PNP_ID_ENVISION_ENV2373 = "ENV2373";

    //Hyundai 3D LCD display W220S 1680x1050 (horizontal interlace)
    static const StString PNP_ID_HYUNDAI_W220S_DSUB = "HIT8002";
    static const StString PNP_ID_HYUNDAI_W220S_DVID = "HIT8D02";

    //Hyundai 3D LCD display W240S 1920x1200 (horizontal interlace)
    static const StString PNP_ID_HYUNDAI_W240S_DSUB = "HIT7003";
    static const StString PNP_ID_HYUNDAI_W240S_DVID = "HIT7D03";
};

// translation resources
enum {
    STTR_HINTERLACE_NAME    = 1000,
    STTR_HINTERLACE_DESC    = 1001,
    STTR_VINTERLACE_NAME    = 1002,
    STTR_VINTERLACE_DESC    = 1003,
    STTR_CHESSBOARD_NAME    = 1006,
    STTR_CHESSBOARD_DESC    = 1007,
    STTR_HINTERLACE_ED_NAME = 1008,
    STTR_HINTERLACE_ED_DESC = 1009,

    // parameters
    STTR_PARAMETER_VSYNC    = 1100,
    STTR_PARAMETER_SHOW_FPS = 1101,
    STTR_PARAMETER_REVERSE  = 1102,
    STTR_PARAMETER_BIND_MON = 1103,

    // about info
    STTR_PLUGIN_TITLE       = 2000,
    STTR_VERSION_STRING     = 2001,
    STTR_PLUGIN_DESCRIPTION = 2002,
};

static const char* ST_SHADER_TEMPLATE[3] = {
    "uniform sampler2D uTexture;\n"
    "varying vec2 fTexCoord;\n"
    "void main(void) {\n",
    "\n",
    "    gl_FragColor = texture2D(uTexture, fTexCoord);\n"
    "}\n"
};

static const StGLVarLocation ST_VATTRIB_VERTEX(0);
static const StGLVarLocation ST_VATTRIB_TCOORD(1);

StProgramFB::StProgramFB(const StString& theTitle)
: StGLProgram(theTitle) {
    //
}

bool StProgramFB::link() {
    StGLProgram::bindAttribLocation("vVertex",   ST_VATTRIB_VERTEX);
    StGLProgram::bindAttribLocation("vTexCoord", ST_VATTRIB_TCOORD);

    if(!StGLProgram::link()) {
        return false;
    }
    StGLVarLocation aTextureLoc = StGLProgram::getUniformLocation("uTexture");
    if(aTextureLoc.isValid()) {
        use();
        glUniform1i(aTextureLoc, StGLProgram::TEXTURE_SAMPLE_0); // GL_TEXTURE0
        unuse();
    }
    return aTextureLoc.isValid();
}

void StOutInterlace::setDevice(const int theDeviceId) {
    if(theDeviceId < 0 || theDeviceId >= DEVICE_NB) {
        ST_DEBUG_LOG("Incorrect device ID!");
        return;
    }
    myDeviceId = theDeviceId;
    if(myOptionsStruct != NULL) {
        myOptionsStruct->curDeviceId = myDeviceId;
    }
}

void StOutInterlace::optionsStructAlloc() {
    StTranslations stLangMap(ST_OUT_PLUGIN_NAME);

    // create device options structure
    myOptionsStruct = (StSDOptionsList_t* )StWindow::memAlloc(sizeof(StSDOptionsList_t)); stMemSet(myOptionsStruct, 0, sizeof(StSDOptionsList_t));
    myOptionsStruct->curRendererPath = StWindow::memAllocNCopy(myPluginPath);
    myOptionsStruct->curDeviceId = myDeviceId;

    myOptionsStruct->optionsCount = 4;
    myOptionsStruct->options = (StSDOption_t** )StWindow::memAlloc(sizeof(StSDOption_t*) * myOptionsStruct->optionsCount);

    // VSync option
    myOptionsStruct->options[DEVICE_OPTION_VSYNC] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDOnOff_t));
    myOptionsStruct->options[DEVICE_OPTION_VSYNC]->optionType = ST_DEVICE_OPTION_ON_OFF;
    ((StSDOnOff_t* )myOptionsStruct->options[DEVICE_OPTION_VSYNC])->value = myIsVSync;
    myOptionsStruct->options[DEVICE_OPTION_VSYNC]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_PARAMETER_VSYNC, "VSync"));

    // Show FPS option
    myOptionsStruct->options[DEVICE_OPTION_SHOWFPS] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDOnOff_t));
    myOptionsStruct->options[DEVICE_OPTION_SHOWFPS]->optionType = ST_DEVICE_OPTION_ON_OFF;
    ((StSDOnOff_t* )myOptionsStruct->options[DEVICE_OPTION_SHOWFPS])->value = myToShowFPS;
    myOptionsStruct->options[DEVICE_OPTION_SHOWFPS]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_PARAMETER_SHOW_FPS, "Show FPS"));

    // Reverse Order option
    myOptionsStruct->options[DEVICE_OPTION_REVERSE] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDOnOff_t));
    myOptionsStruct->options[DEVICE_OPTION_REVERSE]->optionType = ST_DEVICE_OPTION_ON_OFF;
    ((StSDOnOff_t* )myOptionsStruct->options[DEVICE_OPTION_REVERSE])->value = myIsReversed;
    myOptionsStruct->options[DEVICE_OPTION_REVERSE]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_PARAMETER_REVERSE, "Reverse Order"));

    // Bind Monitor option
    myOptionsStruct->options[DEVICE_OPTION_BINDMON] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDOnOff_t));
    myOptionsStruct->options[DEVICE_OPTION_BINDMON]->optionType = ST_DEVICE_OPTION_ON_OFF;
    ((StSDOnOff_t* )myOptionsStruct->options[DEVICE_OPTION_BINDMON])->value = myToBindToMonitor;
    myOptionsStruct->options[DEVICE_OPTION_BINDMON]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_PARAMETER_BIND_MON, "Bind To Supported Monitor"));
}

inline bool isInterlacedMonitor(const StMonitor& theMon) {
    return theMon.getPnPId() == PNP_ID_ZALMAN_TRIMON_M190S
        || theMon.getPnPId() == PNP_ID_ZALMAN_TRIMON_M220W
        || theMon.getPnPId() == PNP_ID_ENVISION_ENV2373
        || theMon.getPnPId() == PNP_ID_HYUNDAI_W220S_DSUB
        || theMon.getPnPId() == PNP_ID_HYUNDAI_W220S_DVID
        || theMon.getPnPId() == PNP_ID_HYUNDAI_W240S_DSUB
        || theMon.getPnPId() == PNP_ID_HYUNDAI_W240S_DVID;
}

StHandle<StMonitor> StOutInterlace::getHInterlaceMonitor() {
    StArrayList<StMonitor> aMonitors = StCore::getStMonitors();
    for(size_t aMonIter = 0; aMonIter < aMonitors.size(); ++aMonIter) {
        const StMonitor& aMon = aMonitors[aMonIter];
        if(isInterlacedMonitor(aMon)) {
            return new StMonitor(aMon);
        }
    }
    return NULL;
}

StOutInterlace::StOutInterlace()
: myDeviceId(DEVICE_AUTO),
  myEDTimer(true),
  myEDIntelaceOn(new StGLProgram("ED Interlace On")),
  myEDOff(new StGLProgram("ED Interlace Off")),
  myVpSizeY(10),
  myVpSizeYOnLoc(),
  myVpSizeYOffLoc(),
  myOptionsStruct(NULL),
  myFPSControl(),
  myToSavePlacement(true),
  myToBindToMonitor(true),
  myIsUserFrmSize(false),
  myIsVSync(true),
  myToShowFPS(false),
  myIsReversed(false),
  myIsStereo(false),
  myIsEDactive(false),
  myIsEDCodeFinished(false) {
    //
    myGlPrograms[DEVICE_HINTERLACE] = new StProgramFB("Row Interlace");
    myGlPrograms[DEVICE_VINTERLACE] = new StProgramFB("Column Interlace");
    myGlPrograms[DEVICE_CHESSBOARD] = new StProgramFB("Chessboard");
    myGlPrograms[DEVICE_HINTERLACE_ED] = myGlPrograms[DEVICE_HINTERLACE];

    myGlProgramsRev[DEVICE_HINTERLACE] = new StProgramFB("Row Interlace Inversed");
    myGlProgramsRev[DEVICE_VINTERLACE] = new StProgramFB("Column Interlace Inversed");
    myGlProgramsRev[DEVICE_CHESSBOARD] = new StProgramFB("Chessboard Inversed");
    myGlProgramsRev[DEVICE_HINTERLACE_ED] = myGlProgramsRev[DEVICE_HINTERLACE];
}

StOutInterlace::~StOutInterlace() {
    for(size_t anIter = 0; anIter < DEVICE_NB; ++anIter) {
        myGlPrograms[anIter].nullify();
        myGlProgramsRev[anIter].nullify();
    }
    myEDIntelaceOn.nullify();
    myEDOff.nullify();
    if(!myStCore.isNull() && !mySettings.isNull()) {
        if(!myFrmBuffer.isNull()) {
            mySettings->saveInt32(ST_SETTING_FRBUFF_X, myFrmBuffer->getSizeX());
            mySettings->saveInt32(ST_SETTING_FRBUFF_Y, myFrmBuffer->getSizeY());
            myFrmBuffer.nullify();
        }
        stMemFree(myOptionsStruct, StWindow::memFree);

        // read windowed placement
        getStWindow()->hide(ST_WIN_MASTER);
        getStWindow()->hide(ST_WIN_SLAVE);
        if(myToSavePlacement) {
            getStWindow()->setFullScreen(false);
            mySettings->saveInt32Rect(ST_SETTING_WINDOWPOS, getStWindow()->getPlacement());
        }
        mySettings->saveBool (ST_SETTING_BIND_MONITOR,   myToBindToMonitor);
        mySettings->saveBool (ST_SETTING_VSYNC,          myIsVSync);
        mySettings->saveBool (ST_SETTING_REVERSE,        myIsReversed);
        mySettings->saveBool (ST_SETTING_USER_FRMB_SIZE, myIsUserFrmSize);
        mySettings->saveInt32(ST_SETTING_DEVICE_ID,      myDeviceId);
    }
    mySettings.nullify();
    myStCore.nullify();
    StSettings::FREE();
    StCore::FREE();
}

bool StOutInterlace::init(const StString&      inRendererPath,
                          const int&           theDeviceId,
                          const StNativeWin_t* nativeParent) {
    myToSavePlacement = nativeParent == NULL;
    myDeviceId   = theDeviceId;
    myPluginPath = inRendererPath;
    if(!StVersionInfo::checkTimeBomb("sView - Interlace Output plugin")) {
        return false;
    }
    ST_DEBUG_LOG_AT("INIT Interlace output plugin");
    // Firstly INIT core library!
    if(StCore::INIT() != STERROR_LIBNOERROR) {
        stError(StString(ST_OUT_PLUGIN_NAME) + " Plugin, Core library not available!");
        return false;
    }

    // INIT settings library
    if(StSettings::INIT() != STERROR_LIBNOERROR) {
        stError(StString(ST_OUT_PLUGIN_NAME) + " Plugin, Settings plugin not available!");
        return false;
    }
    mySettings = new StSettings(ST_OUT_PLUGIN_NAME);
    myStCore   = new StCore();

    // load window position
    StHandle<StMonitor> anInterlacedMon = StOutInterlace::getHInterlaceMonitor();
    StRect<int32_t> loadedRect(256, 768, 256, 1024);
    bool isLoadedPosition = mySettings->loadInt32Rect(ST_SETTING_WINDOWPOS, loadedRect);
    mySettings->loadBool(ST_SETTING_BIND_MONITOR, myToBindToMonitor);
    StMonitor aMonitor = StCore::getMonitorFromPoint(loadedRect.center());
    if(myToBindToMonitor && !anInterlacedMon.isNull() && !isInterlacedMonitor(aMonitor)) {
        aMonitor = *anInterlacedMon;
    }
    if(isLoadedPosition) {
        if(!aMonitor.getVRect().isPointIn(loadedRect.center())) {
            ST_DEBUG_LOG("Warning, stored window position is out of the monitor(" + aMonitor.getId() + ")!" + loadedRect.toString());
            int w = loadedRect.width();
            int h = loadedRect.height();
            loadedRect.left()   = aMonitor.getVRect().left() + 256;
            loadedRect.right()  = loadedRect.left() + w;
            loadedRect.top()    = aMonitor.getVRect().top() + 256;
            loadedRect.bottom() = loadedRect.top() + h;
        }
    } else {
        // try to open window on correct display
        loadedRect = aMonitor.getVRect();
        loadedRect.left()   = loadedRect.left() + 256;
        loadedRect.right()  = loadedRect.left() + 1024;
        loadedRect.top()    = loadedRect.top()  + 256;
        loadedRect.bottom() = loadedRect.top()  + 512;
    }
    getStWindow()->setPlacement(loadedRect);

    mySettings->loadBool(ST_SETTING_VSYNC,   myIsVSync);
    mySettings->loadBool(ST_SETTING_REVERSE, myIsReversed);

    // load framebuffer settings
    GLsizei frbufferSizeX = aMonitor.getVRect().width();
    GLsizei frbufferSizeY = aMonitor.getVRect().height();
    mySettings->loadBool(ST_SETTING_USER_FRMB_SIZE, myIsUserFrmSize);
    if(myIsUserFrmSize) {
        mySettings->loadInt32(ST_SETTING_FRBUFF_X, frbufferSizeX);
        mySettings->loadInt32(ST_SETTING_FRBUFF_Y, frbufferSizeY);
    }

    // load device settings
    if(myDeviceId == StRendererInfo::DEVICE_AUTO) {
        mySettings->loadInt32(ST_SETTING_DEVICE_ID, myDeviceId);
        if(myDeviceId == StRendererInfo::DEVICE_AUTO) {
            myDeviceId = DEVICE_HINTERLACE;
        }
    }

    // allocate and setup the structure pointer
    optionsStructAlloc();
    getStWindow()->setValue(ST_WIN_DATAKEYS_RENDERER, (size_t )myOptionsStruct);

    setDevice(myDeviceId);

    // create our window!
    StWinAttributes_t attribs = stDefaultWinAttributes();
    attribs.isSlave = true;
    attribs.isSlaveHLineTop = true;
    attribs.isSlaveHide = true;
    getStWindow()->stglCreate(&attribs, nativeParent);

    // INIT GLEW library (help library for OpenGL extensions)
    if(!StGLEW::init()) {
        stError(StString(ST_OUT_PLUGIN_NAME) + " Plugin, OpenGL context is broken!\n(OpenGL library internal error?)");
        return false;
    } else if(!GLEW_VERSION_2_0) {
        stError(StString(ST_OUT_PLUGIN_NAME) + " Plugin, OpenGL2.0+ not available!");
        return false;
    }
    if(!stglSetVSync(myIsVSync)) {
        // enable/disable VSync by config
        ST_DEBUG_LOG(ST_OUT_PLUGIN_NAME + " Plugin, VSync extension not available!");
    }

    // INIT framebuffers
    if(!StGLFrameBuffer::isRectangularFBSupported()) {
        if(!myIsUserFrmSize) {
            StGLFrameBuffer::convertToPowerOfTwo(frbufferSizeX, frbufferSizeY);
        }
        ST_DEBUG_LOG(ST_OUT_PLUGIN_NAME + " Plugin, old videocard detected (GLSL 1.1)!");
    }
    myFrmBuffer = new StGLFrameBuffer();
    if(!myFrmBuffer->init(frbufferSizeX, frbufferSizeY)) {
        stError(StString(ST_OUT_PLUGIN_NAME) + " Plugin, Failed to init Frame Buffer");
        return false;
    }

    // INIT shaders
    StString aShadersError(StString(ST_OUT_PLUGIN_NAME) + " Plugin, Failed to init Shaders");
    StGLVertexShader aVertexShader("Interlace"); // common vertex shader
    if(!aVertexShader.init("attribute vec4 vVertex;\n"
                           "attribute vec2 vTexCoord;\n"
                           "varying vec2 fTexCoord;\n"
                           "void main(void) {\n"
                           "  fTexCoord = vTexCoord;\n"
                           "  gl_Position = vVertex;\n"
                           "}\n")) {
        stError(aShadersError);
        return false;
    }

    // row interlaced
    StGLFragmentShader aShaderRow(myGlPrograms[DEVICE_HINTERLACE]->getTitle());
    StGLFragmentShader aShaderRowRev(myGlProgramsRev[DEVICE_HINTERLACE]->getTitle());
    if(!aShaderRow.init(ST_SHADER_TEMPLATE[0],
                        // drop odd horizontal line (starts from bottom)
                        "if(int(mod(gl_FragCoord.y + 1.5, 2.0)) == 1) { discard; }\n",
                        ST_SHADER_TEMPLATE[2])) {
        stError(aShadersError);
        return false;
    } else if(!aShaderRowRev.init(ST_SHADER_TEMPLATE[0],
              // drop even horizontal line (starts from bottom)
              "if(int(mod(gl_FragCoord.y + 1.5, 2.0)) != 1) { discard; }\n",
              ST_SHADER_TEMPLATE[2])) {
        stError(aShadersError);
        return false;
    }
    myGlPrograms   [DEVICE_HINTERLACE]->create().attachShader(aVertexShader).attachShader(aShaderRow).link();
    myGlProgramsRev[DEVICE_HINTERLACE]->create().attachShader(aVertexShader).attachShader(aShaderRowRev).link();

    // column interlaced
    StGLFragmentShader aShaderCol(myGlPrograms[DEVICE_VINTERLACE]->getTitle());
    StGLFragmentShader aShaderColRev(myGlProgramsRev[DEVICE_VINTERLACE]->getTitle());
    if(!aShaderCol.init(ST_SHADER_TEMPLATE[0],
                        // drop odd column (starts from left)
                        "if(int(mod(gl_FragCoord.x + 1.5, 2.0)) != 1) { discard; }\n",
                        ST_SHADER_TEMPLATE[2])) {
        stError(aShadersError);
        return false;
    } else if(!aShaderColRev.init(ST_SHADER_TEMPLATE[0],
              // drop even column (starts from left)
              "if(int(mod(gl_FragCoord.x + 1.5, 2.0)) == 1) { discard; }\n",
              ST_SHADER_TEMPLATE[2])) {
        stError(aShadersError);
        return false;
    }
    myGlPrograms   [DEVICE_VINTERLACE]->create().attachShader(aVertexShader).attachShader(aShaderCol).link();
    myGlProgramsRev[DEVICE_VINTERLACE]->create().attachShader(aVertexShader).attachShader(aShaderColRev).link();

    // chessboard
    StGLFragmentShader aShaderChess(myGlPrograms[DEVICE_CHESSBOARD]->getTitle());
    StGLFragmentShader aShaderChessRev(myGlProgramsRev[DEVICE_CHESSBOARD]->getTitle());
    if(!aShaderChess.init(ST_SHADER_TEMPLATE[0],
                          "bool isEvenX = int(mod(floor(gl_FragCoord.x + 1.5), 2.0)) == 1;\n"
                          "bool isEvenY = int(mod(floor(gl_FragCoord.y + 1.5), 2.0)) != 1;\n"
                          "if((isEvenX && isEvenY) || (!isEvenX && !isEvenY)) { discard; }\n",
                           ST_SHADER_TEMPLATE[2])) {
        stError(aShadersError);
        return false;
    } else if(!aShaderChessRev.init(ST_SHADER_TEMPLATE[0],
              "bool isEvenX = int(mod(floor(gl_FragCoord.x + 1.5), 2.0)) == 1;\n"
              "bool isEvenY = int(mod(floor(gl_FragCoord.y + 1.5), 2.0)) != 1;\n"
              "if(!((isEvenX && isEvenY) || (!isEvenX && !isEvenY))) { discard; }\n",
              ST_SHADER_TEMPLATE[2])) {
        stError(aShadersError);
        return false;
    }
    myGlPrograms   [DEVICE_CHESSBOARD]->create().attachShader(aVertexShader).attachShader(aShaderChess).link();
    myGlProgramsRev[DEVICE_CHESSBOARD]->create().attachShader(aVertexShader).attachShader(aShaderChessRev).link();

    /// TODO (Kirill Gavrilov#3) fix shaders
    const StString aShadersRoot = StProcess::getStCoreFolder() + "shaders" + SYS_FS_SPLITTER
                                + ST_OUT_PLUGIN_NAME + SYS_FS_SPLITTER;
    StGLVertexShader stVShaderED("ED control");
    if(!stVShaderED.initFile(aShadersRoot + VSHADER_ED)) {
        stError(aShadersError);
        return false;
    }

    StGLFragmentShader stFInterlaceOn(myEDIntelaceOn->getTitle());
    if(!stFInterlaceOn.initFile(aShadersRoot + FSHADER_EDINTERLACE_ON)) {
        stError(aShadersError);
        return false;
    }
    myEDIntelaceOn->create().attachShader(stVShaderED).attachShader(stFInterlaceOn).link();

    StGLFragmentShader stFShaderEDOff(myEDOff->getTitle());
    if(!stFShaderEDOff.initFile(aShadersRoot + FSHADER_ED_OFF)) {
        stError(aShadersError);
        return false;
    }
    myEDOff->create().attachShader(stVShaderED).attachShader(stFShaderEDOff).link();

    myVpSizeYOnLoc = myEDIntelaceOn->getUniformLocation("vpSizeY");
    myVpSizeYOffLoc = myEDOff->getUniformLocation("vpSizeY");

    if(myDeviceId == DEVICE_HINTERLACE_ED) {
        // could be eDimensional shuttered glasses
        myEDTimer.restart(2000000.0);
    }

    // create vertices buffers to draw simple textured quad
    const GLfloat QUAD_VERTICES[4 * 4] = {
         1.0f, -1.0f, 0.0f, 1.0f, // top-right
         1.0f,  1.0f, 0.0f, 1.0f, // bottom-right
        -1.0f, -1.0f, 0.0f, 1.0f, // top-left
        -1.0f,  1.0f, 0.0f, 1.0f  // bottom-left
    };

    const GLfloat QUAD_TEXCOORD[2 * 4] = {
        1.0f, 0.0f,
        1.0f, 1.0f,
        0.0f, 0.0f,
        0.0f, 1.0f
    };

    myQuadVertBuf.init(4, 4, QUAD_VERTICES);
    myQuadTexCoordBuf.init(2, 4, QUAD_TEXCOORD);

    return true;
}

void StOutInterlace::callback(StMessage_t* stMessages) {
    myStCore->callback(stMessages);
    for(size_t i = 0; stMessages[i].uin != StMessageList::MSG_NULL; ++i) {
        switch(stMessages[i].uin) {
            case StMessageList::MSG_RESIZE: {
                StRectI_t rect = getStWindow()->getPlacement();
                myVpSizeY = rect.height();
                if(!getStWindow()->isFullScreen()) {
                    if(myMonitor.isNull()) {
                        myMonitor = new StMonitor(StCore::getMonitorFromPoint(rect.center()));
                    } else if(!myMonitor->getVRect().isPointIn(rect.center())) {
                        *myMonitor = StCore::getMonitorFromPoint(rect.center());
                    }
                    getStWindow()->stglMakeCurrent(ST_WIN_SLAVE);
                    StRectI_t edRect;
                    edRect.left()   = 0;
                    edRect.right()  = myMonitor->getVRect().width();
                    edRect.top()    = 0;
                    edRect.bottom() = 10;
                    stglResize(edRect);
                    myVpSizeY = 10;
                }
                getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
                stglResize(rect);
                break;
            }
            case StMessageList::MSG_KEYS: {
                bool* keysMap = ((bool* )stMessages[i].data);
                if(keysMap[ST_VK_F1]) {
                    setDevice(DEVICE_HINTERLACE);    keysMap[ST_VK_F1] = false;
                } else if(keysMap[ST_VK_F2]) {
                    setDevice(DEVICE_VINTERLACE);    keysMap[ST_VK_F2] = false;
                } else if(keysMap[ST_VK_F3]) {
                    setDevice(DEVICE_CHESSBOARD);    keysMap[ST_VK_F3] = false;
                } else if(keysMap[ST_VK_F4]) {
                    setDevice(DEVICE_HINTERLACE_ED); keysMap[ST_VK_F4] = false;
                }
                if(keysMap[ST_VK_F12]) {
                    myToShowFPS = !myToShowFPS;
                    keysMap[ST_VK_F12] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )myOptionsStruct->options[DEVICE_OPTION_SHOWFPS]);
                    option->value = myToShowFPS; msg.data = (void* )option;
                    getStWindow()->appendMessage(msg);
                }
                break;
            }
            case StMessageList::MSG_DEVICE_INFO: {
                if(myOptionsStruct->curDeviceId != myDeviceId) {
                    StString newPluginPath(myOptionsStruct->curRendererPath);
                    if(newPluginPath == myPluginPath) {
                        setDevice(myOptionsStruct->curDeviceId);
                        stMessages[i].uin = StMessageList::MSG_NONE;
                    } // else - another plugin
                }
                break;
            }
            case StMessageList::MSG_DEVICE_OPTION: {
                bool newVSync = ((StSDOnOff_t* )myOptionsStruct->options[DEVICE_OPTION_VSYNC])->value;
                if(newVSync != myIsVSync) {
                    myIsVSync = newVSync;
                    getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
                    stglSetVSync(myIsVSync);
                }
                myToShowFPS          = ((StSDOnOff_t* )myOptionsStruct->options[DEVICE_OPTION_SHOWFPS])->value;
                myIsReversed         = ((StSDOnOff_t* )myOptionsStruct->options[DEVICE_OPTION_REVERSE])->value;
                bool toBindToMonitor = ((StSDOnOff_t* )myOptionsStruct->options[DEVICE_OPTION_BINDMON])->value;

                bool toMovePos = !myToBindToMonitor && toBindToMonitor;
                myToBindToMonitor = toBindToMonitor;
                if(toMovePos && !getStWindow()->isFullScreen()) {
                    StRectI_t aRect = getStWindow()->getPlacement();
                    StMonitor aMon  = StCore::getMonitorFromPoint(aRect.center());
                    StHandle<StMonitor> anInterlacedMon = StOutInterlace::getHInterlaceMonitor();
                    if(!anInterlacedMon.isNull() && !isInterlacedMonitor(aMon)) {
                        int aWidth  = aRect.width();
                        int aHeight = aRect.height();
                        aRect.left()   = anInterlacedMon->getVRect().left() + 256;
                        aRect.right()  = aRect.left() + aWidth;
                        aRect.top()    = anInterlacedMon->getVRect().top() + 256;
                        aRect.bottom() = aRect.top() + aHeight;
                        getStWindow()->setPlacement(aRect);
                    }
                }

                break;
            }
            case StMessageList::MSG_EXIT: {
                if((myDeviceId == DEVICE_HINTERLACE_ED) && myIsEDactive) {
                    // disactivate eDimensional shuttered glasses
                    myEDTimer.restart();
                    myIsEDactive = false;
                    while(myEDTimer.getElapsedTime() <= 0.5) {
                        stglDraw(ST_DRAW_BOTH);
                        StThread::sleep(10);
                    }
                }
                break;
            }
        }
    }
}

void StOutInterlace::stglDrawEDCodes() {
    if(myEDTimer.getElapsedTime() > 0.5) {
        getStWindow()->hide(ST_WIN_SLAVE);
        myIsEDCodeFinished = true;
        return;
    }
    if(!getStWindow()->isFullScreen()) {
        getStWindow()->show(ST_WIN_SLAVE);
        getStWindow()->stglMakeCurrent(ST_WIN_SLAVE);
        // clear the screen and the depth buffer
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    if(myIsEDactive) {
        myEDIntelaceOn->use();
        if(myVpSizeYOnLoc != -1) {
            // TODO (Kirill Gavrilov#9#) use glViewport instead
            glUniform1i(myVpSizeYOnLoc, myVpSizeY);
        }
    } else {
        myEDOff->use();
        if(myVpSizeYOffLoc != -1) {
            // TODO (Kirill Gavrilov#9#) use glViewport instead
            glUniform1i(myVpSizeYOffLoc, myVpSizeY);
        }
    }
    // TODO (Kirill Gavrilov#5) use vertex buffer
    glBegin(GL_QUADS);
        glVertex2f(-1.0f, -1.0f);
        glVertex2f( 1.0f, -1.0f);
        glVertex2f( 1.0f,  1.0f);
        glVertex2f(-1.0f,  1.0f);
    glEnd();
    myEDIntelaceOn->unuse(); // this is global unuse
    glDisable(GL_BLEND);
    if(!getStWindow()->isFullScreen()) {
        getStWindow()->stglSwap(ST_WIN_SLAVE);
    }
}

void StOutInterlace::stglDraw(unsigned int ) {
    myFPSControl.setTargetFPS(getStWindow()->stglGetTargetFps());
    if(myToShowFPS && myFPSControl.isUpdated()) {
        getStWindow()->setTitle(StString("Interlace Rendering FPS= ") + myFPSControl.getAverage());
    }

    // always draw LEFT view into real screen buffer
    getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
    myStCore->stglDraw(ST_DRAW_LEFT);

    if(!getStWindow()->isStereoOutput()) {
        if(myDeviceId == DEVICE_HINTERLACE_ED) {
            // EDimensional disactivation
            // TODO (Kirill Gavrilov#4#) implement logic to sync multiple sView instances
            if(myIsEDCodeFinished) {
                if(myIsStereo) {
                    if(myIsEDactive) {
                        myEDTimer.restart();
                        myIsEDactive = false;
                        myIsEDCodeFinished = false;
                    }
                    myIsStereo = false;
                }
            }
            stglDrawEDCodes();
        }

        // decrease FPS to target by thread sleeps
        myFPSControl.sleepToTarget();
        getStWindow()->stglSwap(ST_WIN_MASTER);
        ++myFPSControl;
        return;
    }

    // reverse L/R according to window position
    bool isPixelReverse = false;
    StRectI_t aWinRect = getStWindow()->getPlacement();

    // odd vertically?
    if(!getStWindow()->isFullScreen() && aWinRect.bottom() % 2 == 1) {
        switch(myDeviceId) {
            case DEVICE_CHESSBOARD:
            case DEVICE_HINTERLACE:
            case DEVICE_HINTERLACE_ED:
                isPixelReverse = !isPixelReverse; break;
        }
    }

    // odd horizontally?
    if(!getStWindow()->isFullScreen() && aWinRect.left() % 2 == 1) {
        switch(myDeviceId) {
            case DEVICE_CHESSBOARD:
            case DEVICE_VINTERLACE:
                isPixelReverse = !isPixelReverse; break;
        }
    }

    // reversed by user?
    if(myIsReversed) {
        isPixelReverse = !isPixelReverse;
    }

    // update FBO rendering dimensions
    GLsizei aVPSizeX = stMin(aWinRect.width(),  myFrmBuffer->getSizeX());
    GLsizei aVPSizeY = stMin(aWinRect.height(), myFrmBuffer->getSizeY());
    if(aVPSizeX != myFrmBuffer->getVPSizeX()
    || aVPSizeY != myFrmBuffer->getVPSizeY()) {
        // reduce viewport to avoid additional aliasing of narrow lines
        GLfloat aDX = GLfloat(aVPSizeX) / GLfloat(myFrmBuffer->getSizeX());
        GLfloat aDY = GLfloat(aVPSizeY) / GLfloat(myFrmBuffer->getSizeY());
        StArray<StGLVec2> aTCoords(4);
        aTCoords[0] = StGLVec2(aDX,  0.0f);
        aTCoords[1] = StGLVec2(aDX,  aDY);
        aTCoords[2] = StGLVec2(0.0f, 0.0f);
        aTCoords[3] = StGLVec2(0.0f, aDY);
        myQuadTexCoordBuf.init(aTCoords);
        myFrmBuffer->setVPSizeX(aVPSizeX);
        myFrmBuffer->setVPSizeY(aVPSizeY);
    }

    // draw into virtual frame buffer
    glPushAttrib(GL_VIEWPORT_BIT); // we save previous (real window) viewport state
        myFrmBuffer->setupViewPort();   // we set TEXTURE sizes here
        myFrmBuffer->bindBuffer();
            myStCore->stglDraw(ST_DRAW_RIGHT);
        myFrmBuffer->unbindBuffer();
    glPopAttrib();                 // return real WINDOW viewport

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_BLEND);

    myFrmBuffer->bindTexture();
    const StHandle<StProgramFB>& aProgram = isPixelReverse ? myGlProgramsRev[myDeviceId] : myGlPrograms[myDeviceId];
    aProgram->use();
        myQuadVertBuf.bindVertexAttrib(ST_VATTRIB_VERTEX);
        myQuadTexCoordBuf.bindVertexAttrib(ST_VATTRIB_TCOORD);

        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

        myQuadTexCoordBuf.unBindVertexAttrib(ST_VATTRIB_TCOORD);
        myQuadVertBuf.unBindVertexAttrib(ST_VATTRIB_VERTEX);
    aProgram->unuse();
    myFrmBuffer->unbindTexture();

    if(myDeviceId == DEVICE_HINTERLACE_ED) {
        // EDimensional activation
        if(myIsEDCodeFinished) {
            if(!myIsStereo) {
                if(!myIsEDactive) {
                    myEDTimer.restart();
                    myIsEDactive = true;
                    myIsEDCodeFinished = false;
                }
                myIsStereo = true;
            }
        }
        stglDrawEDCodes();
    }

    // decrease FPS to target by thread sleeps
    myFPSControl.sleepToTarget();
    getStWindow()->stglSwap(ST_WIN_MASTER);
    ++myFPSControl;
}

// SDK version was used
ST_EXPORT void getSDKVersion(StVersion* ver) {
    *ver = StVersionInfo::getSDKVersion();
}

// plugin version
ST_EXPORT void getPluginVersion(StVersion* ver) {
    *ver = StVersionInfo::getSDKVersion();
}

// plugin description
static StString initAboutInfo() {
    StTranslations aLangMap(ST_OUT_PLUGIN_NAME);
    StString& aTitle = aLangMap.changeValueId(STTR_PLUGIN_TITLE,
        "sView - Interlaced Output library");
    StString& aVerString = aLangMap.changeValueId(STTR_VERSION_STRING,
        "version");
    StString& aDescr = aLangMap.changeValueId(STTR_PLUGIN_DESCRIPTION,
        "(C) 2009-2012 Kirill Gavrilov <kirill@sview.ru>\nOfficial site: www.sview.ru\n\nThis library distributed under LGPL3.0");
    return aTitle + '\n' + aVerString + ": " + StVersionInfo::getSDKVersionString() + "\n \n" + aDescr;
}

ST_EXPORT const stUtf8_t* getAboutInfo() {
    static const StString ABOUT_PLUGIN(initAboutInfo());
    return ABOUT_PLUGIN.toCString();
}

static StStereoDeviceInfo_t* prepareDevicesInfo() {
    int hSupportLevel = ST_DEVICE_SUPPORT_NONE;
    // Firstly INIT core library!
    if(StCore::INIT() != STERROR_LIBNOERROR) {
        ST_DEBUG_LOG(ST_OUT_PLUGIN_NAME + " Plugin, Core library not available!");
    } else {
        StHandle<StMonitor> aMon = StOutInterlace::getHInterlaceMonitor();
        if(!aMon.isNull()) {
            hSupportLevel = ST_DEVICE_SUPPORT_HIGHT;
            aMon.nullify();
        }
        StCore::FREE();
    }

    //
    StTranslations stLangMap(ST_OUT_PLUGIN_NAME);
    static StStereoDeviceInfo ST_DEVICE_INFO_HINTERLACE("StOutHInterlace",
        stLangMap.changeValueId(STTR_HINTERLACE_NAME, "Row Interlaced"),
        stLangMap.changeValueId(STTR_HINTERLACE_DESC, "Row interlaced displays: Zalman, Hyundai,..."),
        hSupportLevel);
    static StStereoDeviceInfo ST_DEVICE_INFO_VINTERLACE("StOutVInterlace",
        stLangMap.changeValueId(STTR_VINTERLACE_NAME, "Column Interlaced"),
        stLangMap.changeValueId(STTR_VINTERLACE_DESC, "Column interlaced displays"),
        ST_DEVICE_SUPPORT_NONE);
    static StStereoDeviceInfo ST_DEVICE_INFO_CHESSBOARD("StOutChessboard",
        stLangMap.changeValueId(STTR_CHESSBOARD_NAME, "DLP TV (chessboard)"),
        stLangMap.changeValueId(STTR_CHESSBOARD_DESC, "DLP TV (chessboard)"),
        ST_DEVICE_SUPPORT_NONE);
    static StStereoDeviceInfo ST_DEVICE_INFO_HINTERLACE_ED("StOutHInterlaceED",
        stLangMap.changeValueId(STTR_HINTERLACE_ED_NAME, "Interlaced ED"),
        stLangMap.changeValueId(STTR_HINTERLACE_ED_DESC, "EDimensional in interlaced mode"),
        ST_DEVICE_SUPPORT_NONE);
    static StStereoDeviceInfo_t ST_DEVICE_INFO_HINTERLACE_STRUCT    = ST_DEVICE_INFO_HINTERLACE.getStruct();
    static StStereoDeviceInfo_t ST_DEVICE_INFO_VINTERLACE_STRUCT    = ST_DEVICE_INFO_VINTERLACE.getStruct();
    static StStereoDeviceInfo_t ST_DEVICE_INFO_CHESSBOARD_STRUCT    = ST_DEVICE_INFO_CHESSBOARD.getStruct();
    static StStereoDeviceInfo_t ST_DEVICE_INFO_HINTERLACE_ED_STRUCT = ST_DEVICE_INFO_HINTERLACE_ED.getStruct();
    ST_DEVICE_INFO_HINTERLACE_STRUCT.nextInList = &ST_DEVICE_INFO_VINTERLACE_STRUCT;
    ST_DEVICE_INFO_VINTERLACE_STRUCT.nextInList = &ST_DEVICE_INFO_CHESSBOARD_STRUCT;
    ST_DEVICE_INFO_CHESSBOARD_STRUCT.nextInList = &ST_DEVICE_INFO_HINTERLACE_ED_STRUCT;
    return &ST_DEVICE_INFO_HINTERLACE_STRUCT;
}

ST_EXPORT const StStereoDeviceInfo_t* getSupportedDevicesInfo() {
    static const StStereoDeviceInfo_t* DEVICES_INFO = prepareDevicesInfo();
    return DEVICES_INFO;
}

ST_EXPORT StRendererInterface* StRenderer_new() {
    return new StOutInterlace(); }
ST_EXPORT void StRenderer_del(StRendererInterface* inst) {
    delete (StOutInterlace* )inst; }
ST_EXPORT StWindowInterface* StRenderer_getStWindow(StRendererInterface* inst) {
    // This is VERY important return libImpl pointer here!
    return ((StOutInterlace* )inst)->getStWindow()->getLibImpl(); }
ST_EXPORT stBool_t StRenderer_init(StRendererInterface* inst, const stUtf8_t* rendererPath, const int& theDeviceId, const StNativeWin_t* nativeParent) {
    return ((StOutInterlace* )inst)->init(StString(rendererPath), theDeviceId, nativeParent); }
ST_EXPORT stBool_t StRenderer_open(StRendererInterface* inst, const StOpenInfo_t* stOpenInfo) {
    return ((StOutInterlace* )inst)->open(StOpenInfo(stOpenInfo)); }
ST_EXPORT void StRenderer_callback(StRendererInterface* inst, StMessage_t* stMessages) {
    ((StOutInterlace* )inst)->callback(stMessages); }
ST_EXPORT void StRenderer_stglDraw(StRendererInterface* inst, unsigned int views) {
    ((StOutInterlace* )inst)->stglDraw(views); }
