/**
 * Copyright © 2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * StOutPageFlip library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StOutPageFlip library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StGLColoredLine.h"

#include <StGL/StGLProgram.h>

class ST_LOCAL StGLColoredLine::StColoredLineProgram : public StGLProgram {

        private:

    StGLVarLocation atrVVertexLoc;
    StGLVarLocation uniColorLoc;
    StGLVarLocation uniLineLenPxLoc;

        public:

    StColoredLineProgram()
    : StGLProgram("StColoredLine"),
      atrVVertexLoc(),
      uniColorLoc(),
      uniLineLenPxLoc() {
        //
    }

    virtual ~StColoredLineProgram() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    virtual bool init() {
        const char VERTEX_SHADER[] =
           "attribute vec4 vVertex; \
            void main(void) { \
                gl_Position = vVertex; \
            }";

        const char FRAGMENT_SHADER[] =
           "uniform vec4 uColor; \
            uniform int uLineLenPx; \
            void main(void) { \
                int yFromBottom = int(floor(gl_FragCoord.y - 0.5)); \
                int xFromLeft = int(floor(gl_FragCoord.x - 0.5)); \
                if(yFromBottom == 0 && xFromLeft < uLineLenPx) { \
                    gl_FragColor = uColor; \
                } else { \
                    gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0); \
                } \
            }";

        StGLVertexShader stVertexShader(StGLProgram::getTitle());
        stVertexShader.init(VERTEX_SHADER);

        StGLFragmentShader stFragmentShader(StGLProgram::getTitle());
        stFragmentShader.init(FRAGMENT_SHADER);
        if(!StGLProgram::create().attachShader(stVertexShader).attachShader(stFragmentShader).link()) {
            return false;
        }

        atrVVertexLoc = StGLProgram::getAttribLocation("vVertex");
        uniColorLoc   = StGLProgram::getUniformLocation("uColor");
        uniLineLenPxLoc = StGLProgram::getUniformLocation("uLineLenPx");

        return atrVVertexLoc.isValid() && uniColorLoc.isValid() && uniLineLenPxLoc.isValid();
    }

    void use(const StGLVec4& colorVec4, const GLint lineLength) {
        StGLProgram::use();
        glUniform4fv(uniColorLoc, 1, colorVec4);
        glUniform1i(uniLineLenPxLoc, lineLength);
    }

};

StGLColoredLine::StGLColoredLine()
: StGLDeviceControl(),
  stProgram(NULL) {
    setWhiteColor();
}

StGLColoredLine::~StGLColoredLine() {
    stglRelease();
}

bool StGLColoredLine::stglInit() {
    const GLfloat QUAD_VERTICES[4 * 4] = {
         1.0f,  1.0f, 0.0f, 1.0f, // top-right
         1.0f, -1.0f, 0.0f, 1.0f, // bottom-right
        -1.0f,  1.0f, 0.0f, 1.0f, // top-left
        -1.0f, -1.0f, 0.0f, 1.0f  // bottom-left
    };
    stVertexBuf.init(4, 4, QUAD_VERTICES);

    stProgram = new StColoredLineProgram();
    return stProgram->init();
}

void StGLColoredLine::stglRelease() {
    stVertexBuf.release();
    delete stProgram;
    stProgram = NULL;
}

void StGLColoredLine::stglDraw(unsigned int view, const int windowWidth, const int ) {
    if(!isActive() || stProgram == NULL || !stProgram->isValid()) {
        return;
    }

    GLint lineLength = (view == ST_DRAW_LEFT) ? (windowWidth / 4) : (3 * windowWidth / 4);

    glEnable(GL_SCISSOR_TEST);
    glScissor(0, 0, lineLength, 1);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    stProgram->use(lineColor, lineLength);
        stVertexBuf.bindVertexAttrib(stProgram->getVVertexLoc());
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
        stVertexBuf.unBindVertexAttrib(stProgram->getVVertexLoc());
    stProgram->unuse();
    glDisable(GL_BLEND);

    glDisable(GL_SCISSOR_TEST);
}
