/**
 * Copyright © 2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * StOutPageFlip library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StOutPageFlip library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StGLControlED.h"

#include <StGL/StGLProgram.h>

class ST_LOCAL StGLControlED::StEDProgram : public StGLProgram {

        private:

    StGLVarLocation atrVVertexLoc;
    StGLVarLocation uniVPSizeYLoc;

        public:

    StEDProgram(const StString& title)
    : StGLProgram(title),
      atrVVertexLoc(),
      uniVPSizeYLoc() {
        //
    }

    virtual ~StEDProgram() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    virtual bool link() {
        if(!StGLProgram::link()) {
            return false;
        }

        atrVVertexLoc = StGLProgram::getAttribLocation("vVertex");
        uniVPSizeYLoc = StGLProgram::getUniformLocation("uVPSizeY");

        return atrVVertexLoc.isValid() && uniVPSizeYLoc.isValid();
    }

    void use(const GLint vpSizeY) {
        StGLProgram::use();
        glUniform1i(uniVPSizeYLoc, vpSizeY);
    }

};

const double StGLControlED::DELAY_MS = 500.0;

StGLControlED::StGLControlED()
: StGLDeviceControl(),
  stProgramOn(NULL),
  stProgramOff(NULL),
  stProgramBlack(NULL),
  stTimerCode(false),
  stTimerBlack(false) {
    //
}

StGLControlED::~StGLControlED() {
    stglRelease();
}

bool StGLControlED::stglInit() {
    const GLfloat QUAD_VERTICES[4 * 4] = {
         1.0f,  1.0f, 0.0f, 1.0f, // top-right
         1.0f, -1.0f, 0.0f, 1.0f, // bottom-right
        -1.0f,  1.0f, 0.0f, 1.0f, // top-left
        -1.0f, -1.0f, 0.0f, 1.0f  // bottom-left
    };
    stVertexBuf.init(4, 4, QUAD_VERTICES);

    stProgramOn    = new StEDProgram("StGLControlED ON");
    stProgramOff   = new StEDProgram("StGLControlED OFF");
    stProgramBlack = new StEDProgram("StGLControlED Black");

    const char VERTEX_SHADER[] =
       "attribute vec4 vVertex; \
        void main(void) { \
            gl_Position = vVertex; \
        }";

    const char FRAGMENT_SHADER_ON[] =
       "uniform int uVPSizeY; \
        void main(void) { \
            int yFromTop = uVPSizeY - int(floor(gl_FragCoord.y - 0.5)) - 1; \
            if(yFromTop >= 0 && yFromTop <= 1) { \
                gl_FragColor = vec4(1.0, 0.0, 0.0, 1.0); \
            } else if(yFromTop >= 2 && yFromTop <= 3) { \
                gl_FragColor = vec4(0.0, 1.0, 0.0, 1.0); \
            } else if(yFromTop >= 4 && yFromTop <= 7) { \
                gl_FragColor = vec4(1.0, 1.0, 0.0, 1.0); \
            } else if(yFromTop >= 8 && yFromTop <= 9) { \
                gl_FragColor = vec4(0.0, 1.0, 0.0, 1.0); \
            } else { \
                gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0); \
            } \
        }";

    const char FRAGMENT_SHADER_OFF[] =
       "uniform int uVPSizeY; \
        void main(void) { \
            int yFromTop = uVPSizeY - int(floor(gl_FragCoord.y - 0.5)) - 1; \
            if(yFromTop >= 0 && yFromTop <= 1) { \
                gl_FragColor = vec4(1.0, 0.0, 0.0, 1.0); \
            } else if(yFromTop >= 2 && yFromTop <= 3) { \
                gl_FragColor = vec4(0.0, 1.0, 0.0, 1.0); \
            } else if(yFromTop >= 4 && yFromTop <= 7) { \
                gl_FragColor = vec4(1.0, 1.0, 0.0, 1.0); \
            } else if(yFromTop >= 8 && yFromTop <= 9) { \
                gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0); \
            } else { \
                gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0); \
            } \
        }";

    const char FRAGMENT_SHADER_BLACK[] =
       "uniform int uVPSizeY; \
        void main(void) { \
            int yFromTop = uVPSizeY - int(floor(gl_FragCoord.y - 0.5)) - 1; \
            if(yFromTop >= 0 && yFromTop <= 9) { \
                gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0); \
            } else { \
                gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0); \
            } \
        }";

    StGLVertexShader stVertexShader("StGLControlED");
    stVertexShader.init(VERTEX_SHADER);

    StGLFragmentShader stFPageFlipOn(stProgramOn->getTitle());
    stFPageFlipOn.init(FRAGMENT_SHADER_ON);
    stProgramOn->create().attachShader(stVertexShader).attachShader(stFPageFlipOn).link();

    StGLFragmentShader stFPageFlipOff(stProgramOff->getTitle());
    stFPageFlipOff.init(FRAGMENT_SHADER_OFF);
    stProgramOff->create().attachShader(stVertexShader).attachShader(stFPageFlipOff).link();

    StGLFragmentShader stFBlack(stProgramBlack->getTitle());
    stFBlack.init(FRAGMENT_SHADER_BLACK);
    stProgramBlack->create().attachShader(stVertexShader).attachShader(stFBlack).link();

    return stProgramOn->isValid() && stProgramOff->isValid() && stProgramBlack->isValid();
}

void StGLControlED::stglRelease() {
    stVertexBuf.release();
    delete stProgramOn;    stProgramOn = NULL;
    delete stProgramOff;   stProgramOff = NULL;
    delete stProgramBlack; stProgramBlack = NULL;
}

void StGLControlED::stglDraw(unsigned int , const int windowWidth, const int windowHeight) {
    if(!isActive() || stProgramOn  == NULL   || !stProgramOn->isValid()
                   || stProgramOff == NULL   || !stProgramOff->isValid()
                   || stProgramBlack == NULL || !stProgramBlack->isValid()) {
        return;
    }

    StEDProgram* stProgram = stTimerBlack.isOn() ? stProgramBlack : (isStereo() ? stProgramOn : stProgramOff);

    glEnable(GL_SCISSOR_TEST);
    glScissor(0, windowHeight - 10, windowWidth, 10);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    stProgram->use(windowHeight);
        stVertexBuf.bindVertexAttrib(stProgram->getVVertexLoc());
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
        stVertexBuf.unBindVertexAttrib(stProgram->getVVertexLoc());
    stProgram->unuse();
    glDisable(GL_BLEND);

    glDisable(GL_SCISSOR_TEST);
}
