/**
 * Copyright © 2012 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGL/StGLFont.h>

StGLFont::StGLFont(StHandle<StFTFont>& theFont,
                   const GLint         theTextureFormat)
: myFont(theFont),
  myAscender(0.0f),
  myLineSpacing(0.0f),
  myTileSizeX(0),
  myTileSizeY(0),
  myLastTileId(-1),
  myTextureFormat(theTextureFormat) {
    stMemSet(&myLastTilePx, 0, sizeof(myLastTilePx));
}

bool StGLFont::stglInit() {
    ///release();
    if(myFont.isNull() || !myFont->isValid()) {
        return false;
    }

    myAscender    = myFont->getAscender();
    myLineSpacing = myFont->getLineSpacing();
    myTileSizeX   = myFont->getGlyphMaxSizeX();
    myTileSizeY   = myFont->getGlyphMaxSizeY();

    myLastTileId = -1;
    return createTexture();
}

bool StGLFont::createTexture() {
    GLint aMaxSize = 1024;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &aMaxSize);

    GLint aGlyphsNb = myFont->getGlyphsNumber() - GLint(myLastTileId) + 1;

    const GLsizei aTextureSizeX = getPowerOfTwo(aGlyphsNb * myTileSizeX, aMaxSize);
    const size_t  aTilesPerRow  = aTextureSizeX / myTileSizeX;
    const GLsizei aTextureSizeY = getPowerOfTwo(GLint((aGlyphsNb / aTilesPerRow) + 1) * myTileSizeY, aMaxSize);

    stMemSet(&myLastTilePx, 0, sizeof(myLastTilePx));
    myLastTilePx.bottom() = myTileSizeY;

    myTextures.add(new StGLTexture(myTextureFormat));
    StHandle<StGLTexture>& aTexture = myTextures[myTextures.size() - 1];
    ///return aTexture->initTrash(aTextureSizeX, aTextureSizeY);
    if(!aTexture->initBlack(aTextureSizeX, aTextureSizeY)) {
        return false;
    }

    aTexture->bind();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    aTexture->unbind();
    return true;
}

bool StGLFont::renderGlyph(const stUtf32_t theChar) {
    if(!myFont->renderGlyph(theChar)) {
        return false;
    }

    StHandle<StGLTexture>& aTexture = myTextures[myTextures.size() - 1];

    const StImagePlane& anImg = myFont->getGlyphImage();
    const size_t aTileId = myLastTileId + 1;
    myLastTilePx.left()  = myLastTilePx.right() + 3;
    myLastTilePx.right() = myLastTilePx.left() + anImg.getSizeX();
    if(myLastTilePx.right() >= aTexture->getSizeX()) {
        myLastTilePx.left()    = 0;
        myLastTilePx.right()   = anImg.getSizeX();
        myLastTilePx.top()    += myTileSizeY;
        myLastTilePx.bottom() += myTileSizeY;

        if(myLastTilePx.bottom() >= aTexture->getSizeY()) {
            if(!createTexture()) {
                return false;
            }
            return renderGlyph(theChar);
        }
    }

    /// TODO
    aTexture->bind();
    glPixelStorei(GL_UNPACK_LSB_FIRST,  GL_FALSE);
    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    glPixelStorei(GL_UNPACK_ALIGNMENT,  1);

    glTexSubImage2D(GL_TEXTURE_2D, 0,
                    myLastTilePx.left(), myLastTilePx.top(), anImg.getSizeX(), anImg.getSizeY(),
                    GL_ALPHA, GL_UNSIGNED_BYTE, anImg.getData());

    StGLTile aTile;
    aTile.uv.left()   = GLfloat(myLastTilePx.left())                   / GLfloat(aTexture->getSizeX());
    aTile.uv.right()  = GLfloat(myLastTilePx.right())                  / GLfloat(aTexture->getSizeX());
    aTile.uv.top()    = GLfloat(myLastTilePx.top())                    / GLfloat(aTexture->getSizeY());
    aTile.uv.bottom() = GLfloat(myLastTilePx.top() + anImg.getSizeY()) / GLfloat(aTexture->getSizeY());
    aTile.texture     = aTexture->getTextureId();
    myFont->getGlyphRect(aTile.px);

    myLastTileId = aTileId;
    myTiles.add(aTile);
    return true;
}

void StGLFont::renderGlyph(const stUtf32_t theUChar,
                           const stUtf32_t theUCharNext,
                           StGLTile&       theGlyph,
                           StGLVec2&       thePen) {
    std::map<stUtf32_t, size_t>::iterator aTileIter = myGlyphMap.find(theUChar);
    size_t aTileId;
    if(aTileIter == myGlyphMap.end()) {
        if(renderGlyph(theUChar)) {
            aTileId = myLastTileId;
        } else {
            thePen.x() += myFont->getAdvanceX(theUChar, theUCharNext);
            return;
        }
        myGlyphMap[theUChar] = aTileId;
    } else {
        aTileId = aTileIter->second;
    }

    const StGLTile& aTile = myTiles[aTileId];
    StGLRect aRect(thePen.y() + aTile.px.top(),  thePen.y() + aTile.px.bottom(),
                   thePen.x() + aTile.px.left(), thePen.x() + aTile.px.right());
    theGlyph.px      = aRect;
    theGlyph.uv      = aTile.uv;
    theGlyph.texture = aTile.texture;

    thePen.x() += myFont->getAdvanceX(theUChar, theUCharNext);
}
