/**
 * Copyright © 2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLMesh/StGLMesh.h>

void StGLMesh::drawKernel() const {
    if(myIndexBuf.isValid()) {
        // has indices
        myIndexBuf.bind();
        glDrawElements(myPrimitives, GLsizei(myIndexBuf.getElemsCount()), myIndexBuf.getDataType(), NULL);
        myIndexBuf.unbind();
    } else {
        // just draw all vertices
        glDrawArrays(myPrimitives, 0, static_cast<GLsizei>(myVertexBuf.getElemsCount()));
    }
}

void StGLMesh::bind(const StGLMeshProgram& theProgram) const {
    myVertexBuf.bindVertexAttrib(theProgram.getLocation(ST_VBO_VERTEX));
    myNormalBuf.bindVertexAttrib(theProgram.getLocation(ST_VBO_NORMAL));
    myTCoordBuf.bindVertexAttrib(theProgram.getLocation(ST_VBO_TCOORD));
    myColorsBuf.bindVertexAttrib(theProgram.getLocation(ST_VBO_COLORS));
}

void StGLMesh::unbind(const StGLMeshProgram& theProgram) const {
    myColorsBuf.unBindVertexAttrib(theProgram.getLocation(ST_VBO_COLORS));
    myTCoordBuf.unBindVertexAttrib(theProgram.getLocation(ST_VBO_TCOORD));
    myNormalBuf.unBindVertexAttrib(theProgram.getLocation(ST_VBO_NORMAL));
    myVertexBuf.unBindVertexAttrib(theProgram.getLocation(ST_VBO_VERTEX));
}

void StGLMesh::bindFixed() const {
    if(myVertexBuf.isValid()) {
        myVertexBuf.bind();
        glEnableClientState(GL_VERTEX_ARRAY);
        glVertexPointer(GLint(myVertexBuf.getElemSize()), myVertexBuf.getDataType(), 0, NULL);
    } else {
        glDisableClientState(GL_VERTEX_ARRAY);
    }
    if(myNormalBuf.isValid()) {
        //ST_DEBUG_ASSERT(myNormalBuf.getElemSize() == 3);
        myNormalBuf.bind();
        glEnableClientState(GL_NORMAL_ARRAY);
        glNormalPointer(myNormalBuf.getDataType(), 0, NULL);
    } else {
        glDisableClientState(GL_NORMAL_ARRAY);
    }
    if(myTCoordBuf.isValid()) {
        myTCoordBuf.bind();
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(GLint(myTCoordBuf.getElemSize()), myTCoordBuf.getDataType(), 0, NULL);
    } else {
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
    if(myColorsBuf.isValid()) {
        myColorsBuf.bind();
        //ST_DEBUG_ASSERT(myColorsBuf.getElemSize() == 3 || myColorsBuf.getElemSize() == 4);
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(GLint(myColorsBuf.getElemSize()), myColorsBuf.getDataType(), 0, NULL);
    } else {
        glDisableClientState(GL_COLOR_ARRAY);
    }
}

void StGLMesh::unbindFixed() const {
    if(myColorsBuf.isValid()) {
        myColorsBuf.unbind();
        glDisableClientState(GL_COLOR_ARRAY);
    }
    if(myTCoordBuf.isValid()) {
        myTCoordBuf.unbind();
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
    if(myNormalBuf.isValid()) {
        myNormalBuf.unbind();
        glDisableClientState(GL_NORMAL_ARRAY);
    }
    if(myVertexBuf.isValid()) {
        myVertexBuf.unbind();
        glDisableClientState(GL_VERTEX_ARRAY);
    }
}

StGLMesh::StGLMesh(GLenum thePrimitives)
: myBndSphere(),
  myVertices(1),
  myNormals(1),
  myTCoords(1),
  myColors(1),
  myIndices(1),
  myPrimitives(thePrimitives) {
    //
}

StGLMesh::~StGLMesh() {
    //
}

bool StGLMesh::computeMesh() {
    // dummy - we can not compute mesh from nowhere
    // however in general alien mesh could be set just as is
    if(!myVertices.isEmpty()) {
        if(myBndSphere.isVoid()) {
            myBndSphere.enlarge(myVertices);
        }
        return true;
    }
    return false;
};

bool StGLMesh::initVBOs() {
    // reset all current VBOs
    clearVRAM();
    if(myVertices.isEmpty() && !computeMesh()) {
        // no vertices - invalid mesh
        return false;
    }
    if(myVertices.isEmpty()) {
        // no vertices - invalid mesh
        return false;
    }
    bool isOK = myVertexBuf.init(myVertices);
    if(!myNormals.isEmpty() && myNormals.size() == myVertices.size()) {
        isOK = isOK && myNormalBuf.init(myNormals);
    }
    if(!myTCoords.isEmpty() && myTCoords.size() == myVertices.size()) {
        isOK = isOK && myTCoordBuf.init(myTCoords);
    }
    if(!myColors.isEmpty() && myColors.size() == myVertices.size()) {
        isOK = isOK && myColorsBuf.init(myColors);
    }
    if(!myIndices.isEmpty()) {
        isOK = isOK && myIndexBuf.init(myIndices);
    }
    return isOK;

};

bool StGLMesh::check(const StGLMeshProgram& theProgram) const {
    return !(theProgram.getLocation(ST_VBO_VERTEX).isValid() && !myVertexBuf.isValid())
        && !(theProgram.getLocation(ST_VBO_NORMAL).isValid() && !myNormalBuf.isValid())
        && !(theProgram.getLocation(ST_VBO_TCOORD).isValid() && !myTCoordBuf.isValid())
        && !(theProgram.getLocation(ST_VBO_COLORS).isValid() && !myColorsBuf.isValid());
}

void StGLMesh::draw(const StGLMeshProgram& theProgram) const {
    bind(theProgram);
    drawKernel();
    unbind(theProgram);
}

void StGLMesh::drawFixed() const {
    bindFixed();
    drawKernel();
    unbindFixed();
}

void StGLMesh::clearRAM() {
    myBndSphere.reset(); // should we?
    myVertices.initList(1);
    myNormals.initList(1);
    myTCoords.initList(1);
    myColors.initList(1);
    myIndices.initList(1);
}

void StGLMesh::clearVRAM() {
    myVertexBuf.release();
    myNormalBuf.release();
    myTCoordBuf.release();
    myColorsBuf.release();
    myIndexBuf.release();
}

bool StGLMesh::computeNormals(size_t theDelta) {
    ST_DEBUG_ASSERT(theDelta > 0);
    myNormals.initArray(myVertices.size());
    if(myVertices.isEmpty()) {
        return false;
    }

    // iterate over each triangle
    // for each node we compute summary of normals for all triangles where this node used
    // normals are NOT normalized per triangle - this allows to interpolate result normal
    // with respect to each triangle dimensions
    GLuint aV1, aV2, aV3;
    StGLVec3 aNorm;
    if(myIndices.size() >= 3) {
        size_t aLimit = myIndices.size() - 3;
        for(size_t anIndexId = 0; anIndexId <= aLimit; anIndexId += theDelta) {
            aV1 = myIndices[anIndexId];
            aV2 = myIndices[anIndexId + 1];
            aV3 = myIndices[anIndexId + 2];
            const StGLVec3& aVert1 = myVertices[aV1];
            const StGLVec3& aVert2 = myVertices[aV2];
            const StGLVec3& aVert3 = myVertices[aV3];
            aNorm = StGLVec3::cross(aVert2 - aVert1, aVert3 - aVert1);
            myNormals.changeValue(aV1) += aNorm;
            myNormals.changeValue(aV2) += aNorm;
            myNormals.changeValue(aV3) += aNorm;
        }
    } else if(myVertices.size() >= 3) {
        size_t aLimit = myVertices.size() - 3;
        for(size_t aVertId = 0; aVertId <= aLimit; aVertId += theDelta) {
            aV1 = GLuint(aVertId);
            aV2 = GLuint(aVertId + 1);
            aV3 = GLuint(aVertId + 2);
            const StGLVec3& aVert1 = myVertices[aV1];
            const StGLVec3& aVert2 = myVertices[aV2];
            const StGLVec3& aVert3 = myVertices[aV3];
            aNorm = StGLVec3::cross(aVert2 - aVert1, aVert3 - aVert1);
            myNormals.changeValue(aV1) += aNorm;
            myNormals.changeValue(aV2) += aNorm;
            myNormals.changeValue(aV3) += aNorm;
        }
    } else {
        return false;
    }

    // normalize normals (important for OpenGL)
    for(size_t aNormId = 0; aNormId < myNormals.size(); ++aNormId) {
        myNormals.changeValue(aNormId).normalize();
    }
    return true;
}
