/**
 * Copyright © 2010-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLStereo/StGLProjCamera.h>

#include <gl/glew.h>

namespace {
    static const GLfloat ST_DTR_HALF = 0.5f * 0.0174532925f; // (degrees -> radians) * 0.5
};

StGLProjCamera::StGLProjCamera(GLfloat theFOVy,
                               GLfloat theZNear,
                               GLfloat theZFar,
                               GLfloat theZScreen)
: myMatrix(),
  myMatrixMono(),
  myFOVy(theFOVy),
  myZoom(1.0f),
  myAspect(1.0f),
  myZScreen(theZScreen),
  myIOD(0.5f),
  myFrustL(),
  myFrustR(),
  myFrustM(),
  myFrust(NULL),
  myIsPersp(true) {
    //
    myFrust = &myFrustM;
    myFrustM.zNear = myFrustL.zNear = myFrustR.zNear = theZNear;
    myFrustM.zFar  = myFrustL.zFar  = myFrustR.zFar  = theZFar;
    updateFrustum();
}

StGLProjCamera::~StGLProjCamera() {
    //
}

void StGLProjCamera::getZParams(GLdouble theZValue, StRectD_t& theSectRect) const {
    if(myIsPersp) {
        theSectRect.top() = GLdouble(myZoom) * theZValue * std::tan(ST_DTR_HALF * myFOVy);
    } else {
        // Z-value doesn't change the section
        theSectRect.top() = GLdouble(myZoom) * myFrustM.zNear;/// * std::tan(ST_DTR_HALF * myFOVy);
    }
    theSectRect.bottom() = -theSectRect.top();
    theSectRect.left()   = -myAspect * theSectRect.top();
    theSectRect.right()  = -theSectRect.left();
}

void StGLProjCamera::setZScreen(GLfloat theZScreen) {
    // save the relationship
    myIOD = myIOD / myZScreen * theZScreen;
    myFrustM.zNear = myFrustL.zNear = myFrustR.zNear = myFrustM.zNear / myZScreen * theZScreen;
    // set new value
    myZScreen = theZScreen;
}

void StGLProjCamera::setPerspective(bool theIsPerspective) {
    myIsPersp = theIsPerspective;
    updateFrustum();
}

void StGLProjCamera::resize(GLsizei sizeX, GLsizei sizeY) {
    if(sizeY == 0) {  // prevent a Divide by Zero by
        sizeY = 1;    // making height equal one
    }
    myAspect = GLfloat(sizeX) / GLfloat(sizeY);

    glViewport(0, 0, sizeX, sizeY); // reset the current viewport
    updateFrustum();
}

GLfloat StGLProjCamera::getFOVyZoomed() const {
    GLfloat aDYHalf = myFrustM.zNear * std::tan(ST_DTR_HALF * myFOVy); /// ??? myIsPersp
    return std::atan2(aDYHalf, myFrustM.zNear * myZoom);
}

void StGLProjCamera::updateFrustum() {
    // sets top of frustum based on FOVy and near clipping plane
    GLfloat aDYHalf = myIsPersp ? (myZoom * myFrustM.zNear * std::tan(ST_DTR_HALF * myFOVy)) : (myZoom * myFrustM.zNear); ///
    // sets right of frustum based on aspect ratio
    GLfloat aDXHalf = myAspect * aDYHalf;
    GLfloat aDXStereoShift = (0.5f * myIOD) * myFrustM.zNear / myZScreen;

    // frustum for left view
    myFrustL.yTop    =  aDYHalf;
    myFrustL.yBottom = -aDYHalf;
    myFrustL.xLeft   = -aDXHalf + aDXStereoShift;
    myFrustL.xRight  =  aDXHalf + aDXStereoShift;
    myFrustL.xTranslation =  0.5f * myIOD; // X translation to cancel parallax
    // frustum for right view
    myFrustR.yTop    =  aDYHalf;
    myFrustR.yBottom = -aDYHalf;
    myFrustR.xLeft   = -aDXHalf - aDXStereoShift;
    myFrustR.xRight  =  aDXHalf - aDXStereoShift;
    myFrustR.xTranslation = -0.5f * myIOD; // X translation to cancel parallax
    // frustum for mono view
    myFrustM.yTop    =  aDYHalf;
    myFrustM.yBottom = -aDYHalf;
    myFrustM.xLeft   = -aDXHalf;
    myFrustM.xRight  =  aDXHalf;
    myFrustM.xTranslation = 0.0f;

    // update current matrix
    setupMatrix();
}

void StGLProjCamera::setView(unsigned int theView) {
    switch(theView) {
        case ST_DRAW_LEFT:
            myFrust = &myFrustL;
            break;
        case ST_DRAW_RIGHT:
            myFrust = &myFrustR;
            break;
        case ST_DRAW_MONO:
        default:
            myFrust = &myFrustM;
    }
    // update current matrix
    setupMatrix();
}

void StGLProjCamera::setupMatrix() {
    if(myIsPersp) {
        myMatrix.initFrustum(*myFrust);
        myMatrixMono.initFrustum(myFrustM);
    } else {
        myMatrix.initOrtho(myFrustM);
        myMatrixMono.initOrtho(myFrustM);
    }
}

void StGLProjCamera::setupFixed() {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if(myIsPersp) {
        // setup current frustum
        glFrustum(myFrust->xLeft, myFrust->xRight,
                  myFrust->yBottom, myFrust->yTop,
                  myFrust->zNear, myFrust->zFar);
        // translate to cancel parallax
        glTranslatef(myFrust->xTranslation, 0.0f, 0.0f);
    } else {
        glOrtho(myFrustM.xLeft, myFrustM.xRight,
                myFrustM.yBottom, myFrustM.yTop,
                myFrustM.zNear, myFrustM.zFar);
    }

    // turn back to model view matrix
    glMatrixMode(GL_MODELVIEW);
}

StString StGLProjCamera::toString() const {
    StRectD_t aSect; getZParams(aSect);
    return StString("ProjCamera, FOV= ") + myFOVy
       + ";\n Z-Near=   " + myFrustM.zNear
       + "; Z-Screen= "   + myZScreen
       + "; Z-Far=    "   + myFrustM.zFar
       + "; IOD= "    + myIOD
       + ";\nscr L= " + aSect.left()
       + "; R= "      + aSect.right()
       + "; B= "      + aSect.bottom()
       + "; T= "      + aSect.top();
}
