/**
 * Copyright © 2009-2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLStereo/StGLStereoFrameBuffer.h>

bool StGLStereoFrameBuffer::StGLStereoProgram::link() {
    if(!StGLProgram::link()) {
        return false;
    }
    StGLVarLocation uniTexLLoc = StGLProgram::getUniformLocation("texL");
    StGLVarLocation uniTexRLoc = StGLProgram::getUniformLocation("texR");
    atrVVertexLoc   = StGLProgram::getAttribLocation("vVertex");
    atrVTexCoordLoc = StGLProgram::getAttribLocation("vTexCoord");

    if(uniTexLLoc.isValid() && uniTexRLoc.isValid()) {
        use();
        glUniform1i(uniTexLLoc, StGLProgram::TEXTURE_SAMPLE_0); // GL_TEXTURE0 in multitexture
        glUniform1i(uniTexRLoc, StGLProgram::TEXTURE_SAMPLE_1); // GL_TEXTURE1 in multitexture
        unuse();
    }

    return uniTexRLoc.isValid() && uniTexLLoc.isValid() && atrVVertexLoc.isValid() && atrVTexCoordLoc.isValid();
}

StGLStereoFrameBuffer::StGLStereoFrameBuffer(GLint textureFormat)
: StGLStereoTexture(textureFormat),
  myVerticesBuf(),
  myTexCoordBuf(),
  myViewPortX(0),
  myViewPortY(0) {
    //
    myGLFBufferIds[StGLStereoTexture::LEFT_TEXTURE ] = StGLFrameBuffer::NO_FRAMEBUFFER;
    myGLFBufferIds[StGLStereoTexture::RIGHT_TEXTURE] = StGLFrameBuffer::NO_FRAMEBUFFER;
}

void StGLStereoFrameBuffer::release() {
    // release texture
    StGLStereoTexture::release();
    // release frame buffers
    if(myGLFBufferIds[StGLStereoTexture::LEFT_TEXTURE] != StGLFrameBuffer::NO_FRAMEBUFFER) {
        glDeleteFramebuffersEXT(1, &myGLFBufferIds[StGLStereoTexture::LEFT_TEXTURE]);
        myGLFBufferIds[StGLStereoTexture::LEFT_TEXTURE] = StGLFrameBuffer::NO_FRAMEBUFFER;
    }
    if(myGLFBufferIds[StGLStereoTexture::RIGHT_TEXTURE] != StGLFrameBuffer::NO_FRAMEBUFFER) {
        glDeleteFramebuffersEXT(1, &myGLFBufferIds[StGLStereoTexture::RIGHT_TEXTURE]);
        myGLFBufferIds[StGLStereoTexture::RIGHT_TEXTURE] = StGLFrameBuffer::NO_FRAMEBUFFER;
    }
    // release render buffers
    if(myGLDepthRBIds[StGLStereoTexture::LEFT_TEXTURE] != StGLFrameBuffer::NO_RENDERBUFFER) {
        glDeleteFramebuffersEXT(1, &myGLDepthRBIds[StGLStereoTexture::LEFT_TEXTURE]);
        myGLDepthRBIds[StGLStereoTexture::LEFT_TEXTURE] = StGLFrameBuffer::NO_RENDERBUFFER;
    }
    if(myGLDepthRBIds[StGLStereoTexture::RIGHT_TEXTURE] != StGLFrameBuffer::NO_RENDERBUFFER) {
        glDeleteFramebuffersEXT(1, &myGLDepthRBIds[StGLStereoTexture::RIGHT_TEXTURE]);
        myGLDepthRBIds[StGLStereoTexture::RIGHT_TEXTURE] = StGLFrameBuffer::NO_RENDERBUFFER;
    }
    // release VBOs
    myVerticesBuf.release();
    myTexCoordBuf.release();
}

bool StGLStereoFrameBuffer::init(GLsizei textureSizeX, GLsizei textureSizeY) {
    // release current objects
    release();

    // create the textures
    if(!StGLStereoTexture::initTrash(textureSizeX, textureSizeY)) {
        release();
        return false;
    }

    // generate RenderBuffers (will be used as depth buffer)
    glGenRenderbuffersEXT(2, myGLDepthRBIds);

    // create left RenderBuffer (will be used as depth buffer)
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, myGLDepthRBIds[StGLStereoTexture::LEFT_TEXTURE]);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, textureSizeX, textureSizeY);

    // build FBOs
    glGenFramebuffersEXT(2, myGLFBufferIds);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, myGLFBufferIds[StGLStereoTexture::LEFT_TEXTURE]);

    // bind left texture to left FBO as color buffer
    StGLStereoTexture::bindLeft();
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,
                              StGLStereoTexture::stTextures[StGLStereoTexture::LEFT_TEXTURE].glTextureId, 0);
    // bind left render buffer to left FBO as depth buffer
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT,
                                 myGLFBufferIds[StGLStereoTexture::LEFT_TEXTURE]);
    if(glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT) {
        release();
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, StGLFrameBuffer::NO_RENDERBUFFER);
        return false;
    }
    unbindBufferLeft();
    StGLStereoTexture::unbindLeft();

    // create right RenderBuffer (will be used as depth buffer)
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, myGLDepthRBIds[StGLStereoTexture::RIGHT_TEXTURE]);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, textureSizeX, textureSizeY);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, myGLFBufferIds[StGLStereoTexture::RIGHT_TEXTURE]);

    // bind right texture to rights FBO as color buffer
    StGLStereoTexture::bindRight();
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,
                              StGLStereoTexture::stTextures[StGLStereoTexture::RIGHT_TEXTURE].glTextureId, 0);

    // bind right render buffer to right FBO as depth buffer
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT,
                                 myGLFBufferIds[StGLStereoTexture::RIGHT_TEXTURE]);
    if(glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT) {
        release();
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, StGLFrameBuffer::NO_RENDERBUFFER);
        return false;
    }
    unbindBufferRight();
    StGLStereoTexture::unbindRight();
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, StGLFrameBuffer::NO_RENDERBUFFER);

    ST_DEBUG_LOG("OpenGL, created StFrameBuffer(WxH= " + getSizeX() + 'x' + getSizeY() + ')');

    // create vertices buffers to draw simple textured quad
    StArray<StGLVec4> aQuad(4);
    aQuad[0] = StGLVec4( 1.0f, -1.0f, 0.0f, 1.0f); // top-right
    aQuad[1] = StGLVec4( 1.0f,  1.0f, 0.0f, 1.0f); // bottom-right
    aQuad[2] = StGLVec4(-1.0f, -1.0f, 0.0f, 1.0f); // top-left
    aQuad[3] = StGLVec4(-1.0f,  1.0f, 0.0f, 1.0f); // bottom-left
    myVerticesBuf.init(aQuad);

    StArray<StGLVec2> aQuadTC(4);
    aQuadTC[0] = StGLVec2(1.0f, 0.0f);
    aQuadTC[1] = StGLVec2(1.0f, 1.0f);
    aQuadTC[2] = StGLVec2(0.0f, 0.0f);
    aQuadTC[3] = StGLVec2(0.0f, 1.0f);
    myTexCoordBuf.init(aQuadTC);

    myViewPortX = textureSizeX;
    myViewPortY = textureSizeY;
    return true;
}

void StGLStereoFrameBuffer::setVPDimensions(GLsizei theSizeX, GLsizei theSizeY) {
    GLsizei aVPSizeX = stMin(theSizeX, getSizeX());
    GLsizei aVPSizeY = stMin(theSizeY, getSizeY());
    if(aVPSizeX != myViewPortX
    || aVPSizeY != myViewPortY) {
        GLfloat aDX = GLfloat(aVPSizeX) / GLfloat(getSizeX());
        GLfloat aDY = GLfloat(aVPSizeY) / GLfloat(getSizeY());
        StArray<StGLVec2> aTCoords(4);
        aTCoords[0] = StGLVec2(aDX,  0.0f);
        aTCoords[1] = StGLVec2(aDX,  aDY);
        aTCoords[2] = StGLVec2(0.0f, 0.0f);
        aTCoords[3] = StGLVec2(0.0f, aDY);
        myTexCoordBuf.init(aTCoords);
        myViewPortX = aVPSizeX;
        myViewPortY = aVPSizeY;
    }
}

void StGLStereoFrameBuffer::drawQuad(const StGLStereoFrameBuffer::StGLStereoProgram* theProgram) const {
    theProgram->use();
    myVerticesBuf.bindVertexAttrib(theProgram->getVVertexLoc());
    myTexCoordBuf.bindVertexAttrib(theProgram->getVTexCoordLoc());
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    myTexCoordBuf.unBindVertexAttrib(theProgram->getVTexCoordLoc());
    myVerticesBuf.unBindVertexAttrib(theProgram->getVVertexLoc());
    theProgram->unuse();
}
