/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGL/StGLTexture.h>
#include <StStrings/StLogger.h>
#include <StImage/StImagePlane.h>

bool StGLTexture::getInternalFormat(const StImagePlane& theData,
                                    GLint& theInternalFormat) {
    switch(theData.getFormat()) {
        case StImagePlane::ImgRGBAF:
        case StImagePlane::ImgBGRAF:
            theInternalFormat = GL_RGBA32F;
            return true;
        case StImagePlane::ImgRGBF:
        case StImagePlane::ImgBGRF:
            theInternalFormat = GL_RGB32F;
            return true;
        case StImagePlane::ImgRGBA:
        case StImagePlane::ImgBGRA:
            theInternalFormat = GL_RGBA8;
            return true;
        case StImagePlane::ImgRGB:
        case StImagePlane::ImgBGR:
        case StImagePlane::ImgRGB32:
        case StImagePlane::ImgBGR32:
            theInternalFormat = GL_RGB8;
            return true;
        case StImagePlane::ImgGrayF:
            //theInternalFormat = GL_R32F;    // OpenGL3+ hardware
            //theInternalFormat = GL_ALPHA32F_ARB;
            theInternalFormat = GL_ALPHA16; // backward compatibility
            return true;
        case StImagePlane::ImgGray:
            // this texture format is deprecated with OpenGL3+, use GL_R8 (GL_RED) instead
            //theInternalFormat = GL_R8;   // OpenGL3+ hardware
            theInternalFormat = GL_ALPHA8; // backward compatibility
            return true;
        default:
            return false;
    }
}

bool StGLTexture::getDataFormat(const StImagePlane& theData,
                                GLenum& thePixelFormat,
                                GLenum& theDataType) {
    thePixelFormat = GL_RGB;
    theDataType = GL_UNSIGNED_BYTE;
    switch(theData.getFormat()) {
        case StImagePlane::ImgGray: {
            // we fill ALPHA channel in the texture!
            //thePixelFormat = GL_RED;
            thePixelFormat = GL_ALPHA;
            theDataType = GL_UNSIGNED_BYTE;
            return true;
        }
        case StImagePlane::ImgRGB: {
            thePixelFormat = GL_RGB;
            theDataType = GL_UNSIGNED_BYTE;
            return true;
        }
        case StImagePlane::ImgBGR: {
            thePixelFormat = GL_BGR;
            theDataType = GL_UNSIGNED_BYTE;
            return true;
        }
        case StImagePlane::ImgRGBA:
        case StImagePlane::ImgRGB32: {
            thePixelFormat = GL_RGBA;
            theDataType = GL_UNSIGNED_BYTE;
            return true;
        }
        case StImagePlane::ImgBGRA:
        case StImagePlane::ImgBGR32: {
            thePixelFormat = GL_BGRA;
            theDataType = GL_UNSIGNED_BYTE;
            return true;
        }
        case StImagePlane::ImgGrayF: {
            //thePixelFormat = GL_RED;
            thePixelFormat = GL_ALPHA;
            theDataType = GL_FLOAT;
            return true;
        }
        case StImagePlane::ImgRGBF: {
            thePixelFormat = GL_RGB;
            theDataType = GL_FLOAT;
            return true;
        }
        case StImagePlane::ImgBGRF: {
            thePixelFormat = GL_BGR;
            theDataType = GL_FLOAT;
            return true;
        }
        case StImagePlane::ImgRGBAF: {
            thePixelFormat = GL_RGBA;
            theDataType = GL_FLOAT;
            return true;
        }
        case StImagePlane::ImgBGRAF: {
            thePixelFormat = GL_BGRA;
            theDataType = GL_FLOAT;
            return true;
        }
        default: return false;
    }
}

#ifdef __ST_DEBUG__
/**
 * Dummy function to display texture internal format.
 */
static StString formatInternalFormat(const GLint theInternalFormat) {
    switch(theInternalFormat) {
        // RED variations (GL_RED, OpenGL 3.0+)
        case GL_R8:       return "GL_R8";
        case GL_R16:      return "GL_R16";
        case GL_R16F:     return "GL_R16F"; // half-float
        case GL_R32F:     return "GL_R32F"; // float
        // RGB variations
        case GL_RGB4:     return "GL_RGB4";
        case GL_RGB5:     return "GL_RGB5";
        case GL_RGB8:     return "GL_RGB8";
        case GL_RGB10:    return "GL_RGB10";
        case GL_RGB12:    return "GL_RGB12";
        case GL_RGB16:    return "GL_RGB16";
        case GL_RGB16F:   return "GL_RGB16F"; // half-float
        case GL_RGB32F:   return "GL_RGB32F"; // float
        // RGBA variations
        case GL_RGBA8:    return "GL_RGBA8";
        case GL_RGB10_A2: return "GL_RGB10_A2";
        case GL_RGBA12:   return "GL_RGBA12";
        case GL_RGBA16:   return "GL_RGBA16";
        case GL_RGBA16F:  return "GL_RGBA16F"; // half-float
        case GL_RGBA32F:  return "GL_RGBA32F"; // float
        // ALPHA variations (deprecated)
        case GL_ALPHA8:   return "GL_ALPHA8";
        case GL_ALPHA16:  return "GL_ALPHA16";
        // unknown...
        default:          return StString("GL_? (") + theInternalFormat + ')';
    }
}
#endif

bool StGLTexture::init(GLsizei textureSizeX, GLsizei textureSizeY, GLenum inDataFormat, const GLubyte* pData) {
    // check texture size is fit dimension maximum
    GLint maxTexDim = 2048;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTexDim);
    if(textureSizeX < 32 || textureSizeY < 32) {
        ST_DEBUG_LOG("Texture size X or Y (" + textureSizeX + " x " + textureSizeY + ") lesser than minimum dimension (32)!");
        return false;
    } else if(textureSizeX > maxTexDim && textureSizeY > maxTexDim) {
        ST_DEBUG_LOG("Texture size X and Y (" + textureSizeX + " x " + textureSizeY + ") greater than maximum dimension (" + maxTexDim + ")!");
        return false;
    } else if(textureSizeX > maxTexDim) {
        ST_DEBUG_LOG("Texture size X (" + textureSizeX + ") greater than maximal dimension (" + maxTexDim + ")!");
        return false;
    } else if(textureSizeY > maxTexDim) {
        ST_DEBUG_LOG("Texture size Y (" + textureSizeY + ") greater than maximal dimension (" + maxTexDim + ")!");
        return false;
    }

    sizeX = textureSizeX;
    sizeY = textureSizeY;

    return create(inDataFormat, pData);
}

bool StGLTexture::initBlack(GLsizei textureSizeX, GLsizei textureSizeY) {
    size_t dataSize = 4 * textureSizeX * textureSizeY;
    GLubyte* pBlackData = new GLubyte[dataSize];
    stMemSet(pBlackData, 0, dataSize);
    bool result = init(textureSizeX, textureSizeY, GL_RGBA, pBlackData);
    delete[] pBlackData;
    return result;
}

bool StGLTexture::isProxySuccess() {
    // use proxy to check texture could be created or not
    glTexImage2D(GL_PROXY_TEXTURE_2D,
                 0,                // LOD number: 0 - base image level; n is the nth mipmap reduction image
                 textFormat,       // internal format
                 sizeX, sizeY, 0,  // border = {0,1}
                 GL_RGBA, GL_UNSIGNED_BYTE, NULL); // no mention (we check graphical RAM here)
    GLint testParamX = 0;
    GLint testParamY = 0;
    glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &testParamX);
    glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &testParamY);
    if(testParamX == 0 || testParamY == 0) {
        ST_DEBUG_LOG("Creation texture with size (" + sizeX + " x " + sizeY + ") FAILED!");
        return false;
    }
    return true;
}



bool StGLTexture::create(GLenum inDataFormat, const GLubyte* pData) {
    if(!isValid()) {
        glGenTextures(1, &glTextureId); // Create The Texture
    }
    bind();

    // texture interpolation parameters - could be overridden later
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, textureFilt);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, textureFilt);

    if(!isProxySuccess()) {
        release();
        return false;
    }

    glTexImage2D(GL_TEXTURE_2D,
                 0,                // LOD number: 0 - base image level; n is the nth mipmap reduction image
                 textFormat,       // texture internal format (driver will choose the nearest to requested thats supported by hardware)
                 sizeX, sizeY, 0,  // border = {0,1}
                 inDataFormat,     // format of the pixel data
                 GL_UNSIGNED_BYTE, // data type of the pixel data
                 pData);

    // detect which texture was actually created
    GLint resultFormat = 0;
    GLint resultSizeX  = 0;
    GLint resultSizeY  = 0;
    glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH,  &resultSizeX);
    glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &resultSizeY);
    glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_INTERNAL_FORMAT, &resultFormat);
#ifdef __ST_DEBUG_TEXTURES__
    ST_DEBUG_LOG("Created StGLTexture " + resultSizeX + " x "+ resultSizeY
               + " (format " + formatInternalFormat(resultFormat) + ')');
#endif

    unbind();
    return true;
}

void StGLTexture::setMinMagFilter(const GLenum newMinMagFilter) {
    if(!isValid()) {
        textureFilt = newMinMagFilter;
        return;
    } else if(textureFilt == newMinMagFilter) {
        return;
    }
    textureFilt = newMinMagFilter;
    bind();
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, textureFilt);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, textureFilt);
    unbind();
}

bool StGLTexture::init(const StImagePlane& theData) {
    if(theData.isNull()) {
        return false;
    }
    if(!initTrash((GLsizei )theData.getSizeX(), (GLsizei )theData.getSizeY())) {
        return false;
    }
    return fill(theData);
}

bool StGLTexture::fill(const StImagePlane& theData,
                       const GLsizei theRowFrom,
                       const GLsizei theRowTo,
                       const GLsizei theBatchRows) {
    if(theData.isNull() || !isValid()) {
        return false;
    }
    GLenum aPixelFormat, aDataType;
    if(!getDataFormat(theData, aPixelFormat, aDataType)) {
        return false;
    }

    GLsizei aRowTo = GLsizei(stMin(theData.getSizeY(), size_t(getSizeY())));
    if(theRowTo > 0) {
        aRowTo = stMin(theRowTo, aRowTo);
    }

    if(theRowFrom >= aRowTo) {
        // out of range
        return false;
    }

    bind();

    // setup the alignment
    size_t anExtraBytes = theData.getRowExtraBytes();
    size_t anAligment = stMin(theData.getMaxRowAligment(), size_t(8)); // limit to 8 bytes for OpenGL
    glPixelStorei(GL_UNPACK_ALIGNMENT, GLint(anAligment));

    if(theData.getSizeX() <= size_t(getSizeX()) && anExtraBytes < anAligment && theBatchRows > 1) {
        // do batch copy (more effective)
        GLsizei aPatchWidth = GLsizei(theData.getSizeX());
        GLsizei aBatchRows = theBatchRows;
        for(GLsizei aRow(theRowFrom), aRowsRemain(aRowTo); aRow < aRowTo; aRow += theBatchRows) {
            aRowsRemain = aRowTo - aRow;
            if(aRowsRemain < aBatchRows) {
                aBatchRows = aRowsRemain;
            }

            glTexSubImage2D(GL_TEXTURE_2D, 0, // 0 = LOD number
                            0, aRow,          // a texel offset in the (x, y) direction
                            aPatchWidth, aBatchRows,
                            aPixelFormat,     // format of the pixel data
                            aDataType,        // data type of the pixel data
                            theData.getData(aRow, 0));
        }
    } else {
        // copy row by row copy (the image plane greater than texture or batch copy is impossible)
        GLsizei aPatchWidth = stMin(GLsizei(theData.getSizeX()), getSizeX());
        for(GLsizei aRow = theRowFrom; aRow < aRowTo; ++aRow) {
            glTexSubImage2D(GL_TEXTURE_2D, 0, // 0 = LOD number
                            0, aRow,          // a texel offset in the (x, y) direction
                            aPatchWidth, 1,   // the (width, height) of the texture sub-image
                            aPixelFormat,     // format of the pixel data
                            aDataType,        // data type of the pixel data
                            theData.getData(aRow, 0));
        }
    }

    // turn back safe alignment...
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    unbind();
    return true;
}
