/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLStereo/StGLTextureQueue.h>

StGLTextureQueue::StGLTextureQueue(size_t queueSizeMax)
: stMutexPop(),
  dataFront(NULL),
  dataSnap(NULL),
  stMutexPush(),
  dataBack(NULL),
  stMutexSize(),
  queueSize(0),
  queueSizeMax(queueSizeMax),
  stQTexture(),
  stSwapFBMutex(),
  swapFBCount(0),
  stMeterMutex(),
  stFPSMeter(),
  stMutexSrcFromat(),
  currentSrcFormat(ST_V_SRC_MONO),
  ptsCurr(0.0),
  evNewSnapShot(false),
  isRightUpd(false),
  isInUpdTexture(false),
  isReadyToSwapFB(false),
  isSwapOnReady(false) {
    //
    ST_DEBUG_ASSERT(queueSizeMax >= 2);

    // we create 'empty' queue
    dataFront = new StGLTextureData();
    StGLTextureData* iter = dataFront;
    for(size_t i = 1; i < queueSizeMax; ++i) {
        iter->setNext(new StGLTextureData());
        iter = iter->getNext();
    }
    iter->setNext(dataFront); // data in loop
    dataBack = dataFront;
}

StGLTextureQueue::~StGLTextureQueue() {
    for(size_t i = 0; i < queueSizeMax; ++i) {
        StGLTextureData* remItem = dataFront;
        dataFront = dataFront->getNext();
        delete remItem;
    }
}

// this function called ONLY from image thread
bool StGLTextureQueue::push(const StImage& srcDataLeft,
                            const StImage& srcDataRight,
                            const StHandle<StStereoParams>& theStParams,
                            StFormatEnum srcFormat,
                            double srcPTS) {
    if(isFull()) {
        return false;
    }

    stMutexPush.lock();
    dataBack = isEmpty() ? dataFront : dataBack->getNext();

    dataBack->updateData(srcDataLeft,
                         srcDataRight,
                         theStParams,
                         srcFormat,
                         srcPTS);
    stMutexSrcFromat.lock();
        currentSrcFormat = dataBack->getSourceFormat();
    stMutexSrcFromat.unlock();

    stMutexSize.lock();
        ++queueSize;
    stMutexSize.unlock();
    stMutexPush.unlock();
    return true;
}

int StGLTextureQueue::swapFBOnReady() {
    if(isReadyToSwapFB) {
        stSwapFBMutex.lock();
        if(swapFBCount != 0) {
            isReadyToSwapFB = false;
            --swapFBCount;
            stSwapFBMutex.unlock();

            stQTexture.swapFB();

            stMeterMutex.lock();
                ++stFPSMeter;
            stMeterMutex.unlock();
            return SWAPONREADY_SWAPPED;
        } else {
            stSwapFBMutex.unlock();
            return SWAPONREADY_WAITLIM;
        }
    }
    return SWAPONREADY_NOTHING;
}

// this function called ONLY from plugin thread
bool StGLTextureQueue::stglUpdateStTextures() {
    getPlaybackFPS(); // debug output

    int swapState = swapFBOnReady();
    if(swapState == SWAPONREADY_WAITLIM) {
        return false;
    }

    if(!stMutexPop.tryLock()) {
        return swapState == SWAPONREADY_SWAPPED;
    }

    // do we already in update cycle?
    if(!isInUpdTexture) {
        // check event from video thread
        if(!isEmpty()) {
            isInUpdTexture = true;
        }
    } else if(isEmpty()) {
        // if we in texture update sequence - check queue not emptied!
        isInUpdTexture = false;
    }

    // still nothing to update? so return
    if(!isInUpdTexture) {
        stMutexPop.unlock();
        return swapState == SWAPONREADY_SWAPPED;
    }

    if(dataFront->fillTexture(stQTexture)) {
        isReadyToSwapFB = true;
        stMutexSize.lock();
            ptsCurr = dataFront->getPTS();
            dataSnap = dataFront; evNewSnapShot.set();
            dataFront = dataFront->getNext();
            ST_DEBUG_ASSERT(queueSize != 0); // critical error!
            --queueSize;
        stMutexSize.unlock();
        isInUpdTexture = false;
    }
    stMutexPop.unlock();

    // try early swap
    bool isAlreadySwapped = swapState == SWAPONREADY_SWAPPED;
    swapState = swapFBOnReady();
    return (swapState == SWAPONREADY_SWAPPED || isAlreadySwapped);
}

void StGLTextureQueue::clear() {
    stMutexPop.lock();
    stMutexPush.lock();
    stMutexSize.lock();
    stSwapFBMutex.lock();
        // decrease StStereoSource counters
        StGLTextureData* iterFront = dataFront;
        for(size_t i = 0; i < queueSize; ++i, iterFront = iterFront->getNext()) {
            iterFront->resetStParams();
        }
        // reset queue
        queueSize = 0;
        dataBack = dataFront;
        dataSnap = NULL;
        swapFBCount = 0;
        // empty texture update sequence
        isInUpdTexture = false;
    stSwapFBMutex.unlock();
    stMutexSize.unlock();
    stMutexPush.unlock();
    stMutexPop.unlock();
}

void StGLTextureQueue::drop(const size_t& count) {
    stMutexPop.lock();
    stMutexPush.lock();
    stMutexSize.lock();
        if(queueSize < 2) {
            // to small queue
            stMutexSize.unlock();
            stMutexPush.unlock();
            stMutexPop.unlock();
            return;
        }
        size_t decr = (count < queueSize) ? count : (queueSize - 1);

        // decrease StStereoSource counters
        for(size_t i = 0; i < decr; ++i, dataFront = dataFront->getNext()) {
            dataFront->resetStParams();
        }
        // reset queue
        queueSize -= decr;
        // empty texture update sequence
        isInUpdTexture = false;
    stMutexSize.unlock();
    stMutexPush.unlock();
    stMutexPop.unlock();
}

int StGLTextureQueue::getSnapshot(StImage* outDataLeft, StImage* outDataRight, bool isForce) {
    if(!evNewSnapShot.check() && !isForce) {
        return SNAPSHOT_NO_NEW;
    }
    stMutexPop.lock();
    if(dataSnap == NULL) {
        stMutexPop.unlock();
        return SNAPSHOT_NO_NEW;
    }
    dataSnap->getCopy(outDataLeft, outDataRight);
    evNewSnapShot.reset();
    stMutexPop.unlock();
    return SNAPSHOT_SUCCESS;
}
