/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLFrameBuffer_h_
#define __StGLFrameBuffer_h_

#include <StGL/StGLTexture.h>

/**
 * Simple class represents Virtual (texture) stereo Frame buffer object.
 * This allow render to texture.
 */
class ST_LOCAL StGLFrameBuffer : public StGLTexture {

        private:

    GLuint myGLFBufferId; //!< FrameBuffer object
    GLuint myGLDepthRBId; //!< RenderBuffer object for depth ID
    GLsizei  myViewPortX; //!< FBO viewport width  <= texture width
    GLsizei  myViewPortY; //!< FBO viewport height <= texture height

        private:

    static bool isRectangularFBSupportedInit() {
        // It seems that all many problematic cards / drivers
        // (FrameBuffer with non-power-of-two size corrupted / slowdown)
        // are GLSL 1.1 limited (and OpenGL2.0).
        // SURE - this is NOT rootcouse of the problem, but just some detection mechanism
        StString strGLSL_old("1.10");
        // latest drivers for GeForce FX return GLSL1.2 vc OpenGL2.1 support level
        // but slowdown rendering - so we detect them here.
        StString strGeForceFX("GeForce FX");

        StString strGLSLVersion((const char* )glGetString(GL_SHADING_LANGUAGE_VERSION));
        StString strGlRenderer((const char* )glGetString(GL_RENDERER));
        return !strGLSLVersion.isContains(strGLSL_old) && !strGlRenderer.isContains(strGeForceFX);
    }

    /**
     * Validate FrameBuffer id.
     */
    bool isValidFrameBuffer() const {
        return myGLFBufferId != NO_FRAMEBUFFER;
    }

    /**
     * Validate RenderBuffer id.
     */
    bool isValidDepthBuffer() const {
        return myGLDepthRBId != NO_RENDERBUFFER;
    }

        public:

    static const GLuint NO_FRAMEBUFFER  = 0;
    static const GLuint NO_RENDERBUFFER = 0;

    /**
     * Empty constructor.
     */
    StGLFrameBuffer(GLint theTextureFormat = GL_RGBA8)
    : StGLTexture(theTextureFormat),
      myGLFBufferId(NO_FRAMEBUFFER),
      myGLDepthRBId(NO_RENDERBUFFER),
      myViewPortX(0),
      myViewPortY(0) {
        //
    }

    /**
     * OpenGL objects will be automatically released.
     * Make sure it is called from thread with bound GL context
     * or release the object in advance.
     */
    virtual ~StGLFrameBuffer() {
        release();
    }

    /**
     * Returns true if FBO was initialized.
     */
    bool isValid() const {
        return isValidFrameBuffer() && StGLTexture::isValid();// && isValidDepthBuffer();
    }

    /**
     * Initialize the FBO with specified dimensions.
     */
    bool init(GLsizei textureSizeX, GLsizei textureSizeY) {
        // release current objects
        release();

        // create the texture
        if(!StGLTexture::initTrash(textureSizeX, textureSizeY)) {
            release();
            return false;
        }

        // create RenderBuffer (will be used as depth buffer)
        glGenRenderbuffersEXT(1, &myGLDepthRBId);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, myGLDepthRBId);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, textureSizeX, textureSizeY);

        // build FBO and setup it as texture
        glGenFramebuffersEXT(1, &myGLFBufferId);
        bindBuffer();
        // bind texture to the FBO as color buffer
        StGLTexture::bind();
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,
                                  StGLTexture::glTextureId, 0);
        // bind render buffer to the FBO as depth buffer
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT,
                                     myGLFBufferId);
        if(glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT) {
            release();
            glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, NO_RENDERBUFFER);
            return false;
        }
        StGLTexture::unbind();
        unbindBuffer();
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, NO_RENDERBUFFER);

        myViewPortX = textureSizeX;
        myViewPortY = textureSizeY;
        return true;
    }

    /**
     * Release OpenGL objects related to this FBO.
     */
    void release() {
        StGLTexture::release();
        if(isValidDepthBuffer()) {
            glDeleteRenderbuffersEXT(1, &myGLDepthRBId);
            myGLDepthRBId = NO_RENDERBUFFER;
        }
        if(isValidFrameBuffer()) {
            glDeleteFramebuffersEXT(1, &myGLFBufferId);
            myGLFBufferId = NO_FRAMEBUFFER;
        }
    }

    /**
     * FBO viewport width.
     */
    GLsizei getVPSizeX() const {
        return myViewPortX;
    }

    /**
     * FBO viewport height.
     */
    GLsizei getVPSizeY() const {
        return myViewPortY;
    }

    /**
     * Set new FBO viewport width. Should be <= texture width.
     */
    void setVPSizeX(GLsizei theSizeX) {
        myViewPortX = theSizeX;
    }

    /**
     * Set new FBO viewport height. Should be <= texture height.
     */
    void setVPSizeY(GLsizei theSizeY) {
        myViewPortY = theSizeY;
    }

    /**
     * Setup OpenGL viewport equal to FBO dimensions
     */
    void setupViewPort() {
        glViewport(0, 0, myViewPortX, myViewPortY);
    }

    /**
     * Bind frame buffer (to render into the texture).
     */
    void bindBuffer() {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, myGLFBufferId);
    }

    /**
     * Unbind frame buffer.
     */
    void unbindBuffer() {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, NO_FRAMEBUFFER);
    }

    static void unbindBufferGlobal() {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, NO_FRAMEBUFFER);
    }

    /**
     * Bind color texture (to render the texture).
     */
    void bindTexture(GLenum textureUnit = GL_TEXTURE0) {
        StGLTexture::bind(textureUnit);
    }

    /**
     * Unbind color texture.
     */
    void unbindTexture() {
        StGLTexture::unbind();
    }

    /**
     * Method to detect obsolete hardware.
     */
    static bool isRectangularFBSupported() {
        static const bool isSupported = isRectangularFBSupportedInit();
        return isSupported;
    }

    /**
     * Upscale input parameters to power of two values.
     */
    static void convertToPowerOfTwo(GLsizei& frSizeX, GLsizei& frSizeY) {
        // compute optimal values...
        GLint maxTexDim = 2048; glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTexDim);
        GLsizei frSizeXGreater = getPowerOfTwo(frSizeX, GLsizei(maxTexDim));
        GLsizei frSizeYGreater = getPowerOfTwo(frSizeY, GLsizei(maxTexDim));
        GLsizei frSizeXLess = frSizeXGreater / 2;
        GLsizei frSizeYLess = frSizeYGreater / 2;
        frSizeX = (frSizeXGreater + frSizeXLess - 2 * frSizeX) >= 0 ? frSizeXLess : frSizeXGreater;
        frSizeY = (frSizeYGreater + frSizeYLess - 2 * frSizeY) >= 0 ? frSizeYLess : frSizeYGreater;
    }

};

#endif //__StGLFrameBuffer_h_
