/**
 * Copyright © 2009-2012 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLShader_h_
#define __StGLShader_h_

#include <gl/glew.h>

#include <StStrings/StLogger.h>
#include <StFile/StRawFile.h>

/**
 * Virtual class represents GLSL shader.
 */
class ST_LOCAL StGLShader {

        public:

    static const GLuint NO_SHADER = 0;

        public:

    /**
     * Implementations should define this method to return:
     * GL_VERTEX_SHADER, GL_FRAGMENT_SHADER or GL_GEOMETRY_SHADER.
     */
    virtual GLenum getType() const = 0;

    virtual StString getTypeString() const {
        switch(getType()) {
            case GL_VERTEX_SHADER:   return StString("Vertex Shader");
            case GL_FRAGMENT_SHADER: return StString("Fragment Shader");
            ///case GL_GEOMETRY_SHADER: return StString("Geometry Shader");
            default:                 return StString("Unknown Shader");
        }
    }

    /**
     * @return true if shader object is valid.
     */
    bool isValid() const {
        return myShaderId != NO_SHADER;
    }

    /**
     * @return user-specified title for this shader.
     */
    virtual const StString& getTitle() const {
        return myTitle;
    }

    /**
     * Vitual method could be overridden by classes contained
     * shader program text internally.
     */
    virtual bool init() {
        return false;
    }

    virtual ~StGLShader() {
        release();
    }

    /**
     * Delete shader object and invalidate its id.
     * You should detach this shader from all programs before deleting!
     */
    void release() {
        if(isValid()) {
            glDeleteShader(myShaderId);
            myShaderId = NO_SHADER;
        }
    }

    /**
     * Initialize the shader program from text buffer.
     * @param theSrcLines0 (const char* ) - shader program source;
     * @return true on success.
     */
    virtual bool init(const char* theSrcLines0,
                      const char* theSrcLines1 = NULL,
                      const char* theSrcLines2 = NULL) {
        if(theSrcLines0 == NULL) {
            return false;
        }
        GLsizei aCount = (theSrcLines1 == NULL) ? 1 : ((theSrcLines2 == NULL) ? 2 : 3);
        const char* aSrcArray[3] = {theSrcLines0, theSrcLines1, theSrcLines2};

        myShaderId = glCreateShader(getType());
        glShaderSource(myShaderId, aCount, aSrcArray, NULL);

        // compile shaders
        glCompileShader(myShaderId);

        StString compileInfo = getCompileInfo();
    #ifndef __ST_DEBUG_SHADERS__
        if(!isCompiled()) {
    #endif
        ST_DEBUG_LOG("Compiling " + getTypeString()
                     + " '" + myTitle + "' "
                     + (isCompiled() ? StString("done") : StString("FAILED"))
                     + (!compileInfo.isEmpty() ? (StString(". Log:\n") + compileInfo) : (StString())));
    #ifndef __ST_DEBUG_SHADERS__
        }
    #endif

        // check compile success
        if(!isCompiled()) {
            release();
            return false;
        }
        return true;
    }

    /**virtual bool init(const StString& sourceUtfString) {
        std::string sourceString = sourceUtfString.ansiText();
        return init(sourceString.c_str());
    }*/

    /**
     * Initialize the shader from text file.
     * @param theFileName (const StString& ) - file to read;
     * @return true on success.
     */
    bool initFile(const StString& theFileName) {
        StRawFile aTextFile(theFileName);
        if(!aTextFile.readFile()) {
            ST_DEBUG_LOG("Shader file '" + theFileName + "' can not be read");
            return false;
        }
        return init((const char* )aTextFile.getBuffer());
    }

        protected:

    StGLShader(const StString& theTitle)
    : myTitle(theTitle),
      myShaderId(NO_SHADER) {
        //
    }

    /**
     * Check compilation state.
     * @return true if shader object compiled successfully.
     */
    bool isCompiled() const {
        GLint isSuccess = GL_FALSE;
        glGetShaderiv(myShaderId, GL_COMPILE_STATUS, &isSuccess);
        return isSuccess == GL_TRUE;
    }

    /**
     * Returns compilation information, provided by driver.
     * This string is driver's and vendor's specific.
     * Usually it contains compilation errors and warnings.
     */
    StString getCompileInfo() const {
        GLint anInfoLen = 0;
        glGetShaderiv(myShaderId, GL_INFO_LOG_LENGTH, &anInfoLen);
        if(anInfoLen > 0) {
            GLchar* anInfoStr = new GLchar[anInfoLen];
            GLsizei aCharsWritten = 0;
            glGetShaderInfoLog(myShaderId, anInfoLen, &aCharsWritten, anInfoStr);
            StString aCompileInfo(anInfoStr);
            delete[] anInfoStr;
            return aCompileInfo;
        }
        return StString();
    }

        protected:

    StString myTitle;    //!< just shader title
    GLuint   myShaderId; //!< OpenGL shader ID

        private:

    friend class StGLProgram;

};

/**
 * Class represents GLSL Vertex Shader.
 */
class ST_LOCAL StGLVertexShader : public StGLShader {

        public:

    StGLVertexShader(const StString& theTitle)
    : StGLShader(theTitle) {
        //
    }

    virtual ~StGLVertexShader() {
        //
    }

    virtual GLenum getType() const {
        return GL_VERTEX_SHADER;
    }

};

/**
 * Class represents GLSL Fragment Shader.
 */
class ST_LOCAL StGLFragmentShader : public StGLShader {

        public:

    StGLFragmentShader(const StString& theTitle)
    : StGLShader(theTitle) {
        //
    }

    virtual ~StGLFragmentShader() {
        //
    }

    virtual GLenum getType() const {
        return GL_FRAGMENT_SHADER;
    }

};

#endif //__StGLShader_h_
