/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLTexture_h_
#define __StGLTexture_h_

#include <gl/glew.h>

#include <stTypes.h>

class StImagePlane;

/**
 * This class represent OpenGL texture.
 */
class ST_LOCAL StGLTexture {

        private:

    friend class StGLStereoTexture;
    friend class StGLStereoFrameBuffer;

        protected:

    GLsizei      sizeX; // texture width
    GLsizei      sizeY; // texture height
    GLint   textFormat; // GL_RGB, GL_RGBA,...
    GLuint glTextureId; // GL texture ID
    GLenum textureUnit; // texture unit
    GLenum textureFilt; // current texture filter

    /**
     * Help function for GL texture creation.
     * @param inDataFormat (const GLenum& ) - data format;
     * @param pData (const GLubyte* ) - initialization data (INPUT);
     * @return true on success.
     */
    bool create(GLenum inDataFormat, const GLubyte* pData);
    bool isProxySuccess();

        public:

    static const GLuint NO_TEXTURE = 0;

    /**
     * Function setup the best internal texture format to fit the image data.
     * Currently GL_ALPHA is returned for grayscale image data.
     * @return true if internal format was found.
     */
    static bool getInternalFormat(const StImagePlane& theData,
                                  GLint& theInternalFormat);

    /**
     * Function convert StImagePlane format into OpenGL data format.
     * @return true if format supported.
     */
    static bool getDataFormat(const StImagePlane& theData,
                              GLenum& thePixelFormat,
                              GLenum& theDataType);

    StGLTexture(GLint textureFormat = GL_RGBA8)
    : sizeX(0),
      sizeY(0),
      textFormat(textureFormat),
      glTextureId(NO_TEXTURE),
      textureUnit(GL_TEXTURE0),
      textureFilt(GL_LINEAR) {
        //
    }

    virtual ~StGLTexture() {
        release();
    }

    bool isValid() const {
        return glTextureId != NO_TEXTURE;
    }

    void release() {
        if(isValid()) {
            glDeleteTextures(1, &glTextureId);
            glTextureId = NO_TEXTURE;
        }
        sizeX = sizeY = 0;
    }

    /**
     * Get texture format.
     */
    GLint getTextureFormat() {
        return textFormat;
    }

    /**
     * Set texture format. Texture should be re-initialized after this change.
     */
    void setTextureFormat(GLint textureFormat) {
        this->textFormat = textureFormat;
    }

    /**
     * @param textureSizeX (GLsizei ) - texture width;
     * @param textureSizeY (GLsizei ) - texture height;
     * @param inDataFormat (GLenum );
     * @param data (const GLubyte* );
     * @return true on success.
     */
    bool init(GLsizei textureSizeX, GLsizei textureSizeY, GLenum inDataFormat, const GLubyte* pData);
    bool init(const StImagePlane& theData);
    bool initBlack(GLsizei textureSizeX, GLsizei textureSizeY);

    /**
     * In GL version 1.1 or greater, pData may be a NULL pointer.
     * However data of the texture will be undefined!
     */
    bool initTrash(GLsizei textureSizeX, GLsizei textureSizeY) {
        return init(textureSizeX, textureSizeY, GL_RGBA, NULL);
    }

    /**
     * Bind the texture to specified unit.
     */
    void bind(GLenum textureUnit = GL_TEXTURE0) {
        this->textureUnit = textureUnit;
        glActiveTexture(textureUnit);
        glBindTexture(GL_TEXTURE_2D, glTextureId);
    }

    void unbind() const {
        glActiveTexture(textureUnit);
        glBindTexture(GL_TEXTURE_2D, NO_TEXTURE);
    }

    static void unbindGlobal(GLenum textureUnit = GL_TEXTURE0) {
        glActiveTexture(textureUnit);
        glBindTexture(GL_TEXTURE_2D, NO_TEXTURE);
    }

    /**
     * @return sizeX (GLsizei ) - texture width.
     */
    GLsizei getSizeX() const {
        return sizeX;
    }

    /**
     * @return sizeY (GLsizei ) - texture height.
     */
    GLsizei getSizeY() const {
        return sizeY;
    }

    /**
     * Change Min and Mag filter.
     * After this call current texture will be undefined.
     */
    void setMinMagFilter(const GLenum newMinMagFilter);

    /**
     * Fill the texture with the image plane.
     * @param theData (const StImagePlane& ) - the image plane to copy data from;
     * @param theRowFrom (const GLsizei ) - fill data from row (for both - input image plane and the texture!);
     * @param theRowTo (const GLsizei ) - fill data up to the row (if zero - all rows);
     * @param theBatchRows (const GLsizei ) - maximal step for GL function call (greater - more effective);
     * @return true on success.
     */
    bool fill(const StImagePlane& theData,
              const GLsizei theRowFrom = 0,
              const GLsizei theRowTo = 0,
              const GLsizei theBatchRows = 128);

    /**
     * @return GL texture ID.
     */
    GLuint getTextureId() const {
        return glTextureId;
    }

};

#endif //__StGLTexture_h_
