/**
 * Copyright © 2010-2012 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLVertexBuffer_h_
#define __StGLVertexBuffer_h_

#include <gl/glew.h>

#include <StStrings/StString.h>
#include <StGL/StGLVarLocation.h>
#include <StGL/StGLVec.h>

/**
 * Enumeration indicates the data stored in VBO.
 */
typedef enum tagStGLVBOType {
    ST_VBO_VERTEX, //!< vertices
    ST_VBO_NORMAL, //!< normals
    ST_VBO_TCOORD, //!< texture coordinates
    ST_VBO_COLORS, //!< colors
    ST_VBO_INDEX,  //!< indices
} StGLVBOType;

class ST_LOCAL StGLVertexBuffer {

        public:

    StGLVertexBuffer()
    : myBufferId(0),
      myElemSize(4),
      myElemsCount(0),
      myDataType(GL_FLOAT) {}

    virtual ~StGLVertexBuffer() {
        release();
    }

    bool operator==(const StGLVertexBuffer& compare) const {
        return myBufferId == compare.myBufferId &&
               getTarget() == compare.getTarget();
    }

    bool operator!=(const StGLVertexBuffer& compare) const {
        return myBufferId != compare.myBufferId ||
               getTarget() != compare.getTarget();
    }

    bool operator>(const StGLVertexBuffer& compare) const {
        return myBufferId > compare.myBufferId;
    }

    bool operator<(const StGLVertexBuffer& compare) const {
        return myBufferId < compare.myBufferId;
    }

    bool operator>=(const StGLVertexBuffer& compare) const {
        return myBufferId >= compare.myBufferId;
    }

    bool operator<=(const StGLVertexBuffer& compare) const {
        return myBufferId <= compare.myBufferId;
    }

    bool isValid() const {
        return myBufferId != 0;
    }

    bool init() {
        if(!isValid()) {
            glGenBuffers(1, &myBufferId);
        }
        return isValid();
    }

    void release() {
        if(isValid()) {
            glDeleteBuffers(1, &myBufferId);
            myBufferId = 0;
        }
    }

    void bind() const {
        glBindBuffer(getTarget(), myBufferId);
    }

    void unbind() const {
        glBindBuffer(getTarget(), 0);
    }

    /**
     * VBO should be binded before call.
     * @param theElemSize   (GLsizeiptr ) - specifies the number of components per generic vertex attribute. Must be 1, 2, 3, or 4;
     * @param theElemsCount (GLsizeiptr ) - elements count;
     * @param theData (const GLfloat* ) - data pointer.
     */
    void setData(GLsizeiptr     theElemSize,
                 GLsizeiptr     theElemsCount,
                 const GLfloat* theData) {
        myElemSize   = theElemSize;
        myElemsCount = theElemsCount;
        glBufferData(getTarget(), myElemsCount * myElemSize * sizeof(GLfloat), theData, GL_STATIC_DRAW);
        myDataType = GL_FLOAT;
    }

    void setData(GLsizeiptr    theElemSize,
                 GLsizeiptr    theElemsCount,
                 const GLuint* theData) {
        myElemSize   = theElemSize;
        myElemsCount = theElemsCount;
        glBufferData(getTarget(), myElemsCount * myElemSize * sizeof(GLuint), theData, GL_STATIC_DRAW);
        myDataType = GL_UNSIGNED_INT;
    }

    void setData(GLsizeiptr     theElemSize,
                 GLsizeiptr     theElemsCount,
                 const GLubyte* theData) {
        myElemSize   = theElemSize;
        myElemsCount = theElemsCount;
        glBufferData(getTarget(), myElemsCount * myElemSize * sizeof(GLubyte), theData, GL_STATIC_DRAW);
        myDataType = GL_UNSIGNED_BYTE;
    }

    bool init(const StArray<StGLVec2>& theArray) {
        return init(2, GLsizeiptr(theArray.size()), theArray.getFirst().getData());
    }

    bool init(const StArray<StGLVec3>& theArray) {
        return init(3, GLsizeiptr(theArray.size()), theArray.getFirst().getData());
    }

    bool init(const StArray<StGLVec4>& theArray) {
        return init(4, GLsizeiptr(theArray.size()), theArray.getFirst().getData());
    }

    bool init(GLsizeiptr     theElemSize,
              GLsizeiptr     theElemsCount,
              const GLfloat* theData) {
        if(!init()) {
            return false;
        }
        bind();
        setData(theElemSize, theElemsCount, theData);
        unbind();
        return true;
    }

    bool init(GLsizeiptr    theElemSize,
              GLsizeiptr    theElemsCount,
              const GLuint* theData) {
        if(!init()) {
            return false;
        }
        bind();
        setData(theElemSize, theElemsCount, theData);
        unbind();
        return true;
    }

    bool init(GLsizeiptr     theElemSize,
              GLsizeiptr     theElemsCount,
              const GLubyte* theData) {
        if(!init()) {
            return false;
        }
        bind();
        setData(theElemSize, theElemsCount, theData);
        unbind();
        return true;
    }

    /**
     * @return elemSize (GLsizeiptr ) - specifies the number of components per generic vertex attribute. Must be 1, 2, 3, or 4;
     */
    GLsizeiptr getElemSize() const {
        return myElemSize;
    }

    GLsizeiptr getElemsCount() const {
        return myElemsCount;
    }

    /**
     * @return the data type of each component in the array.
     */
    GLenum getDataType() const {
        return myDataType;
    }

    virtual GLenum getTarget() const {
        return GL_ARRAY_BUFFER;
    }

    void bindVertexAttrib(StGLVarLocation attribLoc) const {
        if(isValid() && attribLoc.isValid()) {
            bind();
            glEnableVertexAttribArray(attribLoc);
            glVertexAttribPointer(attribLoc, GLint(getElemSize()), getDataType(), GL_FALSE, 0, NULL);
        }
    }

    void unBindVertexAttrib(StGLVarLocation attribLoc) const {
        if(isValid() && attribLoc.isValid()) {
            glDisableVertexAttribArray(attribLoc);
            unbind();
        }
    }

    StString toString() const {
        return StString("OpenGL buffer #") + myBufferId;
    }

        private:

    GLuint     myBufferId;
    GLsizeiptr myElemSize;
    GLsizeiptr myElemsCount;
    GLenum     myDataType;

};

template<> inline void StArray<GLuint>::sort() {}

class ST_LOCAL StGLIndexBuffer : public StGLVertexBuffer {

        public:

    StGLIndexBuffer()
    : StGLVertexBuffer() {}

    virtual ~StGLIndexBuffer() {
        //
    }

    virtual GLenum getTarget() const {
        return GL_ELEMENT_ARRAY_BUFFER;
    }

    bool init(const StArray<GLuint>& theArray) {
        return StGLVertexBuffer::init(GLsizeiptr(1), GLsizeiptr(theArray.size()), &theArray.getFirst());
    }

};

#endif //__StGLVertexBuffer_h_
