/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLQuadTexture_h_
#define __StGLQuadTexture_h_

#include "StGLStereoTexture.h"
#include <StImage/StImage.h>
#include <StGL/StParams.h>

/**
 * Frame texture extend GL texture with several attributes:
 *  - data size (represents data position and size in the texture);
 *  - data DAR (Display Aspect Ratio).
 */
class ST_LOCAL StGLFrameTexture : public StGLTexture {

        private:

    StGLVec2    dataSize; // data size in the texture (x()=right and y()=bottom)
    GLfloat displayRatio; // display aspect ratio

        public:

    StGLFrameTexture()
    : StGLTexture(GL_RGB8),
      dataSize(1.0f, 1.0f),
      displayRatio(1.0f) {
        //
    }

    virtual ~StGLFrameTexture() {
        //
    }

    /**
     * @return dataSize (const StGLVec2& ) - data size in the texture.
     */
    const StGLVec2& getDataSize() {
        return dataSize;
    }

    /**
     * @param dataSize (StGLVec2& ) - setted data size in the texture.
     */
    void setDataSize(const StGLVec2& dataSize) {
        this->dataSize = dataSize;
    }

    /**
     * @return ratio (GLfloat ) - display aspect ratio.
     */
    GLfloat getDisplayRatio() const {
        return displayRatio;
    }

    /**
     * @param ratio (GLfloat ) - setted display aspect ratio.
     */
    void setDisplayRatio(GLfloat value) {
        displayRatio = value;
    }

};

class ST_LOCAL StGLFrameTextures {

        private:

    StHandle<StStereoParams> myParams; // source pointer
    StGLFrameTexture    stTextures[4]; // texture planes
    StImage::ImgColorModel      imgCM; // color model

        public:

    StGLFrameTextures()
    : myParams(),
      imgCM(StImage::ImgColor_RGB) {
        //
    }

    virtual ~StGLFrameTextures() {
        //
    }

    StHandle<StStereoParams> getSource() {
        return myParams;
    }

    void setSource(const StHandle<StStereoParams>& theParams) {
        myParams = theParams;
    }

    StImage::ImgColorModel getColorModel() const {
        return imgCM;
    }

    void setColorModel(const StImage::ImgColorModel theColorModel) {
        imgCM = theColorModel;
    }

    /**
     * Release OpenGL texture objects.
     */
    void release() {
        stTextures[0].release();
        stTextures[1].release();
        stTextures[2].release();
        stTextures[3].release();
    }

    /**
     * @return true if main data is valid.
     */
    bool isValid() const {
        return stTextures[0].isValid();
    }

    /**
     * @return sizeX (GLsizei ) - main texture width.
     */
    GLsizei getSizeX() const {
        return stTextures[0].getSizeX();
    }

    /**
     * @return sizeY (GLsizei ) - main texture height.
     */
    GLsizei getSizeY() const {
        return stTextures[0].getSizeY();
    }

    /**
     * Access to texture plane using id.
     */
    StGLFrameTexture& getPlane(size_t thePlaneId = 0) {
        ST_DEBUG_ASSERT(thePlaneId < 4);
        return stTextures[thePlaneId];
    }

    /**
     * Bind as multitexture.
     */
    void bind(GLenum textureUnit = GL_TEXTURE0) {
        if(stTextures[0].isValid()) { stTextures[0].bind(textureUnit);     }
        if(stTextures[1].isValid()) { stTextures[1].bind(textureUnit + 1); }
        if(stTextures[2].isValid()) { stTextures[2].bind(textureUnit + 2); }
        if(stTextures[3].isValid()) { stTextures[3].bind(textureUnit + 3); }
    }

    /**
     * Unbind textures.
     */
    void unbind() {
        if(stTextures[3].isValid()) { stTextures[3].unbind(); }
        if(stTextures[2].isValid()) { stTextures[2].unbind(); }
        if(stTextures[1].isValid()) { stTextures[1].unbind(); }
        if(stTextures[0].isValid()) { stTextures[0].unbind(); }
    }

    void increaseSize(StGLFrameTexture& stTexture,
                      GLsizei textureSizeX, GLsizei textureSizeY) {
        // test existing size / new size
        /// TODO (Kirill Gavrilov#8) we can automatically reduce texture size here
        if((stTexture.getSizeX() < textureSizeX) ||
           (stTexture.getSizeY() < textureSizeY) ||
           !stTexture.isValid()) {
            ST_DEBUG_LOG("Requested texture size (" + textureSizeX + 'x' + textureSizeY
                       + ") larger than current texture size(" + stTexture.getSizeX() + 'x' + stTexture.getSizeY() + ')');
            GLsizei originalSizeX = stTexture.getSizeX();
            GLsizei originalSizeY = stTexture.getSizeY();

            GLint maxTexDim = 2048; glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTexDim);
            GLsizei newSizeX = getPowerOfTwo(textureSizeX, GLsizei(maxTexDim));
            GLsizei newSizeY = getPowerOfTwo(textureSizeY, GLsizei(maxTexDim));
            if((newSizeY != originalSizeY) || (newSizeX != originalSizeX)) {
                if(!stTexture.initTrash(newSizeX, newSizeY)) {
                    stTexture.initTrash((originalSizeX > 0) ? originalSizeX : 512,
                                        (originalSizeY > 0) ? originalSizeY : 512);
                    ST_DEBUG_LOG("FAILED to Increase the texture size to (" + newSizeX + 'x' +  newSizeY + ")!");
                } else {
                    ST_DEBUG_LOG("Increase the texture size to (" + newSizeX + 'x' +  newSizeY + ") success!");
                }
            } else {
                ST_DEBUG_LOG("Not possible to Increase the texture size!");
            }
        }
    }

    void preparePlane(size_t  thePlaneId,
                      GLsizei theSizeX, GLsizei theizeY,
                      GLint   theInternalFormat) {

        if(stTextures[thePlaneId].getTextureFormat() != theInternalFormat) {
            // wrong texture format
            stTextures[thePlaneId].release();
            stTextures[thePlaneId].setTextureFormat(theInternalFormat);
        }
        increaseSize(stTextures[thePlaneId], theSizeX, theizeY);
    }

    /**
     * Change Min and Mag filter.
     * After this call current bound texture will be undefined.
     */
    void setMinMagFilter(const GLenum newMinMagFilter) {
        stTextures[0].setMinMagFilter(newMinMagFilter);
        /// TODO (Kirill Gavrilov#4) investigate
        stTextures[1].setMinMagFilter(GL_LINEAR);
        stTextures[2].setMinMagFilter(GL_LINEAR);
        stTextures[3].setMinMagFilter(GL_LINEAR);
    }

};

/**
 * This class represent float quad texture buffer. Consists of 4 lists of textures:
 *  - Front Left and Front Right (currently displayed);
 *  - Back Left and Back Right (prepared);
 * Float means each list is independent.
 */
class ST_LOCAL StGLQuadTexture {

        public:

    typedef enum tagFrontOrBack {
        FRONT_TEXTURE = 0,
        BACK_TEXTURE  = 2,
    } FrontOrBack;

    typedef enum tagLeftOrRight {
        LEFT_TEXTURE  = 0,
        RIGHT_TEXTURE = 1,
    } LeftOrRight;

        private:

    StGLFrameTextures stTextures[4]; // Front and Back stereo textures
    bool                     active; // Front/Back flag

    inline size_t getFrontId() const {
        return active ? FRONT_TEXTURE : BACK_TEXTURE;
    }

    inline size_t getBackId() const {
        return active ? BACK_TEXTURE : FRONT_TEXTURE;
    }

        public:

    StGLQuadTexture()
    : active(true) {
        //
    }

    virtual ~StGLQuadTexture() {
        //
    }

    /**
     * @return texture (StGLFrameTextures& ) - texture from front pair (for rendering).
     */
    StGLFrameTextures& getFront(const LeftOrRight leftOrRight) {
        return stTextures[getFrontId() + leftOrRight];
    }

    /**
     * @return texture (StGLFrameTextures& ) - texture from back pair (for filling).
     */
    StGLFrameTextures& getBack(const LeftOrRight leftOrRight) {
        return stTextures[getBackId() + leftOrRight];
    }

    /**
     * Swap FRONT / BACK.
     */
    void swapFB() {
        active = !active; // process swap itself
    }

    /**
     * Change Min and Mag filter.
     * After this call current bound texture will be undefined.
     */
    void setMinMagFilter(const GLenum newMinMagFilter) {
        stTextures[FRONT_TEXTURE +  LEFT_TEXTURE].setMinMagFilter(newMinMagFilter);
        stTextures[FRONT_TEXTURE + RIGHT_TEXTURE].setMinMagFilter(newMinMagFilter);
        stTextures[BACK_TEXTURE  +  LEFT_TEXTURE].setMinMagFilter(newMinMagFilter);
        stTextures[BACK_TEXTURE  + RIGHT_TEXTURE].setMinMagFilter(newMinMagFilter);
    }

    /**
     * Release OpenGL texture objects.
     */
    void release() {
        stTextures[0].release();
        stTextures[1].release();
        stTextures[2].release();
        stTextures[3].release();
    }

};

#endif //__StGLQuadTexture_h_
