/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLStereoFrameBuffer_h_
#define __StGLStereoFrameBuffer_h_

#include <StGL/StGLVertexBuffer.h>
#include <StGL/StGLProgram.h>
#include <StGL/StGLFrameBuffer.h>
#include <StGLStereo/StGLStereoTexture.h>

/**
 * Simple class represents Virtual (texture) stereo Frame buffer object.
 * This allow render to texture.
 */
class ST_LOCAL StGLStereoFrameBuffer : public StGLStereoTexture {

        public:

    /**
     * Class represents simple stereo program.
     * Each program should contains:
     * - "attribute vec4 vVertex" in Vertex Shader (vertex coordinates);
     * - "attribute vec2 vTexCoord" in Vertex Shader (texture coordinates per vertex);
     * - "uniform sampler2D texR" in Fragment Shader (texture sample for Right view - texture should be bound to GL_TEXTURE0);
     * - "uniform sampler2D texL" in Fragment Shader (texture sample for Left view - texture should be bound to GL_TEXTURE1).
     */
    class StGLStereoProgram : public StGLProgram {

            private:

        StGLVarLocation atrVVertexLoc;
        StGLVarLocation atrVTexCoordLoc;

            public:

        StGLStereoProgram(const StString& title)
        : StGLProgram(title),
          atrVVertexLoc(),
          atrVTexCoordLoc() {
            //
        }

        virtual ~StGLStereoProgram() {
            //
        }

        StGLVarLocation getVVertexLoc() const {
            return atrVVertexLoc;
        }

        StGLVarLocation getVTexCoordLoc() const {
            return atrVTexCoordLoc;
        }

        virtual bool link();

    };

        private:

    StGLVertexBuffer myVerticesBuf; //!< buffers to draw simple fullsreen quad
    StGLVertexBuffer myTexCoordBuf;
    GLuint       myGLFBufferIds[2]; //!< FrameBuffer objects
    GLuint       myGLDepthRBIds[2]; //!< RenderBuffer objects for depth ID
    GLsizei            myViewPortX; //!< FBO viewport width  <= texture width
    GLsizei            myViewPortY; //!< FBO viewport height <= texture height

        private:

    /**
     * Validate FrameBuffer ids.
     */
    bool isValidFrameBuffer() const {
        return myGLFBufferIds[StGLStereoTexture::LEFT_TEXTURE ] != StGLFrameBuffer::NO_FRAMEBUFFER
            && myGLFBufferIds[StGLStereoTexture::RIGHT_TEXTURE] != StGLFrameBuffer::NO_FRAMEBUFFER;
    }

    /**
     * Validate RenderBuffer ids.
     */
    bool isValidDepthBuffer() const {
        return myGLDepthRBIds[StGLStereoTexture::LEFT_TEXTURE ] != StGLFrameBuffer::NO_RENDERBUFFER
            && myGLDepthRBIds[StGLStereoTexture::RIGHT_TEXTURE] != StGLFrameBuffer::NO_RENDERBUFFER;
    }

        public:

    /**
     * Empty constructor.
     */
    StGLStereoFrameBuffer(GLint textureFormat = GL_RGBA8);

    /**
     * OpenGL objects will be automatically released.
     * Make sure it is called from thread with bound GL context
     * or release the object in advance.
     */
    virtual ~StGLStereoFrameBuffer() {
        release();
    }

    const StGLVertexBuffer& getQuadVertices() const {
        return myVerticesBuf;
    }

    const StGLVertexBuffer& getQuadTexCoords() const {
        return myTexCoordBuf;
    }

    /**
     * Returns true if FBO was initialized.
     */
    bool isValid() const {
        return isValidFrameBuffer() && StGLStereoTexture::isValid();// && isValidDepthBuffer();
    }

    /**
     * Initialize the FBO with specified dimensions.
     */
    bool init(GLsizei textureSizeX, GLsizei textureSizeY);

    /**
     * Release OpenGL objects related to this FBO.
     */
    void release();

    /**
     * FBO viewport width.
     */
    GLsizei getVPSizeX() const {
        return myViewPortX;
    }

    /**
     * FBO viewport height.
     */
    GLsizei getVPSizeY() const {
        return myViewPortY;
    }

    /**
     * Set new FBO viewport width x height. Should be <= texture dimensions.
     */
    void setVPDimensions(GLsizei theSizeX, GLsizei theSizeY);

    /**
     * Setup OpenGL viewport equal to FBO dimensions
     */
    void setupViewPort() {
        glViewport(0, 0, myViewPortX, myViewPortY);
    }

    void bindTextureLeft(GLenum textureUnit = GL_TEXTURE0) {
        StGLStereoTexture::bindLeft(textureUnit);
    }

    void unbindTextureLeft() {
        StGLStereoTexture::unbindLeft();
    }

    void bindTextureRight(GLenum textureUnit = GL_TEXTURE0) {
        StGLStereoTexture::bindRight(textureUnit);
    }

    void unbindTextureRight() {
        StGLStereoTexture::unbindRight();
    }

    /**
     * Bind left frame buffer (to render into the left texture).
     */
    void bindBufferLeft() {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, myGLFBufferIds[StGLStereoTexture::LEFT_TEXTURE]);
    }

    void unbindBufferLeft() {
        StGLFrameBuffer::unbindBufferGlobal();
    }

    /**
     * Bind right frame buffer (to render into the right texture).
     */
    void bindBufferRight() {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, myGLFBufferIds[StGLStereoTexture::RIGHT_TEXTURE]);
    }

    void unbindBufferRight() {
        StGLFrameBuffer::unbindBufferGlobal();
    }

    void bindMultiTexture(GLenum textureUnit0 = GL_TEXTURE0, GLenum textureUnit1 = GL_TEXTURE1) {
        bindTextureLeft(textureUnit0);
        bindTextureRight(textureUnit1);
    }

    void unbindMultiTexture() {
        unbindTextureLeft();
        unbindTextureRight();
    }

    void drawQuad(const StGLStereoFrameBuffer::StGLStereoProgram* theProgram) const;

};

#endif //__StGLStereoFrameBuffer_h_
