/**
 * Copyright © 2009-2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLStereoTexture_h_
#define __StGLStereoTexture_h_

#include <StGL/StGLTexture.h>
#include <StGL/StGLVec.h>

/**
 * This class represent OpenGL stereo multitexture (Left + Right).
 */
class ST_LOCAL StGLStereoTexture {

        protected:

    enum {
        LEFT_TEXTURE  = 0,
        RIGHT_TEXTURE = 1,
    };

    StGLTexture stTextures[2];

    void setTextureFormat(GLint textureFormat) {
        stTextures[LEFT_TEXTURE ].setTextureFormat(textureFormat);
        stTextures[RIGHT_TEXTURE].setTextureFormat(textureFormat);
    }

        public:

    StGLStereoTexture(GLint textureFormat = GL_RGBA8) {
        setTextureFormat(textureFormat);
    }

    virtual ~StGLStereoTexture() {
        release();
    }

    bool isValid() const {
        return stTextures[LEFT_TEXTURE].isValid() && stTextures[RIGHT_TEXTURE].isValid();
    }

    /**
     * @param textureSizeX (GLsizei ) - each texture width;
     * @param textureSizeY (GLsizei ) - each texture height;
     * @param inDataFormat (GLenum );
     * @param pDataR (const GLubyte* );
     * @param pDataL (const GLubyte* );
     * @return true on success.
     */
    bool init(GLsizei textureSizeX, GLsizei textureSizeY, GLenum inDataFormat,
              const GLubyte* pDataR, const GLubyte* pDataL) {
        return stTextures[LEFT_TEXTURE ].init(textureSizeX, textureSizeY, inDataFormat, pDataL) &&
               stTextures[RIGHT_TEXTURE].init(textureSizeX, textureSizeY, inDataFormat, pDataR);
    }

    bool initBlack(GLsizei textureSizeX, GLsizei textureSizeY) {
        return stTextures[LEFT_TEXTURE ].initBlack(textureSizeX, textureSizeY) &&
               stTextures[RIGHT_TEXTURE].initBlack(textureSizeX, textureSizeY);
    }

    /**
     * In GL version 1.1 or greater, pData may be a NULL pointer.
     * However data of the texture will be undefined!
     */
    bool initTrash(GLsizei textureSizeX, GLsizei textureSizeY) {
        return stTextures[LEFT_TEXTURE ].initTrash(textureSizeX, textureSizeY) &&
               stTextures[RIGHT_TEXTURE].initTrash(textureSizeX, textureSizeY);
    }

    void release() {
        stTextures[LEFT_TEXTURE ].release();
        stTextures[RIGHT_TEXTURE].release();
    }

    /**
     * Bind left texture.
     */
    void bindLeft(GLenum textureUnit = GL_TEXTURE0) {
        stTextures[LEFT_TEXTURE].bind(textureUnit);
    }

    /**
     * Bind right texture.
     */
    void bindRight(GLenum textureUnit = GL_TEXTURE0) {
        stTextures[RIGHT_TEXTURE].bind(textureUnit);
    }

    void unbindLeft() const {
        stTextures[LEFT_TEXTURE].unbind();
    }

    void unbindRight() const {
        stTextures[RIGHT_TEXTURE].unbind();
    }

    /**
     * @return sizeX (GLsizei ) - texture width.
     */
    GLsizei getSizeX() const {
        return stTextures[LEFT_TEXTURE].getSizeX();
    }

    /**
     * @return sizeY (GLsizei ) - texture height.
     */
    GLsizei getSizeY() const {
        return stTextures[LEFT_TEXTURE].getSizeY();
    }

    /**
     * Change Min and Mag filter.
     * After this call current texture will be undefined.
     */
    void setMinMagFilter(const GLenum newMinMagFilter) {
        stTextures[LEFT_TEXTURE ].setMinMagFilter(newMinMagFilter);
        stTextures[RIGHT_TEXTURE].setMinMagFilter(newMinMagFilter);
    }

};

#endif //__StGLStereoTexture_h_
