/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLTextureQueue_h_
#define __StGLTextureQueue_h_

#include <StThreads/StThreads.h> // threads header (mutexes, threads,...)
#include <StThreads/StFPSMeter.h>

#include "StGLQuadTexture.h"
#include "StGLTextureData.h"

/**
 * This is specialized class to maintain continuous frames queue.
 * Notice that it designed to be accessed from multiple threads but you should use it carefully!
 * One thread should be bound valid to OpenGL context (where the frames will be rendered)
 * and another thread(s) to continuously fill this queue.
 * Method stglUpdateStTextures() should be called each rendering call from GL thread to update textures.
 * Method push() should be used to fill in queue with new frames and stglSwapFB() to pop frame from queue
 * to display.
 */
class ST_LOCAL StGLTextureQueue {

        private:

    enum {
        SWAPONREADY_NOTHING = 0,
        SWAPONREADY_SWAPPED = 1,
        SWAPONREADY_WAITLIM = 2,
    };

    StMutex         stMutexPop;
    StGLTextureData* dataFront; // queue pointers
    StGLTextureData*  dataSnap; // snapshot pointer
    StMutex        stMutexPush;
    StGLTextureData*  dataBack;

    mutable StMutex stMutexSize;
    size_t           queueSize;
    size_t        queueSizeMax;

    StGLQuadTexture stQTexture; // quad stereo texture

    StMutex      stSwapFBMutex;
    size_t         swapFBCount;

    StMutex       stMeterMutex;
    StFPSMeter      stFPSMeter;

    StMutex   stMutexSrcFromat;
    int       currentSrcFormat; // current source format

    double             ptsCurr;

    StEvent      evNewSnapShot;
    bool            isRightUpd;
    bool        isInUpdTexture; // private bools for plugin thread
    bool       isReadyToSwapFB;
    bool         isSwapOnReady; // force swap-on-ready

    int swapFBOnReady();

        public:

    StGLTextureQueue(size_t queueSizeMax);
    ~StGLTextureQueue();

    /**
     * @return stQTexture (StQuadTexture& ).
     */
    StGLQuadTexture& getQTexture() {
        return stQTexture;
    }

    /**
     * Function push stereo frame into queue.
     * This function called ONLY from video thread.
     * @param srcDataLeft  (const StImage& ) - first  INPUT data (Both or Left);
     * @param srcDataRight (const StImage& ) - second INPUT data (NULL or Right);
     * @param theStParams  (const StHandle<StStereoParams>& ) - stereo parameters;
     * @param srcFormat (StFormatEnum ) - source data format;
     * @param srcPTS (double ) - PTS (presentation timestamp);
     * @return true on success.
     */
    bool push(const StImage& srcDataLeft,
              const StImage& srcDataRight,
              const StHandle<StStereoParams>& theStParams,
              StFormatEnum srcFormat,
              double srcPTS);

    double getPlaybackFPS() {
        stMeterMutex.lock();
        double fps = stFPSMeter.getAverage();
        bool isUpdated = stFPSMeter.isUpdated();
        stMeterMutex.unlock();
        if(isUpdated) {
            ST_DEBUG_LOG(
                + "Queue playback FPS " + fps
                + ", buffers: " + (queueSize + 1) + "/" + queueSizeMax
            );
        }
        return fps;
    }

    /**
     * Function called in loop from general GL draw loop
     * and do update quad texture data / state (display frame).
     * @return true on each full update.
     */
    bool stglUpdateStTextures();

    size_t getSize() {
        stMutexSize.lock();
            size_t result = queueSize;
        stMutexSize.unlock();
        return result;
    }

    /**
     * @return true if queue is EMPTY.
     */
    bool isEmpty() {
        stMutexSize.lock();
            bool result = (queueSize == 0);
        stMutexSize.unlock();
        return result;
    }

    /**
     * @return true if queue is FULL.
     */
    bool isFull() {
        stMutexSize.lock();
            bool result = ((queueSize + 1) == queueSizeMax);
        stMutexSize.unlock();
        return result;
    }

    double getPTSCurr() const {
        stMutexSize.lock();
            double p = ptsCurr;
        stMutexSize.unlock();
        return p;
    }

    /**
     * @param outPts (double& ) - next (front) stereo frame PTS (presentation timestamp);
     * @return false if next PTS not available.
     */
    bool popPTSNext(double& outPts) {
        bool res = false;
        stMutexPop.lock();
        stMutexPush.lock();
        stMutexSize.lock();
            // make sure front data is valid
            if(queueSize < 1) {
                res = false;
            } else {
                res = true;
                outPts = dataFront->getPTS();
            }
        stMutexSize.unlock();
        stMutexPush.unlock();
        stMutexPop.unlock();
        return res;
    }

    /**
     * Send event to swap texture buffer Back->Front (increase swap counter).
     * @param limit (size_t ) - swap counter limit;
     * @return true if swap counter increased.
     */
    bool stglSwapFB(size_t limit) {
        stSwapFBMutex.lock();
        if(limit == 0 || swapFBCount < limit) {
            ++swapFBCount;
            stSwapFBMutex.unlock();
            return true;
        }
        stSwapFBMutex.unlock();
        return false;
    }

    /**
     * Function process TOTAL queue clean up.
     */
    void clear();

    /**
     * This function clean up only requested number of frames but prevents queue emptying.
     * At least one frame will remain in queue.
     */
    void drop(const size_t& count);

    /**
     * Function used to get current showed source format.
     * At this moment function used just for stereo/mono recognizing.
     */
    int getSrcFormat() {
        stMutexSrcFromat.lock();
            // TODO (Kirill Gavrilov#4#) source format should be defined like front PTS to prevent early changes
            int srcFrmt = currentSrcFormat;
        stMutexSrcFromat.unlock();
        return srcFrmt;
    }

    enum {
        SNAPSHOT_NO_NEW = 0,
        SNAPSHOT_SUCCESS = 1,
    };

    int getSnapshot(StImage* outDataLeft, StImage* outDataRight, bool isForce = false);

};

#endif //__StGLTextureQueue_h_
