/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __stTypes_h_
#define __stTypes_h_

#include <stconfig.conf>

/**
 * Below are compiler-dependent useful macros to warn
 * when some parts of code are used in wrong way.
 * Note that to enable annotations using MSVC you should add compiler option /analyze.
 *
 * Available attributes:
 *  - ST_ATTR_CHECK_RETURN. To warn if a caller of the function does not use its return value.
 *    Notice that this works only with primitive types (useless for returned structure).
 *    Usage:
 *      ST_ATTR_CHECK_RETURN int doInit() { return 0; }
 *      doInit(); // warning will appear
 */
#if defined(_MSC_VER) && (_MSC_VER >= 1400)
    //#define _USE_ATTRIBUTES_FOR_SAL 1 // doesn't work without /analyze option!
    #include <CodeAnalysis/SourceAnnotations.h>
    //#define ST_ATTR_CHECK_RETURN [returnvalue:vc_attributes::Post(MustCheck=vc_attributes::Yes)]
    //#include <sal.h>
    #define ST_ATTR_CHECK_RETURN _Check_return_
#elif defined(__GNUC__)
    #define ST_ATTR_CHECK_RETURN __attribute__ ((warn_unused_result))
#else
    #define ST_ATTR_CHECK_RETURN
#endif

#include <cmath>       // fabs
#include <cstddef>     // size_t, NULL
#include <cstdlib>
#include <cstring>     // for memcpy
#include <xmmintrin.h> // for memory alignment
#include <stdio.h>     // for _snprintf on MinGW

/**
 * Define the default alignment boundary.
 * We define it to 16 bytes (needed for SSE instructions).
 */
#ifndef ST_ALIGNMENT
    #define ST_ALIGNMENT 16
#endif

/**
 * Macro for function arguments not used in function body.
 */
#define ST_UNUSED(theParam)

// activate some C99 macros like UINT64_C in "stdint.h" which used by FFmpeg
#ifndef __STDC_CONSTANT_MACROS
    #define __STDC_CONSTANT_MACROS
#endif

// int16_t/int32_t and so on
#if((defined(_WIN32) || defined(__WIN32__)) && (_MSC_VER > 0 && _MSC_VER < 1600))
    // old MSVC - hasn't stdint header
    #include <sysForVC/stdint.h>
#else
    #include <stdint.h>
#endif
//#include <cstdint> // C++0x

#ifdef __cplusplus
    typedef bool stBool_t;
    static const stBool_t ST_TRUE  = true;
    static const stBool_t ST_FALSE = false;
#else
    // C-consistent bool type (C doesn't have true 'bool' type)
    typedef unsigned char stBool_t;
    static const stBool_t ST_TRUE  = (stBool_t )1;
    static const stBool_t ST_FALSE = (stBool_t )0;
#endif

// some primitives
typedef signed   int  stInt_t;
typedef unsigned int  stUInt_t;
typedef signed   char stByte_t;
typedef unsigned char stUByte_t;

// fixed-size primitives
typedef int32_t  stInt16_t;
typedef uint32_t stUInt16_t;
typedef int32_t  stInt32_t;
typedef uint32_t stUInt32_t;
typedef int64_t  stInt64_t;
typedef uint64_t stUInt64_t;

/**
 * Mordern compilers provide fixed-size primitives like int32_t.
 * Some developers can use these types in function overloads
 * to logically cover all variations depending on primitive size.
 * However some language primitives can recall to the same
 * absolute type but distinct by compiler.
 * For example int and long int both are 32-bit integers on ILP32.
 * Defines below could be used to extend overloaded functions list
 * to support both reincarnations available in compiler when
 * only bitness distinctness is needed.
 * Notice that language can provide more similar types (like char32_t).
 */
#if (defined(_LP64) || defined(__LP64__))
    #define ST_HAS_INT64_EXT
    #if (defined(__APPLE__))
        typedef signed   long stInt64ext_t;
        typedef unsigned long stUInt64ext_t;
    #else
        typedef signed   long long stInt64ext_t;
        typedef unsigned long long stUInt64ext_t;
    #endif
#else
    #define ST_HAS_INT32_EXT
    typedef signed   long stInt32ext_t;
    typedef unsigned long stUInt32ext_t;
#endif

/**
 * In the Windows API, the maximum length for a path is MAX_PATH, which is defined as 260 characters.
 * A local path is structured in the following order: drive letter, colon, backslash, components separated by backslashes,
 * and a terminating null character. For example, the maximum path on drive D is "D:\<some 256 character path string><NUL>"
 * where "<NUL>" represents the invisible terminating null character for the current system codepage.
 * (The characters < > are used here for visual clarity and cannot be part of a valid path string.)
 * The Windows API has many functions that also have Unicode versions to permit an extended-length path for a maximum total path length
 * of 32,767 characters. This type of path is composed of components separated by backslashes, each up to the value returned
 * in the lpMaximumComponentLength parameter of the GetVolumeInformation function (this value is commonly 255 characters).
 * To specify an extended-length path, use the "\\?\" prefix. For example, "\\?\D:\<very long path>". (The characters < > are used here
 * for visual clarity and cannot be part of a valid path string.)
 * Note: The maximum path of 32,767 characters is approximate, because the "\\?\" prefix may be expanded to a longer string by the system
 * at run time, and this expansion applies to the total length.
 **/
#define ST_MAX_PATH 4096

/**
 * Unicode primitives.
 */
typedef char          stUtf8_t;     //!< signed   UTF-8 char is just a byte
typedef unsigned char stUtf8u_t;    //!< unsigned UTF-8 char is just a byte
#if (!defined(__GNUC__) || ((__GNUC__ == 4) && (__GNUC_MINOR__ > 2)) || (__GNUC__ > 4)) \
 || ((defined(_WIN32) || defined(__WIN32__)) && (_MSC_VER >= 1600))
    typedef char16_t  stUtf16_t;    //!< UTF-16 char (always unsigned)
    typedef char32_t  stUtf32_t;    //!< UTF-32 char (always unsigned)
#else // obsolete compilers compatibility
    typedef uint16_t  stUtf16_t;    //!< UTF-16 char (always unsigned)
    typedef uint32_t  stUtf32_t;    //!< UTF-32 char (always unsigned)
#endif
typedef wchar_t       stUtfWide_t;  //!< wide char (unsigned UTF-16 on Windows platform and signed UTF-32 on Linux)

/**
 * Let len be the length of the formatted data string (not including the terminating null). len and count are in bytes for _snprintf, wide characters for _snwprintf.
 * If len < count, then len characters are stored in buffer, a null-terminator is appended, and len is returned.
 * If len = count, then len characters are stored in buffer, no null-terminator is appended, and len is returned.
 * If len > count, then count characters are stored in buffer, no null-terminator is appended, and a negative value is returned.
 * @param argBuffer (char* ) - string buffer;
 * @param argCount (size_t ) - maximum number of characters to store;
 * @param argFormat (const char* ) - string format;
 * @param ... - arguments to format;
 * @return len (int ).
 */
#if(defined(_WIN32) || defined(__WIN32__))
    #define stsprintf(argBuffer, argCount, argFormat, ...); _snprintf(argBuffer, argCount, argFormat, __VA_ARGS__);
#else
    #define stsprintf(argBuffer, argCount, argFormat, ...);  snprintf(argBuffer, argCount, argFormat, __VA_ARGS__);
#endif

/**
 * Define the big-endian flag. If your compiler doesn't provide identification defines
 * - use stconfig.conf to override this. This macro is really important for some functions
 * to work correctly on big-endian CPUs (or bi-endian CPUs with big-endian enabled in OS)!
 */
#ifndef ST_HAVE_BIGENDIAN
    #if (defined(BYTE_ORDER) && BYTE_ORDER==BIG_ENDIAN) || \
        (defined(__BYTE_ORDER) && __BYTE_ORDER==__BIG_ENDIAN) || \
        defined(__BIG_ENDIAN__)
        #define ST_HAVE_BIGENDIAN
    #endif
#endif

/**
 * OpenGL types, just typedef them here (be sure them always in sync - check GLEW headers after update!)
 */
typedef unsigned int GLenum;
typedef unsigned int GLuint;
typedef int GLint;
typedef int GLsizei;
typedef unsigned char GLboolean;
typedef signed char GLbyte;
typedef unsigned char GLubyte;
typedef float GLfloat;
typedef double GLdouble;

/**
 * Dummy structure definition.
 */
typedef struct stNoType {} stNoType;

/**
 * @param value1 (const GLfloat ) - first value for equal test;
 * @param value2 (const GLfloat ) - second value for equal test;
 * @param tolerance (const GLfloat ) - precision for equal test;
 * @return true if test values equal with given precision.
 */
inline bool stAreEqual(const GLfloat value1, const GLfloat value2,
                       const GLfloat tolerance) {
    return (std::abs(value1 - value2) <= tolerance);
}

inline bool stAreEqual(const void* theBuffer1, const void* theBuffer2,
                       size_t theBytes) {
    if(theBytes % 4 == 0) {
        theBytes /= 4;
        int32_t* aBuffer1 = (int32_t* )theBuffer1;
        int32_t* aBuffer2 = (int32_t* )theBuffer2;
        for(; theBytes != 0; --theBytes, ++aBuffer1, ++aBuffer2) {
            if(*aBuffer1 != *aBuffer2) {
                return false;
            }
        }
    } else {
        stUByte_t* aBuffer1 = (stUByte_t* )theBuffer1;
        stUByte_t* aBuffer2 = (stUByte_t* )theBuffer2;
        for(; theBytes != 0; --theBytes, ++aBuffer1, ++aBuffer2) {
            if(*aBuffer1 != *aBuffer2) {
                return false;
            }
        }
    }
    return true;
}

/**
 * Copies the values from the location pointed by source directly to the memory block pointed by destination.
 * @param destinationPtr (void* ) - pointer to the destination array where the content is to be copied, type-casted to a pointer of type void*;
 * @param sourcePtr (const void* ) - pointer to the source of data to be copied, type-casted to a pointer of type void*;
 * @param bytesCount (const size_t& ) - number of bytes to copy;
 * @return destinationPtr.
 */
inline void* stMemCpy(void* destinationPtr, const void* sourcePtr, const size_t& bytesCount) {
    return std::memcpy(destinationPtr, sourcePtr, bytesCount);
}

/**
 * Sets the first bytes of the block of memory pointed by ptr to the specified value (interpreted as an unsigned char).
 * @param ptr (void* ) - pointer to the block of memory to fill;
 * @param value (int ) - value to be set, the value is passed as an int, but the function fills the block of memory using the unsigned char conversion of this value;
 * @param bytesCount (const size_t& ) - number of bytes to be set to the value;
 * @return ptr.
 */
inline void* stMemSet(void* ptr, int value, const size_t& bytesCount) {
    return std::memset(ptr, value, bytesCount);
}

/**
 * Allocates a block of 'bytesCount' bytes of memory, returning a pointer to the beginning of the block.
 * You should deallocate memory with stMemFree().
 * @param bytesCount (const size_t& ) - size of the memory block, in bytes;
 * @return on success, a pointer to the memory block allocated by the function; null pointer otherwise.
 */
inline void* stMemAlloc(const size_t& bytesCount) {
    return std::malloc(bytesCount);
}

inline void* stMemRealloc(void* thePtr, const size_t& bytesCount) {
    return std::realloc(thePtr, bytesCount);
}

/**
 * Deallocate space in memory.
 * @param ptr (void* ) - pointer to a memory block previously allocated with stMemAlloc() to be deallocated.
 */
typedef void (*stMemFree_t)(void* );
inline void stMemFree(void* ptr) {
    return std::free(ptr);
}

/**
 * Allocates an ALIGNED block in memory.
 * For SSE commands alignment should be 16.
 * You should use stMemFreeAligned() to deallocate memory.
 * @param bytesCount (const size_t& ) - size of the memory block, in bytes;
 * @param align (const size_t& ) - alignment constraint, must be a power of two;
 * @return on success, a pointer to the memory block allocated by the function; null pointer otherwise.
 */
inline void* stMemAllocAligned(const size_t& bytesCount,
                               const size_t& align = ST_ALIGNMENT) {
#if defined(_MSC_VER)
    return _aligned_malloc(bytesCount, align);
#else
    return _mm_malloc(bytesCount, align);
#endif
}

/**
 * Same as stMemAllocAligned() but fill the allocated block with zeros;
 */
inline void* stMemAllocZeroAligned(const size_t& bytesCount,
                                   const size_t& align = ST_ALIGNMENT) {
    void* aPtr = stMemAllocAligned(bytesCount, align);
    if(aPtr != NULL) {
        stMemSet(aPtr, 0, bytesCount);
    }
    return aPtr;
}

inline void* stMemReallocAligned(void* ptrAligned, const size_t& bytesCount,
                                 const size_t& align = ST_ALIGNMENT) {
#if defined(_MSC_VER)
    return _aligned_realloc(ptrAligned, bytesCount, align);
#else
    /// TODO (Kirill Gavrilov#4) what should we call here???
    return realloc(ptrAligned, bytesCount);
#endif
}

/**
 * Deallocate space in memory.
 * @param ptr (void* ) - pointer to a memory block previously allocated with stMemAllocAligned() to be deallocated.
 */
inline void stMemFreeAligned(void* ptrAligned) {
#if defined(_MSC_VER)
    _aligned_free(ptrAligned);
#else
    _mm_free(ptrAligned);
#endif
}

// namespaces in export-headers
#ifdef __cplusplus
    #define ST_NAMESPACE_START(name) namespace name {
    #define ST_NAMESPACE_END(name) }
#else
    #define ST_NAMESPACE_START(name)
    #define ST_NAMESPACE_END(name)
#endif

// Add EXPORT or IMPORT prefix before every exported function
// C-style used
#ifdef __cplusplus
    #define EXTERN_C_PREFIX extern "C"
#else
    #define EXTERN_C_PREFIX
#endif

#if(defined(_WIN32) || defined(__WIN32__))
    #define ST_EXPORT EXTERN_C_PREFIX __declspec(dllexport)
    #define ST_IMPORT EXTERN_C_PREFIX __declspec(dllimport)
    #define ST_LOCAL
#elif(__GNUC__ >= 4)
    #define ST_EXPORT EXTERN_C_PREFIX __attribute__ ((visibility("default")))
    #define ST_IMPORT EXTERN_C_PREFIX __attribute__ ((visibility("default")))
    #define ST_LOCAL __attribute__ ((visibility("hidden")))
#else
    #define ST_EXPORT EXTERN_C_PREFIX
    #define ST_IMPORT EXTERN_C_PREFIX
    #define ST_LOCAL
#endif

/**
 * Returns nearest (greater or equal) aligned number.
 */
inline size_t getAligned(const size_t theNumber, const size_t theAlignment = ST_ALIGNMENT) {
    return theNumber + (theNumber % theAlignment);
}

/**
 * Check number for alignment.
 */
inline bool isAligned(const size_t theNumber, const size_t theAlignment = ST_ALIGNMENT) {
    return (theNumber % theAlignment) == 0;
}

/**
 * Simple function for getting power of to number
 * larger or equal to input number.
 * @param inNumber (const size_t& ) - number to 'power of two';
 * @param threshold (const size_t& ) - upper threshold;
 * @return (size_t ) power of two number.
 */
inline size_t getPowerOfTwo(const size_t inNumber, const size_t threshold) {
    for(size_t p2 = 2; p2 <= threshold; p2 <<= 1) {
        if(inNumber <= p2) {
            return p2;
        }
    }
    return threshold;
}

inline GLsizei getPowerOfTwo(const GLsizei inNumber, const GLsizei threshold) {
    for(GLsizei p2 = 2; p2 <= threshold; p2 <<= 1) {
        if(inNumber <= p2) {
            return p2;
        }
    }
    return threshold;
}

inline bool isOddNumber(const int64_t theNumber) {
    return theNumber & 0x01;
}

inline bool isEvenNumber(const int64_t theNumber) {
    return !(isOddNumber(theNumber));
}

inline bool isOddNumber(const size_t number) {
    return number & 0x01;
}

inline bool isEvenNumber(const size_t number) {
    return !(isOddNumber(number));
}

/**
 * Simple function for getting even number larger then input number.
 * @param number (const size_t ) - number to even;
 * @return (size_t ) even number.
 */
inline size_t getEvenNumber(const size_t number) {
    return isOddNumber(number) ? (number + 1) : number;
}

inline bool isOddNumber(const int number) {
    return number & 0x01;
}

inline bool isEvenNumber(const int number) {
    return !(isOddNumber(number));
}

inline int getEvenNumber(const int number) {
    return isOddNumber(number) ? (number + 1) : number;
}

#ifdef __cplusplus
    #include <StTemplates/StTemplates.h> // include commonly-used templates
#endif

#endif //__stTypes_h_
